# 二、Java 有而 Go 没有的东西

Java 有一些 Go 没有的特性，反之亦然。所以，在我们看一些在 Go 中有相同功能的 Java 特性之前，让我们先简单看一下 Go 没有的特性。并不是每一个 Java 拥有而 Go 没有的特性都可以列出来，但是下面总结了一些关键的特性。

请注意，许多 Go“缺失”的特性被故意省略，以保持语言的简单和高效，而不是因为它们难以提供。这被认为是一种美德。

## 多重任务

Java 可以在一条语句中将多个变量赋给相同的值。例如:

```go
int x, y, z;
x = y = z = 10;

```

最接近的 Go 是

```go
var x, y, z int = 10, 10, 10

```

在 Go 中，分配的类型和值可以不同。

## 语句和运算符

Go 和 Java 操作符的优先级不同。在作者看来，Go 的优先级更少，也更自然。当有疑问时，使用括号将表达式括起来以确保正确。

一个关键的区别是，在 Go 中`x++`(表示:x = x + 1)和`x--`(表示:x = x - 1)是语句，而不是运算符。而且根本没有`--x`或者`++x`的表情。

Go 不支持*三元表达式*。需要使用 if/else 语句。例如，获取较大值的 Java 表达式

```go
var z = x > y ? x : y;

```

在 Go 中需要像下面这样的东西:

```go
var z = y
if x > y {
   z = x
}

```

相似但不相同。你也可以这样做:

```go
var z int
if x > y { z = <some expensive int expression> }
else { z = <some other expensive int expression>}

```

注意前面的`if/else`必须在一个源代码行中输入。

Go 不支持赋值表达式，只支持赋值语句。

## 断言语句

Go 没有`assert`语句。一般来说，Go 有 *panics* 可以用来实现类似的功能，但是它们不能像断言那样在编译时被禁用。因此，不鼓励这样使用恐慌。

## While 和 Do 语句

Java `while`语句被 Go `for`语句取代(即`for`的行为类似于`while`)。Java `do`语句没有直接的等价物，但是`for`语句可以用来代替它。

注意，Java `for`语句也可以用作`while`语句。

例如:

```go
var x = 0; for(; x < 10;) { ... ; x++; }

```

与...相同

```go
var x = 0; while(x < 10) { ... ; x++; }

```

## Throw 语句/Throws 子句

Go 没有`throw`语句(或`throws`子句)。Go `panic` `(...)`功能的作用与投掷动作类似。

## Strictfp、瞬态、易变、同步、抽象、静态

Go 没有这些 Java 修饰符的等价物。大多数是不需要的，因为在 Java 中需要它们的问题在 Go 中以不同的方式解决了。例如，通过将声明的值作为顶级(也称为包)值来实现静态值的等效。

## 对象和类(OOP)和内部类，Lambdas，this，super，Explicit 构造函数

Go 不像 Java 那样完全支持*面向对象编程* (OOP)。因此，它不支持这些 Java 结构。Go 具有本文稍后描述的特性，可以类似于这些 OOP 特性中的大部分来使用。因此，更好的描述是，Go 是一种基于对象的语言*。Go 确实允许一个人实现 OOP 的关键目标，但是以不同于严格的 OOP 语言通常会做的方式。*

Go 不支持真正的*类*(即 Java `class`声明)。Go 确实支持*结构*，它们类似于类，但没有继承。Go 确实允许嵌套结构，这有点像内部类。

Go 在类型声明中没有`extends`或`implements`子句。按照这些条款的规定，Go 没有继承。Go 的接口类型确实有一种隐含形式的`implements`。

Go 不支持 Java *Lambdas* (编译成类实例的函数签名)。相反，Go 支持可以作为参数传递的*一级函数*(通常是文字)。Go 不支持*方法引用*(作为参数传递的 lambdas 的简单名称)。

Go 支持接口的方式与 Java 不同。Go 的接口允许*鸭子打字*。Go 的接口不要求显式实现(Go 中不需要`implements`子句)；任何具有与接口的所有方法相匹配的方法的类型都会隐式实现接口。总的来说，Go 的做法更加灵活。

Java 8 和更高版本允许在接口中实现(具体的)方法。Go 不会。Java 允许在接口中声明常量；Go 不会。Java 允许在接口中声明子类型。Go 不会。

考虑 OOP 的这些租户:

1.  一个对象有一个标识(它可以与所有其他对象区分开来)。

2.  一个对象可能(通常确实)有状态(也称为实例数据、属性或字段)。

3.  一个对象可能(通常确实)有行为(也称为成员函数或方法)。

4.  一个对象由一个称为类的模板来描述/定义。

5.  类可以被安排在一个(继承)层次结构中；实例是层次结构中类的组合。

6.  对象实例被封装；状态通常仅通过方法可见。

7.  变量可以在类层次结构中的任何级别声明；子类的实例可以分配给这些变量(多态性)。

Java 支持(但不一定强制)所有前面的租户。Go 不会。Go 对这些租户的支持如下:

1.  struct 实例有一个地址，该地址通常可以作为其标识(但可能不总是唯一的)；结构实例类似于对象实例，但并不完全相同。

2.  结构实例可能(通常确实)有状态。

3.  结构实例可能(通常)有行为。

4.  像类一样，结构实例由称为结构类型的模板描述/定义。

5.  不直接支持；结构可以嵌入提供类似组合的其他结构。

6.  支持但通常不使用(结构字段通常是公共的)。

7.  不支持。

从历史上看，面向对象程序设计语言源于计算机模拟 <sup>[1](#Fn1)</sup> 和改善人机交互的愿望。<sup>T5】2</sup>OOP 语言被设想用来实现模拟对象之间的消息传递以影响行为。随着 OOP 改进的行为重用可能性(即继承)变得众所周知，它作为一种编程风格越来越受欢迎。大多数现代语言都提供了这种能力。

对许多人来说，Go 缺乏完整的 OOP 可能是它最大的缺点。但是作者希望，一旦你习惯于做习惯性的 Go 编程，你就不会像最初想的那样怀念 OOP 的特性。Go 是一种设计良好、功能丰富的语言，它支持 OOP 的目标，而不包含其他语言(如 Java)复杂的 OOP 特性。

请考虑 OOP 并不是写好程序所必须的。所有现存的 C 程序，一些大而丰富的，如操作系统 <sup>[3](#Fn3)</sup> 和网络浏览器，证明并非如此。事实上，有时候 OOP 思维会在程序上强加不适当的结构。再说一遍，Go 是一种类 C 语言。

实现高水平的重用不需要 OOP。函数可以很好地扮演这个角色，尤其是当它们是第一流的时候。

## 泛型类型和方法

Go 目前不支持*泛型*类型和任意类型上的方法。这里，泛型意味着能够持有/使用多种类型。在 Java 中，`Object`类型的变量是泛型的，因为它可以保存任何引用类型的值。在 Go 中，`interface{}`类型的变量是通用的，因为它可以保存任何类型的值。

Java 5 细化了这个概念，声明的类型(比如容器类)可以被指定为只支持特定的(而不是所有的)类型(比如字符串或数字)作为容器类型的修饰符，例如，`List<String>`(而不是仅仅`List`)类型。Go 的内置集合类型(切片、地图和通道)在这方面是通用的。

最初，Java 不支持特定类型的泛型类型。它们是在 Java 5 中引入的，主要是为了缓解该语言中存在的集合的某些可用性问题。由于向后兼容性，Java 的通用设计有一些不理想的特性/妥协。

目前，有一个关于在 Go 中添加泛型的提议得到了批准，其原因与在 Java 中添加泛型的原因大致相同。看起来 Go 将会步 Java 的后尘。

Java(和 Go)定义的泛型类型主要是为了去除重复编码的语法糖。在 Java 中，它们根本不会影响运行时代码(因为运行时类型擦除)。在 Go 中，它们可能会导致可执行文件中存在更多的二进制代码，但不会比手动模拟更多。

## 广泛的函数式编程能力

Go 支持一级函数，但不支持典型的广义效用函数(map、reduce、select、exclude、forEach、find 等。)最具功能性(强烈支持*功能性编程*范例)的语言和 Java(通过其 Lambdas 和 Streams 支持)提供。这种省略是 Go 语言设计者故意做出的决定。当包含泛型时，Go 可能会添加一些这样的实用函数。

## 原始价值观的拳击

Java 集合(数组除外)不能包含原始值，只能包含对象。因此，Java 为每种原始类型提供了包装器类型。为了使集合更容易使用，Java 会自动将一个原语包装(装箱)到一个包装器类型中，以将其插入到一个集合中，并在从集合中取出该值时将其解包(取消装箱)。Go 支持可以保存原语的集合，所以不需要这样的装箱。注意需要使用装箱是 Java 在内存使用方面不如 Go 有效的地方。

## 源注释

Go 没有*注释*。Go Struct 字段可以有*标签*，它提供了一个类似的但是更加有限的角色。

注释，以及函数流和 lambdas，使 Java(至少部分地)成为一种声明性语言。 <sup>[4](#Fn4)</sup> Go 几乎纯粹是一种命令式语言。 <sup>[5](#Fn5)</sup> 这是通过选择。这往往会使 Go 代码变得更加明显和冗长。

Note Go 与 Java 编译时注释有类似的概念，源文件可以包含特殊的注释(称为*构建约束*),构建器解释这些注释以改变代码的处理方式。例如，要为其生成代码的目标操作系统可以在源文件的最开头通过这样的注释来指定:

```go
// +build linux,386

```

这将导致该文件仅适用于 Linux 操作系统(OS)和基于 386 的体系结构。

有一种替代的(通常是首选的)语法；前面的注释也可以写成

```go
//go:build linux,386

```

注意，一些约束条件，比如目标操作系统和/或硬件架构，可以嵌入到 Go 文件名中。例如

```go
xxx_windows.go

```

将只为 Windows 操作系统构建。

## 多重可见性

Java 支持四种可视性:

1.  private–只有包含类型中的代码可以看到它。

2.  默认–只有同一个包中的代码可以看到它。

3.  protected–只有同一包或该类型的子类中的代码才能看到它。

4.  public–任何代码都可以看到它。

Go 只支持默认可见性(在 Go 中通常称为 private 或 package)和公共可见性。地鼠通常将公共可见性称为“导出可见性”，将私有可见性称为“未导出可见性”

## 过载/覆盖的函数

在 Java 中，可以在同一个作用域中定义名称相同但签名不同(参数数量和/或类型不同)的函数。这些被称为(通过一种参数多态性的形式)*重载*函数。Go 不允许重载。

在 Java 中，具有相同名称和签名的函数可以在继承层次结构中被重新定义。这种重新定义的功能被称为(通过继承多态性)*覆盖*。因为 Go 不支持继承，所以不允许这样的覆盖。

## 正式列举

Java 有正式的*枚举*类型，它们是特殊用途的类类型，具有离散的静态实例，以便于与 sameness ( `==`)操作符进行比较。Go 不会。相反，它对整数类型的常量使用了`iota`运算符。在 Java 中，枚举值可以基于几种类型(但整数是常见的)；在 Go 中，只允许整数类型。

注意，Java 枚举是类，可以像任何其他类一样拥有字段和方法。他们也支持继承。Go 枚举没有类似的特性。

## 内置二进制数据自序列化

Java 提供了以二进制形式序列化(转换为字节序列，在这个用例中通常称为*八位字节* <sup>[6](#Fn6)</sup> )数据和对象的能力。`Data{Input|Output}Stream`和(子类)`Object{Input|Output}Stream`类型提供 API 来做这件事。序列化数据通常被写入文件或通过网络传输，有时存储在数据库中。序列化可以为原本短暂的对象提供一种持久性形式。序列化也是大多数*远程过程调用* <sup>[7](#Fn7)</sup> (RPC)机制的基础。

Java 支持原语值、数组和任何包含原语类型或任何标有`Serializable`接口的类型以及这些类型的任何集合的数据结构(类实例)的序列化。Java 甚至支持带有引用循环的结构。

Go 没有提供这种完全对象序列化的直接等价物。在 Go 中，人们通常将数据序列化为某种文本格式(比如 JSON 或 XML ),然后保存/发送该格式。使用文本通常不如二进制表示有效(需要更多的字节和时间)。这些文本形式通常不支持数据结构中的引用循环。

Go 提供社区支持，比如针对二进制数据的 Google *协议缓冲区*、<sup>、T3】 8 、T5。有了标准的 Go 库，人们可以创建定制的二进制格式，这有点乏味。</sup>

## 并发收款

Java 有许多集合实现，每一个都为不同的用例提供了细微的优化。Go 采用了一种更简单的方法，像 Python 和 JavaScript 等其他语言一样，在所有用例中使用单个集合实现，比如一个列表或地图。这在运行时可能不是最理想的，但是它更容易学习和使用。

除了标准的等价物之外，Java 还有几个并发(在多线程中使用时性能良好(低争用))类型和集合。`ConcurrentHashMap`大概是最通俗的例子。Go 有一些标准的等价库，比如`sync.Map`类型。一般来说，这种并发类型在 Go 中的使用频率较低。经常使用替代方法，如通道。

<aside aria-label="Footnotes" class="FootnoteSection" epub:type="footnotes">Footnotes [1](#Fn1_source)

[T2`https://en.wikipedia.org/wiki/Simula`](https://en.wikipedia.org/wiki/Simula)

  [2](#Fn2_source)

[T2`https://en.wikipedia.org/wiki/Smalltalk`；Smalltalk 推出了图形用户界面(GUI)。](https://en.wikipedia.org/wiki/Smalltalk)

  [3](#Fn3_source)

[T2`www.toptal.com/c/after-all-these-years-the-world-is-still-powered-by-c-programming`](http://www.toptal.com/c/after-all-these-years-the-world-is-still-powered-by-c-programming)

  [4](#Fn4_source)

[T2`https://en.wikipedia.org/wiki/Declarative_programming`](https://en.wikipedia.org/wiki/Declarative_programming)

  [5](#Fn5_source)

[T2`https://en.wikipedia.org/wiki/Imperative_programming`](https://en.wikipedia.org/wiki/Imperative_programming)

  [6](#Fn6_source)

[T2`https://en.wikipedia.org/wiki/Octet_(computing)`](https://en.wikipedia.org/wiki/Octet_%2528computing%2529)

  [7](#Fn7_source)

[T2`https://en.wikipedia.org/wiki/Remote_procedure_call`](https://en.wikipedia.org/wiki/Remote_procedure_call)

  [8](#Fn8_source)

[T2`https://en.wikipedia.org/wiki/Protocol_Buffers`](https://en.wikipedia.org/wiki/Protocol_Buffers)

 </aside>