# 三、Go 和 Java 的深入比较

这一章深入探讨了 Go 与 Java 的早期介绍。它更详细地描述了 Java 和 Go 之间的显著差异。通过比较 Go 和 Java，人们可以更容易地吸收 Go 的特性。

Go 是(在作者看来)比 Java 简单得多的语言；可以说，Go 甚至是比 c 更简单的语言。例如， *Java 语言规范*目前大约有 800 页长，而 *Go 语言规范*目前大约有 85 <sup>[1](#Fn1)</sup> 页长。显然，Java 比 Go 有更多的语言复杂性。

Go 标准库也是如此。就提供的类型和函数的数量以及纯粹的代码行而言，它们比 Java 标准库小得多。在某些方面，Go 库功能较少，但即使如此，它们的功能一般足以编写许多有用的程序。

与 Java 社区一样，标准库中未包含的功能通常由社区成员提供。在作者看来，Java 库，尤其是社区提供的库，通常比许多相应的 Go 库更成熟。

Java 库通常也更重(做得更多),比相应的 Go 库更难学习和使用。一般来说，对于典型的 Go 用例，Go 库更“大小合适” <sup>[2](#Fn2)</sup> ，因此，Go 并不缺乏其适用性。考虑到标准 Java 库的大代码库大小迫使 Java 9 将它们分成可选择的模块，这样可以减少 Java 运行时的内存占用。此外，为了进一步减小运行时的大小，许多旧的库已经被弃用(有些现在已经被删除)。

Go 社区大多由 Google 和许多个人或小团队组成。它拥有更少的审查机构，如阿帕奇软件基金会、为 Java 开发关键的第三方库和框架的机构，如阿帕奇软件基金会、阿帕奇软件基金会、阿帕奇软件基金会、阿帕奇软件基金会、阿帕奇软件基金会、阿帕奇软件基金会、阿帕奇软件基金会、阿帕奇软件基金会、阿帕奇软件基金会、阿帕奇软件基金会、阿帕奇软件基金会、阿帕奇软件基金会、阿帕奇软件基金会、阿帕奇软件基金会、阿帕奇软件基金会和阿帕奇软件基金会。

Go 和 Java <sup>[5](#Fn5)</sup> 支持相似但不同的语句和数据类型。它们总结如下。它们将在本文后面更详细地描述。

Go 和 Java 都支持布尔值和字符、整数和浮点数。在 Go 中，一个字符称为一个`rune`，为 4 个字节；Java 里叫`char`，2 个字节。两者都使用 Unicode 编码。一般来说，Go 对符文的使用比 Java 的`char`类型更好，因为任何字符变量都可以表示任何合法的 Unicode 字符。

Java 和 Go 都支持字符串类型，它们实际上是字符数组。在 Go 中，字符串是一种原始类型。Go 在字符串中使用的 *Unicode* *转换格式* <sup>[7](#Fn7)</sup> (UTF-8)允许许多字符串，特别是对于英文文本，使用比同等的 Java 字符串更少的字节。

在每种语言中，这些类型上的运算符都是相似的。Go 还支持复杂的浮点数，而 Java 不支持。Java 支持大形式的整数和十进制浮点数。Go 支持大形式的整数和二进制浮点数。Go 和 Java 都支持同质值数组。Java 聚合类中的异构值；Go 使用结构。

Java 支持对类实例的引用。Go 使用指针，可以定位任何类型的值。

Java 和 Go 有很多相似的语句。

两者都有赋值语句。两者都增加了(操作员参与的)任务。Go 有多重赋值。

都有`if`、`switch`等条件语句。Go 添加了`select`。两者都支持循环。Java 有`while`、`do`和`for`语句。Go 只有`for`。

两者都有变量声明语句。Go 为局部变量添加了一个方便的声明和赋值组合。Go 提供了基于任何现有类型的通用类型声明。Java 只能声明类、接口或枚举类型。

Go 和 Java 都有异常能力。Java 可以抛出和捕获`Throwable`实例；Go 可以从恐慌中崛起和恢复。

在哲学上，Go 与 Java 有一些不同之处:

*   Go 往往遵循“少即是多”的哲学。

*   Java 诞生的最初动机是简化 C++的复杂性。Go 可以从那个角度来看，但是为了简化 C(以及 Java)。例如，在 Go 语言中通常只有一种方法(而 Java 通常有几种方法)来做一些事情。

*   请注意，Java 的大部分语法都是从 C++语法派生出来的，C++语法是 C 语法的超集，所以 Java 语法也是基于 C 语法的。在较小的程度上，Java 的很多语义都是基于 C++语义的。Go 更面向 C 功能及其支持库。

*   创建 Go 是为了适应 C 语言这样的利基市场。

*   与 C++语言相比，Go 与 C 有更多的共同之处(C++是 C 语言的一个大型超集，Java 就是从 C 语言中派生出来的)。它旨在成为一种类似 C 语言的“系统编程”语言，但具有改进的安全性和语义，以满足现代计算机系统的需求，特别是具有改进的多核处理器易用性。Java 就是这样，但它旨在支持更广泛的用例集。

*   Go 在源语法和格式(符号、操作符、标点和空格的使用)上类似于 C(因此也类似于 Java)。因为 Java 也是基于 C 的，所以 Go 和 Java 在这方面也很相似。

*   Go 的语法更简单。

*   例如，Go 允许分号(“；”的大部分使用)语句结束符在可以隐含时被省略(不存在)。注意，使用省略的语句终止符是惯用的，也是首选的。与 Java 相比，这可以使代码读/写得更干净、更容易。此外，Java 中圆括号(`(...)`)的许多用法在 Go 中都被取消了。在关联类型之外定义方法可以使代码更具可读性。

*   Go 与 Java 有不同的优化点/目标。

*   Java 更多的是一种应用(尤其是商业)语言。Go 更面向系统。这些优化点强烈地影响了 Go 语言的设计/本质。像所有的图灵完全语言一样，Java 和 Go 有重叠的适用性领域，在这些领域中任何一个都是合适的选择。

*   Go 通常比 Java 更具命令性和明确性。

*   Java，尤其是如果使用了 Java 8(以及更高版本)的特性，可以比 Go 更具声明性和抽象性。在某些方面，Go 更像 Java 的第一个(1.0)版本，而不是 Java 的当前定义。

*   在 Go 中，大多数行为都是显式编码的。

*   行为并不隐藏在 Java Streams 和 Lambdas 所支持的函数式编程特性中。这可能会使 Go 代码在风格上更加重复。错误被显式处理(比如在每次函数返回时),而不是像 Java 那样远程/系统地处理异常。

*   除了(在功能上受到限制)struct field *tags* 之外，Go 没有 Java 所具有的注释概念。同样，这是为了让 Go 代码更加透明和明显。注释和任何声明性/后处理(命令性)方法一样，倾向于隐藏或推迟行为。

*   Java 注释驱动方法的一个很好的例子是 *Spring* *MVC* 和 *JAX-RS* 如何在 web 应用服务器中定义 REST API 端点。通常，注释不是在编译时解释，而是在运行时由第三方框架解释。

*   另一个例子是数据库*实体* <sup>[8](#Fn8)</sup> 如何被典型地定义为*对象关系映射器* <sup>[9](#Fn9)</sup> (ORM)。在这种有限的情况下，Go 通过 struct 标签提供选项，这些标签通常用于通知这些工具。社区提供的*GORM*<sup>[10](#Fn10)</sup>ORM 就是一个例子。内置的 JSON 和 XML 处理器也使用标签。

*   Go 支持(源)*生成器*的概念。

*   生成器就是写 Go 代码的 Go 代码。生成器可以由 Go builder 有条件地运行。发电机有许多用例。例如，可以使用生成器来机械地创建集合类型(比如为列表<t>、堆栈<t>、队列<t>、映射<k t="">等的每个所需 T/K 生成一个类型)。)模仿 Java 泛型类型，但通过预处理器完成。Go 社区提供了这样的选择。</k></t></t></t>

*   Go 支持*指针*，Java 支持*引用*。

*   对计算机来说，指针和引用是相似的，但对人类来说，它们是不同的。引用是一个比指针更抽象的概念。指针是保存其他值的机器地址的变量。引用是保存另一个值的定位器(可能是地址或其他东西)的变量。

*   在 Java 中，引用在使用时总是自动解引用的(除了在赋值中)。指针可以是也可以不是。使用指针，可以获得一些数据的地址，并将其保存在指针变量中，还可以将指针转换为其他类型，比如整数。这对于引用是不可能的。

*   与 C(或 C++)不同，Java 和 Go 都限制指针/引用来处理特定类型的数据。没有什么比 c 语言的“void”指针更好的了，也没有什么比 c 语言允许的“指针算法”更好的了。因此，Go 和 Java 一样，比 c 语言更安全(不太可能因为寻址错误而失败)

<aside aria-label="Footnotes" class="FootnoteSection" epub:type="footnotes">Footnotes [1](#Fn1_source)

在当前 HTML 表单上使用另存为 PDF。

  [2](#Fn2_source)

有些人可能会说这些库是精简的，意思是 T1。

  [3](#Fn3_source)

[T2`www.apache.org/`](http://www.apache.org/)

  [4](#Fn4_source)

[T2`https://spring.io/`](https://spring.io/)

  [5](#Fn5_source)

Java 的好总结可以在 [`www.artima.com/objectsandjava/webuscript/ExpressionsStatements1.html`](http://www.artima.com/objectsandjava/webuscript/ExpressionsStatements1.html) 找到

  [6](#Fn6_source)

涉及问题见本文: [`www.oracle.com/technical-resources/articles/javase/supplementary.html`](http://www.oracle.com/technical-resources/articles/javase/supplementary.html)

  [7](#Fn7_source)

[T2`https://en.wikipedia.org/wiki/UTF-8`](https://en.wikipedia.org/wiki/UTF-8)

  [8](#Fn8_source)

[T2`https://en.wikipedia.org/wiki/Entity%E2%80%93relationship_model`](https://en.wikipedia.org/wiki/Entity%25E2%2580%2593relationship_model)

  [9](#Fn9_source)

[T2`https://en.wikipedia.org/wiki/Object%E2%80%93relational_mapping`](https://en.wikipedia.org/wiki/Object%25E2%2580%2593relational_mapping)

  [10](#Fn10_source)

[T2`https://gorm.io/index.html`](https://gorm.io/index.html)

 </aside>