# 五、Go 基本功能

在这一章中，我们将深入探究一些让 Go 成为现在这个样子的重要特征。当我们读完这一章时，你应该能够描述出 Go 语言与 Java 语言的相似之处和不同之处。

## 语言关键字

Go 保留字(不能用于任何其他目的)是

```go
break, case, chan, const, continue, default, defer, else, fallthrough, for, func, go, goto, if, import, interface, map, package, range, return, select, struct, switch, type, var

```

表 10 Go 保留字

Java 和 Go 都有关键字，有些是保留字。它们在表 [5-1](#Tab1) 中列出。如果它们具有相同/相似的目的，则在同一行中列出。某些关键字是保留字(只能按照语言中的定义使用，不能作为变量名)。

表 5-1

保留字和关键字比较

<colgroup><col class="tcol1 align-left"> <col class="tcol2 align-left"> <col class="tcol3 align-left"></colgroup> 
| 

爪哇

 | 

去

 | 

目的

 |
| --- | --- | --- |
|   | `_` | 可废弃的价值；Java 里没有 |
| `abstract` |   | Go 没有对等物 |
| `assert` |   | Go 没有直接的对等词；使用 panics 是类似的 |
| `boolean` | `bool` | 相同的 |
| `break` | `break` | 相同的 |
| `byte` | `byte` | Go 中未签名；用 Java 签名 |
| `case` | `case` | 一样；Go 有一些扩展 |
| `catch` |   | Go 在延迟函数中有一个内置的`recover()`函数，而不是`try/catch` |
|   | `chan` | Java 没有对等物 |
| `char` |   | Go 有一个`rune` (32 位)类型 |
| `class` |   | Go 没有 OOP 一个`struct`类型是最接近的近似值 |
| `const` | `const` | 在 Java 中未使用 |
| `continue` | `continue` | 相同的 |
| `default` | `default` | `switch`中相同；Go 中没有类似的能见度。在 Go 中，函数没有用处 |
|   | `defer` | 喜欢 Java `try/finally` |
| `do` |   | Go 没有直接的对等词 |
| `double` |   | Go 有一个`float64`型 |
| `enum` |   | Go 使用`int` (iota)常量代替(如 C) |
| `else` | `else` | 相同的 |
| `extends` |   | Go 没有传承 |
|   | `fallthrough` | 创建 Java `switch`失败行为 |
| `final` |   | Go 有一个`const`声明 |
| `finally` |   | Go 使用延迟函数代替`try/finally` |
| `float` |   | Go 有一个`float32`型 |
| `for` | `for` | 类似的 |
|   | `func` | Java 有 Lambdas |
|   | `go` | Java 有线程 |
| `goto` | `goto` | Java 中未使用；在 Go 中，目的与`break`相似，但可以在循环之外使用 |
| `if` | `if` | 相同的 |
| `implements` |   | Go 隐式地这样做 |
| `import` | `import` | 类似的 |
| `int` | `int` | Go 也有一个`int32`类型 |
| `instanceof``(type)x` | `x.(type)` | Go 有一个类型断言测试；断言的强制转换部分 |
| `interface` | `interface` | 类似角色 |
| `long` |   | Go 有一个`int64`类型 |
|   | `map` | Java 有一个 *HashMap* (以及其他)库类型 |
| `native` | 没有主体的函数定义 | Go 没有直接的对等词；但是 CGo 做了类似的事情 |
| `new` | `new` | 创建一个对象；Go 内置了一个类似的新功能；Go 没有与类型相关联的构造函数的概念 |
| `package` | `package` | 类似角色 |
| `private` |   | Go 使用小写名称(更多的包保护) |
| `protected` |   | 不需要；Go 没有传承 |
| `public` |   | Go 使用大写名称 |
|   | `range` | Java 有一个`for`语句 |
| `return` | `return` | 一样；Go 可以返回多个值 |
|   | `select` | Java 没有对等物 |
| `short` |   | Go 有一个`int16`类型 |
| `static` |   | Go 有全局(相对于类)变量 |
| `strictfp` |   | Go 没有对等物 |
|   | `struct` | Java `class`可以以类似的方式使用 |
| `super` |   | Go 没有对等物 |
| `switch` | `switch` | 类似；Go 有扩展名 |
| `synchronized` |   | Go 没有直接的对等词；库可以提供类似的行为 |
| `this` |   | Go 可以为此角色使用任何名称 |
| `throw` |   | Go 有一个内置的应急功能 |
| `throws` |   | 不需要；Go 没有要声明的例外 |
| `transient` |   | Go 中没有对等词 |
| `try` |   | Go 没有直接的等价物，但是支持类似于`try/catch`和`try/finally`的行为 |
|   | `type` | Java 没有对等物 |
| `var` | `var` | Java 使用类型名(块局部变量除外) |
| `void` |   | 省略的返回类型在 Go 中起到了同样的作用 |
| `volatile` |   | Go 没有对等物 |
| `while` |   | Go 有`for` |

## 运算符和标点符号

Java 和 Go 都有运算符和标点符号。许多语言有相同或相似的目的，但是每种语言都有一些独特的目的。由于 Go 支持有符号和无符号整数类型，一些操作符的工作方式略有不同(Java 不支持无符号整数)。还有，Go 不会自动将较小的数字转换成较大的数字(如`byte`->-`short`->->-`double`)；这种转换在 Go 中必须是显式的。表 [5-2](#Tab2) 总结了 Go 和 Java 操作符以及它们之间的比较。

表 5-2

Java 和 Go 运算符比较

<colgroup><col class="tcol1 align-left"> <col class="tcol2 align-left"> <col class="tcol3 align-left"></colgroup> 
| 

爪哇

 | 

去

 | 

目的

 |
| --- | --- | --- |
| `+` | `+` | 一样；(二元)加法和(一元)正数或字符串连接 |
| `-` | `-` | 一样；(二进制)减法和(一进制)负数 |
| `*` | `*` | 同(二进制)乘法；在 Go 中，还有指针声明和(一元)解引用 |
| `/` | `/` | 一样；分开 |
| `%` | `%` | 一样；以…为模 |
| `&` | `&` | 一样；逐位 and 在 Go 中，也(一元)用作地址 |
| `&#124;` | `&#124;` | 一样；按位或 |
| `^` | `^` | 一样；逐位异或；(一元)不在 Go 中的布尔值上 |
| `<<` | `<<` | 一样；按位左移 |
| `>>` | `>>` | 一样；按位右移 |
| `>>>` |   | 无符号逐位右移；在 Go 中，对无符号 int 类型使用>> |
|   | `&^` | 位清零(和不清零)；Java 里没有 |
| `=` | `=` | 分配；在 Go 中，一种语句，而不是一种运算符 |
| `+=` | `+=` | 加赋值；在 Go 中，一种语句，而不是一种运算符 |
| `-=` | `-=` | 减去赋值；在 Go 中，一种语句，而不是一种运算符 |
| `*=` | `*=` | 乘法赋值；在 Go 中，一种语句，而不是一种运算符 |
| `/=` | `/=` | 划分作业；在 Go 中，一种语句，而不是一种运算符 |
| `%=` | `%=` | 模数分配；在 Go 中，一种语句，而不是一种运算符 |
| `&=` | `&=` | 和赋值；在 Go 中，一种语句，而不是一种运算符 |
| `&#124;=` | `&#124;=` | 或者赋值；在 Go 中，一种语句，而不是一种运算符 |
| `^=` | `^=` | 异或赋值；在 Go 中，一种语句，而不是一种运算符 |
| `<<=` | `<<=` | 左移赋值；在 Go 中，一种语句，而不是一种运算符 |
| `>>=` | `>>=` | 右移位赋值；在 Go 中，一种语句，而不是一种运算符 |
| `>>>=` |   | 无符号右移位赋值；不在 Go 中 |
|   | `&^=` | 位清零分配；Java 里没有 |
| `&&` | `&&` | 一样；逻辑与；短路 |
| `&#124;&#124;` | `&#124;&#124;` | 一样；逻辑或；短路 |
| `++` | `++` | 自动递增；在 Go 中，只有后缀；在 Go 中，一种语句，而不是一种运算符 |
| `--` | `--` | 自动减量；在 Go 中，只有后缀；在 Go 中，一种语句，而不是一种运算符 |
| `==` | `==` | 一样；等于测试 |
| `!=` | `!=` | 一样；不等于测试 |
| `<` | `<` | 一样；小于测试 |
| `<=` | `<=` | 一样；小于或等于测试 |
| `>` | `>` | 一样；大于测试 |
| `>=` | `>=` | 一样；大于或等于测试 |
|   | `:=` | 简单声明；Java 里没有 |
| `...` | `...` | 类似；varargs 声明；在 Go 中，列出函数参数的展开 |
| `(` | `(` | 一样；打开参数列表或打开子表达式 |
| `)` | `)` | 一样；关闭参数列表或关闭子表达式 |
| `[` | `[` | 一样；开放索引 |
| `]` | `]` | 一样；收盘指数 |
| `{` | `{` | 一样；打开块或初始化列表 |
| `}` | `}` | 一样；关闭块或初始化列表 |
| `;` | `;` | 一样；在 Go 中，如果放在行尾，通常可以省略 |
| `:` | `:` | 一样；分离器 |
| `@` |   | 批注指示器；不在 Go 中 |
| `::` |   | 方法参考；不在 Go 中 |
| `.` | `.` | 一样；字段参考 |
| `,` | `,` | 一样；列表或参数分隔符；不是 Go 中的运算符 |
| `~` |   | 没有比特智慧；不在 Go 中 |
| `?:` |   | 三元选择；不在 Go 中 |
| `!` | `!` | 一样；逻辑非 |
| `->` |   | Lambda 表达式声明；不在 Go 中 |
|   | `<-` | 向通道发送或从通道接收(基于位置)；Java 里没有 |
| `instanceof``(type)value` | `x. (y)` | 测试类型；不在 Go；Go 有一个断言表达式，用于转换类型并返回一个布尔值`if`,这种转换是可能的；can if 测试布尔型 |
| `new` | `New``make` | 分配和构造一个对象；Go 有新功能，make 有。此外，可以声明 Go 结构并获取地址，这将导致相同的操作。在 Go 中，`new`不运行任何构造函数；那就是`make`的作用 |

Java 和 Go 都有关系运算符(`==`、`!=`、`<`、`<=`、`>`、`>=`)，但它们的工作方式并不总是相同。例如，要在 Java 中比较两个字符串`s1`和`s2`(或任何其他引用类型)是否相等，必须使用

```go
if(s1.equals(s2)) { ... }

```

在 Go 中，人们会使用

```go
if s1 == s2 { ... }

```

只有当类型可比较时，这种比较才有效。大多数内置类型都是。有关详细信息，请参见 Go 语言规范。切片、贴图和函数值除了与`nil`值比较之外是不可比的。指针、通道和接口值也可以与`nil`进行比较。

在比较引用的 Java 中，以下内容具有不同的含义(即，相同性测试):

```go
if(s1 == s2) { ... }

```

这是一个测试，看看`s1`和`s2`引用是否指向同一个对象(即*是彼此的别名*)。为了在 Go 中得到等价的测试，需要这样的测试来比较字符串的地址(不是值):

```go
if &s1 == &s2 { ... }

```

Java 不隐式支持引用类型的关系测试；类型本身必须提供一些方法来做到这一点。对于字符串，测试可以是

```go
if(s1.compareTo(s2) < 0) { ... }

```

在 Go 中，这将是

```go
if s1 < s2 { ... }

```

只有当类型通过某种方式排序时，这种比较才有效。在 Go 中，很多类型都是。有关详细信息，请参见 Go 语言规范。例如，字符串被排序为字节数组，较短的字符串意味着附加的零单元。

像在 Java 中一样，`&&`和`||`操作符是短路的，可能只计算左边的参数。

Go *take address* ( `&`，又名的*地址)一元运算符返回其操作数的地址，该地址必须是可寻址的(有存储位置；例如，常数，而许多表达式不是)。Java 没有与这个动作等价的动作。对于任何类型为`T`的值，`&`操作符返回类型为`*T`的值。我们不能接受一个`nil`值的地址。*

可寻址的值有

*   已声明的变量

*   指针解引用(`*p`)–返回 p

*   数组或切片索引表达式

*   结构字段选择器表达式

*   复合(数组、切片、映射、结构)文字

注意，对于任何导致恐慌的表达式`exp`，表达式`&exp`也将。

## Go 运算符优先级

Java 操作符优先级很复杂，这里不再赘述(更多细节参见 Java 语言规范)。Go 优先一般比较简单。一元运算符优先于二元运算符。

一元从高到低:

*   `Wrapping ( ... )`

*   `Prefix + - * &`

*   `Suffix [ ... ] ( ... )`

注意 Go 中的`++`和`--`是语句，不是运算符。

二进制从高到低:

*   `/ % << >> & &^`

*   `+ - | ^`

*   `== != < <= > >=`

*   `&&`

*   `||`

最佳做法是在有疑问时使用括号(...)来阐明优先顺序，尤其是对于一元运算符。

注意 Go 有 Java 没有的位清除(`&^`)运算符。

表达式`x &^ y`实际上是`x AND (NOT y)`。

注意 Go 没有 Java 的二进制 not ( `~`)运算符。请改用异或运算符。例如:

```go
func not32(x uint32) uint32 {
       return x ^ 0XFF_FF_FF_FF
}
func not64(x uint64) uint64 {
       return x ^ 0XFF_FF_FF_FF_FF_FF_FF_FF
}

or:
func not32(x uint32) uint32 {
       y := int32(-1)
       return x ^ uint32(y)
}
func not64(x uint64) uint64 {
       y := int64(-1)
       return x ^ uint64(y)
}

```

当运行时

```go
fmt.Printf("%X\n", not32(10))
fmt.Printf("%x\n", not64(10))

```

产生(注意大小写差异)

```go
FFFFFFF5
fffffffffffffff5

```

## Go 内置函数

Go 有几个内置函数，总结在表 [5-3](#Tab3) 中，用于访问常见行为。这些函数通常是通用的(或重载的),因为它们处理不同的数据类型。Java 通常有执行类似行为的特定于类型的方法。

注意，Go 内置函数名不是保留关键字；这些名称可能用于其他目的(可能隐藏了内置函数)。

表 5-3

Java 和 Go 常用函数比较

<colgroup><col class="tcol1 align-left"> <col class="tcol2 align-left"> <col class="tcol3 align-left"></colgroup> 
| 

爪哇

 | 

去

 | 

目的

 |
| --- | --- | --- |
| `.length, .length(), .size(), ...` | `len(...)` | 获取字符串、数组、切片、映射、通道的长度 |
| 因收藏类型而异 | `cap(...)` | 获取切片、地图、频道的容量；对于某些集合，`cap`和`len`是相同的 |
| `new`。。。或工厂方法 | `make(...)`或`new(...)`或`&<structureType>{}` | 创建(并为`make`初始化)一个集合或结构；返回一个指向分配内存的指针；对于`new`,在 Go 中没有构造函数调用 |
| `System.arraycopy` | `copy(...)` | 在相同/不同数组之间复制/移动数组元素 |
| 因收藏类型而异 | `delete(...)` | 从地图中删除元素；通常用作陈述 |
| Java 里没有；有些类型有这样做的方法 | `close(...)` | 关闭一个频道；通常用作陈述 |
| `(<type>)...` | `<type>(...)` | 将参数转换为指定的类型(即强制转换) |
| `throw <throwable>` | `panic(...)` | 引起恐慌；可以发送任何类型作为紧急值，但是最好发送一个`error`实例；一般用作陈述；避免使用`panic(nil)` |
| `try/catch` | `v := recover()` | 惊慌失措。通常用在延迟函数中 |
| 因收藏类型而异；经常`add()` | `append(...)` | 向切片追加值；如果需要，重新分配切片；应该将结果分配给输入切片 |
| Java 里没有 | `complex(...)` | 生成一个复数值 |
| Java 里没有 | `real(...)` | 获取复数值的实部 |
| Java 里没有 | `imag(...)` | 获取一个复数值的虚部 |