# 八、Go 语句

在这一章中，我们将更详细地描述 Go 的各种语言语句。当我们完成这一章时，你应该能够清楚地识别 Go 和 Java 语言语句及其功能之间的异同。

与 Java 非常相似，在 Go 中，计算是基于命令式模型的。计算按顺序执行，并保存在变量中。Go 几乎没有 Java 也支持的函数式编程计算风格。控制流只基于条件语句和循环语句，而不是像 Java 可以用它的流库支持的那样嵌入在函数调用中。关于在 Go 中尝试函数方法的一些讨论可以在 [`https://github.com/robpike/filter`](https://github.com/robpike/filter) 找到。

Go 有几个条件语句:

*   单向或双向条件句(也可用于构成多路条件句)-if/else

*   多路值条件开关

*   多路通道条件–选择

Go 有一个循环语句(for ),其中包含几个子窗体:

*   无限循环

*   带调整索引的循环

*   当条件为真时循环

*   在集合中循环

Go 可以用不同的方式退出/迭代循环:

*   回路条件测试失败

*   突然退出——中断还是返回

*   前进到下一个迭代–继续

像在 Java 中一样，所有的 Go 代码必须被分组到可重用的单元中，这些单元被称为*函数*。在 Go 中，最好的做法是保持函数简短(比如几十行，最多几行)，并根据需要生成更多的函数。Go 可以通过名字调用函数，也可以通过函数值间接调用函数。Java 只能通过名字调用方法。有些函数 Java 是通过语句做的，Go 是通过内置函数调用做的。

Go 可以为每个函数返回零个或多个结果。Java 只支持零或一。像在 Java 中一样，返回可以出现在函数中的任何地方。

## 打包和导入语句

像 Java 一样，每个 Go 源文件都需要一个 Package 语句作为第一个声明源代码所属包的语句，比如

```go
package main

```

一个包中可以有任意数量的源文件。Go 源文件名不需要与包名匹配(如果一个包有多个源文件，通常不需要)，但是为了更好地组织代码，建议它们匹配，尤其是对于包含主入口点的目录。例如，建议您使用一个`main.go`文件来保存包含`main()`函数的`main`包源代码。注意`main`包约定中的`main`函数是必需的，这样 Go builder 就可以识别出必须构建一个可执行文件。

如果源文件使用另一个包中的任何公共声明，则必须导入该包，例如

```go
import "math"
import "net/http"

```

或者这样分组:

```go
import (
  "math"
  "net/http"
)

```

源文件中可以有多组导入。所有 import 语句必须在 package 语句之后，任何其他语句之前。导入可以按任何顺序进行，但通常按导入路径中的姓氏排序，尤其是在同一个导入组中。如果源文件中没有引用包中的公共项，则不能导入该包(编译器将报告错误)。

导入是在文件级而不是包级完成的，因此必须像在 Java 中一样，在使用导入的每个源文件中重复进行。同一个包中的不同源文件可以并且经常有不同的导入列表。

导入的包中对公共名称的所有引用都必须以包名为前缀，如下所示:

```go
r := new(http.Request)

```

默认情况下，任何导入的包路径中的姓都用作导入的前缀名。有时，您可能希望对一个包使用不同的(比如说更短的)名称。您可以在导入过程中为包指定一个别名，如下所示:

```go
import net "net/http"

```

Go 包可以有几个`init()`功能。有时，即使不使用包中的符号，也需要运行这些函数。为此，在导入中添加空白的别名(下划线)，如下所示:

```go
import _ "net/http"

```

一个包的`init()`函数只运行一次，不管有多少源文件导入这个包。

## 赋值语句

也许在 Go 中最基本的动作就是给一个变量赋值。在 Go 中，像在 Java 中一样，这是通过*赋值*语句显式完成的。也可以通过向函数传递参数或从函数返回值来实现。赋值可以是常量、其他变量或涉及这些项目的表达式。

最基本的任务是

```go
<variable> = <expression>

```

虽然是声明而不是赋值，但也有一种方便的方法来声明和赋值，类似于赋值语句:

```go
 <variable> := <expression>

```

还有这种形式的扩充(也称为复合)赋值:

```go
<variable> <binaryOperation>= <expression>

```

它们被解释为

```go
<variable> = <variable> <binaryOperation> <expression>

```

像在 Java 中一样，并不是所有受支持的二元运算符都可以与赋值运算符结合使用。例如，逻辑运算符`(&&`和`||`不能使用，因为它们具有短路行为。

请注意以下声明:

`<variable>++`相当于`<variable> += 1`

`<variable>--`相当于`<variable> -= 1`

Go 允许以下形式的并行(即元组)赋值:

```go
<variable1>,<variable2>,...,<variableN> = <expression1>,<expression2>,...,<expressionN>

```

其中 N 在每一侧必须相同。任何(但通常不是全部)`<variableX>`都可以用下划线(“_”)替换，以忽略表达式位置，这通常是在函数调用结果中进行的。

所有右侧的值必须与左侧的相应变量兼容(能够被赋值)，没有任何隐含的转换(除了一些数字文字值)。通常，这意味着相应位置的左侧变量和右侧值必须是同一类型。

如果左侧至少有一个变量是新声明的，则允许使用声明形式:

```go
<variable1>,<variable2>,...,<variableN> := <expression1>,<expression2>,...,<expressionN>

```

在前面所有的例子中，`<variableX>`是定义一个可赋值目标(又名左值)的任何表达式。通常，这些是简单的标识符(变量名)，但也可以是索引数组、切片、映射或指针变量解引用。

## 声明变量

Java 允许一次声明一个变量并分组声明。Go 也是如此。在 Java 和 Go 中，任何初始值都是可选的。注:在 Java 中，可以创建没有初始值的块/方法局部变量。同样的情况在 Go 中是不可能的；如果没有指定，所有声明的值都有一个初始值(称为*零*)。

Java 的声明:

```go
{<vis>} {<mod>}... <type> <id> {= <value>} {, <id> {= <value>}}...;

```

类型是任何内置或声明的类型(类、接口、枚举等)。).这些值必须可转换为类型。如果省略，则使用默认值(块/方法局部变量除外)。该值可以是一个表达式。

仅在字段声明中允许使用`<vis>`修饰符。它是`public`、`private`、`protected`中的一个，或者省略(意味着默认或包受保护)。Java 的`<mod>`修饰符，像`abstract`和`final`，通常只允许在字段声明中使用，没有 Go 等价物。

Go 相当于一个声明语句:

```go
var <id> {, <id>}... <type>

```

或者

```go
var <id> {, <id>}... <type> = <value> {, <value>}...

```

或者

```go
var <id> {, <id>}... = <value> {, <value>}...

```

该类型是任何内置或声明的类型。每个值必须属于同一类型。如果该值是文本，则它必须可转换为类型。如果省略，则使用零值。只有当所有值都被省略时，类型才是必需的；如果一个值存在，它的类型将用于推断任何缺少的类型。每个位置的推断类型可能不同。任何值都可以是表达式。id 和值的数量必须相同。

如前所述，Go 没有可见性修改器。如果 id 以大写字母开头，则它是 public 否则，它是包私有的(只能被同一个包中的代码看到)。

Go 允许更简洁的声明形式:

```go
var ({<xxx> {, <xxx>...})

```

其中 xxx 是没有“var”前缀的声明。结束语“)”通常单独在一行中。这是声明变量的常规方式。

例如:

```go
var (
      p = 1
      q = "hello"
      l int
      f float64 = 0
)

```

在顶级声明中，任何关于 var 的注释都由组中的所有成员共享。

Go 有另一种声明形式用于块局部(非字段)声明:

```go
<id> {, <id>}... := <value> {, <value>}...

```

其中 id 和值的计数必须匹配。此外，在同一个块中至少不能声明一个 id。id 的类型可以不同，由值来表示。

元组赋值(或声明)有许多用途，但一些常见的用途是

*   不使用临时变量交换值

例如:

*   拆分`range`操作的结果

```go
var x, y = 1, 2
x, y = y, x // after x==2, y == 1

```

例如:

*   拆分函数或运算符返回的结果

```go
for index, next := range collection { ... }
- or -
for _, next := range collection { ... }

```

例如:

```go
file, err := os.Open(...)
- or -
if v, ok := map[key]; ok { ... }

```

## 声明命名常数

Java 允许声明类似常量的 <sup>[1](#Fn1)</sup> ( `static final`)值。Go 有真常数。Go 支持一次定义一个常量并分组定义。在 Java 和 Go 中，初始值都是必需的。

Java 的声明(在某种类型内部):

```go
{<vis>} static final <type> <id> {= <value>} {, <id> {= <value>}}...;

```

这些值必须可转换为类型。该值必须是常量表达式。

仅在字段声明中允许使用`<vis>`。它是`public`、`private`、`protected`中的一个，或者省略(意味着默认或包受保护)。

Go 相当于一个声明语句:

```go
const <id> {, <id>}... <type> = <value> {, <value>}...

```

或者

```go
const <id> {, <id>}... = <value> {, <value>}...

```

该类型是任何具有文字初始值设定项的内置或声明的类型。该值必须属于同一类型。如果该值是文本，则它必须可转换为类型。该值必须是可以在编译时计算的表达式(即，所有引用的标识符都指向没有循环引用的其他常数)。id 和值的数量必须相同。

Go 没有可见性修改器。如果 id 以大写字母开头，则它是 public 否则，它是包私有的(只能被同一个包中的代码看到)。

Go 允许更简洁的声明形式:

```go
const ({<xxx> {, <xxx>...})

```

其中 xxx 是没有“const”前缀的声明。结束语“)”通常单独在一行中。这是声明常数的常规方式。

例如:

```go
const (
      p = 1
      q = "hello"
      f float64 = 0
)

```

## If/Else 语句

If/Else 是最基本的条件测试机制。它允许代码序列中的交替流。

Java 的 if 语句:

```go
if(<cond>) <block>

```

或者

```go
if(<cond>) <block> else <block>

```

Java 允许除了块之外的任意可执行语句作为 if/else 目标。

Go 的 if 语句:

```go
if {<simpleStmt>;} <cond> <block>

```

或者

```go
if {<simpleStmt>;} <cond> <block>
else (<ifStmt>|<block>)

```

If/else 目标是语句块(这也是 Java 中的最佳实践)。Else 语句还允许另一个 if 语句作为目标；这允许多条件测试。在 Go 中，多条件测试最好通过使用 Switch 语句来完成。

可选的简单语句是

*   空(省略–无分号)语句

*   表达式语句

*   发送(通道

*   Inc/dec 报表

*   分配

*   短变量声明(最常见的选项)

if 语句创建了一个隐含块，所以任何声明都隐藏了这样的名字，使其不包含作用域。例如:

```go
var x, y = 0, 0
if t := x; t < 0 {  // t in new scope
      var x = 1 // a new x variable; hides x above
      y = t + x
} else {
      y = -1
}

```

注意`else`子句，如果存在，必须在与`if`块结束相同的行开始。

在惯用的 Go 中，`else`的使用被最小化。因此，从条件(比如 If)块返回是很常见的。当这样做时，使用`else`子句是非常规的(这是多余的)。例如:

```go
if t < 0 {
      return true
} else {
      return false
}

```

更常规的写法是

```go
if t < 0 {
      return true
}
return false

```

它也可以更简洁地表达为

```go
return t < 0

```

通过这种方式，惯用的 Go 代码倾向于在包含函数的左边对齐，而不会嵌套太深。如果你的代码嵌套超过(比方说)两层，考虑使用`return`、`break`或`continue`语句或者通过提取深度嵌套的代码作为一个新函数来重写它以减少层数。

Go 有很强的源代码风格规则。一个是如何测试布尔值。考虑(常见的)例子:

```go
if v, ok := aMap[someKey]; !ok {
      return
}

```

和...相对

```go
if v, ok := aMap[someKey];  ok == false {
      return
}

```

第一种形式(直接使用布尔值)是惯用的，通常用在第二种形式(比较布尔值)上。

Java 有一个*三元*表达式(？:)允许(通常很方便)条件测试。例如:

```go
int x = input < 0 ? -input : input;  // a simple abs(input)

```

这是一个简短的形式

```go
if(input < 0) x = -input; else x = input;

```

而是作为一种表达(相对于陈述)。

Go 没有这个表达式的对等词。人们必须这样做:

```go
var x int
if input < 0 {
      x = -input
} else {
      x = input
}

```

或者

```go
var x int = input // (or x := input)
if input < 0 {
      x = -input
}

```

或者，对于简单的物体(比如变量或常数)，更简洁地说:

```go
var x int; if input < 0 { x = -input } else { x = input }

```

或者

```go
x := input; if input < 0 { x = -input }

```

注意，即使如前面所示输入，大多数 Go 源代码格式化程序也会在分号处拆分这些行。

## 交换语句

和 Java 一样，Go 也有 Switch 语句。总的来说，Go 的 Switch 语句更加灵活。Java Switch 语句遵循以下一般形式:

```go
switch (<expr>) {
  case <value1>:
    :
  case <value2>:
    <statements>
    break;
  :
  default:
    <statements>
}

```

每组陈述可以由一个或多个案例介绍者进行。将 expr 值与每个 case 值(必须是唯一的)进行匹配(测试相等性)，并执行匹配后的任何代码。如果代码中没有提供，流程继续执行以下情况，直到找到`break`。expr 可以是任何整数类型、字符串类型或任何枚举类型。如果没有匹配，并且`default`介绍器存在，则运行该代码。

上例的 Go 对应物是

```go
switch <expr> {
  case <value> {, <value>}...:
    <statements>
  default:
    <statements>
}

```

switch 语句和每个 case 都创建了一个隐含的块，所以任何声明都隐藏了这样的名字，使其不包含作用域。

在 Go 案例中，每个案例使用多个匹配值，而不是多个案例介绍器。同样，在 Go 情况下，每组语句的末尾都有一个隐式的`break`。像在 Java 中一样，这些值必须是不同的。同样，在 Go 中，每个 case 都是它自己的块，就好像它被输入为(这在 Java 中是需要的)

```go
case <value>: {
    <statements>
}

```

这意味着变量可以在那组语句中声明为局部变量。

要获得类似 Java 的无中断失败，请用 fall through 语句结束这组语句，如下所示:

```go
switch <expr> {
  case <value1>:
    <statements>
    fallthrough
  case <value2>
    <statements>
  default:
    <statements>
}

```

Java 支持如下级联 if 语句:

```go
if(<expr1>) {
  :
} else if(<expr2>) {
  :
} ... else if(<exprN>) {
:
} else {
  :
}

```

Go 也支持这种方法，但是惯用的方法是使用不同形式的开关:

```go
switch {
  case <expr1>:
    <statements>
  case <expr2>
    <statements>
  :
  case <exprN>
    <statements>
  default:
    <statements>
}

```

表达式可以是任意的，除非它们必须是布尔类型。案例(默认案例除外)按照输入的顺序进行测试。

所以，这个 switch 语句:

```go
var c, ditto rune = 'c', '\0'
switch c {
case 'a', 'b', 'c':
      ditto = c
default:
      ditto = 'x'
}

```

和这个 switch 语句是等效的:

```go
var c, ditto rune = 'c', '\0'
switch {
case c == 'a', c == 'b', c == 'c':
      ditto = c
default:
      ditto = 'x'
}

```

Java 最近增加了 Switch 语句的表达式(有结果值)形式(一种增强的三元表达式)。开关可以是任何表达式中的术语。Go 没有这样的。这些开关表达式添加了不失败案例样式 Go has。此外，案件创建自己的块像 Go。与这个新的`switch`相关联的是返回开关值的新的`yield`语句。

## While 语句

While 是一种基本的循环机制。它允许在代码序列中有条件地(预先测试)重复流。

Java 的 while 语句:

```go
while (<cond>) <block>

```

Java 允许任意可执行语句作为 while 目标。

Go 相当于 while 语句:

```go
for <cond> <block>

```

该语句的目标是一个块(这是 Java 中的最佳实践)。

例如:

```go
var x, y = 10, 0
for x > 0 {
  y++
  x--
}

```

## Do-While 语句

Do-While 是一种基本的循环机制。它允许在代码序列中有条件地(事后测试)重复流动。

Java 的 do-while 语句:

```go
do <block> while (<cond>);

```

Go 没有与 Do-While 语句直接等效的语句。可以通过在块中包含测试的 for 语句生成一个，如下所示:

```go
var x, y = 10, 0
for {
      y++
      x--
      if x < 0 {
            break
      }
}

```

## For with Index 语句

For 是主要的索引循环机制。它允许一个索引跨越一个范围，并由代码序列中的重复流来操作。Go 的 For 语句提供了与 Java 的 For 语句类似的功能。

Java 的 for 语句:

```go
for({<init>};{<cond>};{<inc>}) <block>

```

Java 允许任意的可执行语句作为目标。

Go 相当于 for 语句:

```go
for {<init>};{<cond>};{<inc>} <block>

```

与 Java 不同，Go 不支持在`<init>`和`<inc>`子句中使用逗号("，")分隔的表达式。

该语句的目标是一个块(这是 Java 中的最佳实践)。`<cond>`子句是可选的，如果省略则为`true`。可选的`<init>`和`<inc>`组可以是

*   空(省略–无分号)语句

*   表达式语句

*   发送(通道

*   Inc/dec 报表

*   分配

*   短变量声明

该语句创建了一个隐含块，因此任何声明都隐藏了这样的名称，使其不包含范围。

例如:

```go
var x, y = 10, 0
for i := 0; i < 10; i++ {
      y++
}

```

## 对于超过一个集合声明

For 是迭代(可能为空)集合(或其他值流)的主要循环机制。它允许通过代码序列中的重复流程一次处理一个集合的元素。处理顺序由集合决定。

Java 的 for 语句:

```go
for(<varDecl>: <iterable>) <block>

```

或者(更直白地)一个人能做的

```go
Collection<SomeType> c = <some collection>;
Iterator<SomeType> it = c.iterator();
for(; it.hasNext();) {  // could use while here instead
  <varDecl> = it.next();
  :
}

```

或者(更直白地)一个人可以做的(在可索引集合上)

```go
Collection<SomeType> c = <some collection>;
for(int i = 0, count = c.size(); i < count; i++) {
  <varDecl> = c.get(i);
  :
}

```

Go 相当于集合上的 for 语句:

```go
for <indexVar>,<valueVar> := range <collection> <block>

```

该语句的目标是一个块(这是 Java 中的最佳实践)。可选的(至少需要一个)`<indexVar>`和`<valueVar>`接收下一项的索引(或键)和下一项的值。`<collection>`必须是某种集合或流类型，比如数组、切片、映射或通道。

Go 要求在块体中使用所有声明的变量(在`:=`的左边)。为了避免这一要求，如果没有引用，可以用下划线(“_”)来替换。

例如:

```go
for _, v := range []string{"x", "y", "z"} {
      fmt.Println(v)
}

```

或者

```go
aMap := make(map[string]string)
:
for k, v := range aMap {
      fmt.Printf("%s = %s", k, v)
}

```

Java 的等价物可能是

```go
Map<String,String> m = <some map>;
for(Iterator<String> it = m.keySet().iterator(); it.hasNext();) {
  var k = it.next();
  var v = m.get(k);
  System.out.printf("%s = %s", k, v);
}

```

`range`为地图呈现键的顺序是不确定的，并且对于每个地图实例可以是不同的。这是故意的。要按某种顺序处理键，必须首先对它们进行显式排序，比如说排序。例如:

```go
aMap := make(map[string]string)
:
keys := make([]string, 0, len(aMap)) // note created empty
for k, _ := range aMap {
      keys = append(keys, k)
}
sort.Strings(keys)
for _, k := range keys {
      fmt.Printf("%s = %s", k, aMap[k])
}

```

注意 Java 的`TreeMap`类型使这变得简单多了。

钥匙片也可以做成这样:

```go
keys := make([]string, len(aMap))  // note created full size
index := 0
for k, _ := range aMap {
      keys[index] = k  // "keys[index++] = k" not supported
      index++
}

```

前面的方法可能更节省时间。

## 永远的声明

For 是主要的无限循环机制。它允许在代码序列中无限重复地流动。

Java 的 for 语句:

```go
for(;;) <block>  // while (true) also works

```

Go 相当于 for 语句:

```go
for <block>

```

目标是一个块(这是 Java 中的最佳实践)。

例如:

```go
var x, y = 10, 0
for {
      y++
      x--
      if x < 0 {
            break
      }
}

```

## 中断和继续语句

像 Java 一样，Go 也有基本上以相同方式工作的`break`和`continue`语句。Break 退出循环，而 continue 移动到循环的下一次迭代。通常，这些语句位于一些条件语句的主体中，如 if 或 f or。语法是

```go
break {<label>}
continue {<label>}

```

如果标签存在，它一定是附加在某个包含循环上的标签。这允许从多层嵌套循环中退出。如果省略，则假定最嵌套的循环。任何循环都可以标记如下(但是该标记必须由一些 break 或 continue 引用):

```go
{<label> :}... <forStatement>

```

注意，Java 在 Go 中自动使用`switch`语句中的`break`语句来退出案例，因此在 Go 的`switch`(或`select`)语句中不需要`break`来避免失败。可以使用`break`或`continue`(比如说`if`主体)在结束前退出箱子。

## Goto 语句

Go 支持 Go-To(无条件跳转)语句；Java 没有(虽然是保留字)。它允许在同一块内跳转(但不允许跳出块或进入嵌套块)。不能使用 Go-To 跳过声明。同一块中任何带标签的语句都可以是目标。格式是

```go
goto <label>

```

可以用 Go-To 代替更结构化的表单。例如，这种概念形式的循环:

```go
for cur:=0; cur < 10; cur++ {
      : body of loop
}

```

可以这样创建:

```go
cur := 0
L1: if cur >= 10 {
      goto L2
}
: body of loop
cur++
goto L1
L2:

```

注意在作者看来，绝对不应该使用`goto`语句；`if`、`switch`、`for`提供足够的本地控制流量。您的代码应该遵循结构化编程的原则 <sup>[2](#Fn2)</sup> 及其 Goto <sup>[3](#Fn3)</sup> -less 方法。

## 返回语句

在 Java 中，每个方法都用一个`return`语句退出(可能在`void`方法上隐式运行，因为从`void`函数的末尾隐式返回)。return 语句提供要返回的值，如下所示:

```go
return {<value>}  // <value> present only on non-void methods

```

在 Go 中，Return 几乎是相同的，只是可以像这样返回多个值:

```go
return {<value>{,<value>...}}  // <value>... only on non-void methods

```

返回值的数量必须与函数原型上声明的返回值的数量相匹配。如果返回值是命名的，return 语句可以省略它们。

例如:

```go
func threeInts() (int, int, int) {
      :
      return 1, 2, 3  // required explicit return values
}

or:
func threeInts() (x, y, z int) {
      :
      return 1, 2, 3  // explicit return values (ignore names)
}

or:
func threeInts() (x, y, z int) {
      x, y, z = 1, 2, 3   // set return values before returning
      :
      return // implicit return values
}

```

这最后一种形式是这位作者普遍推荐的。其他人可能会采取不同的立场。

## 延期声明

Java 有两种流行的资源清理机制:

1.  Try/Finally(或 Try/Catch/Finally)

2.  尝试使用资源

try/最后，一般来说，看起来是这样的:

```go
try <block>
finally <block>

```

无论`try`子句如何结束，都执行`finally`子句(通常，通过`return`或通过某些异常)。

对资源的尝试通常如下所示:

```go
try (<declaration> = <new Resource>{;<declaration> = <new Resource>}...) {
    // use the resource(s)
}

```

当`try`结束时(通常，通过返回或通过一些异常)，在`try`子句中分配的任何资源被自动释放(在编译器编写的`finally`子句中)。

Go 有一个类似 try/finally 的特性，但是没有类似 try with resources 的特性。Go 使用 Defer 语句，其行为很像一个`finally`子句。这个语句看起来像

```go
defer <function call>

```

每次执行 defer 语句时(即使是在循环中)，对所提供的函数的调用都会放在调用堆栈中。当包含 defer 语句的函数退出时，延迟的函数调用以相反的顺序执行。可以有许多延迟函数。即使包含函数以`return`或死机(相当于抛出异常)结束，也会发生这种情况。

典型的方法如下例所示:

```go
func someFunction() {
      // acquire some resource
      defer func() {
            // release the resource
      }()  // note the function is called
      : use the resource
}

```

在这个模式中，在获得任何资源之后，立即注册一个释放资源的延迟函数。延迟函数继续运行，最终返回或死机，导致延迟函数被调用。

请注意，延迟函数可以访问延迟函数(它是一个闭包)的局部变量(必须在编码`defer`之前声明),并且在延迟函数返回到其调用者之前被调用，这允许它更改延迟函数的返回值。这很有用，尤其是在紧急情况下(比如被零除)或其他错误恢复时。例如:

```go
func someFunction() (result int, err error) {
      defer func() {
            if result == 0 {   // default value
                  result = -1
                  err = errors.New("bad value")
            }
      }()
      :
      result = 1
      :
      return
}

```

## Go 语句

Go 语句启动一个 goroutine。goroutine 只是一个普通的 Go 函数，通常不返回任何值(如果返回，则被丢弃)。创建一个 goroutine，并使用 Go 语句启动，如下所示:

```go
go <func>({arg, {arg}....})

```

Go 语句立即返回，函数与调用者异步(可能并行)运行。使用调用方提供的不同 goroutine 中的任何参数调用该函数。

注意，Go 中的所有代码都在某个 goroutine 中运行，包括`main()`函数。

通常，使用函数文字，而不是预先声明的函数，例如

```go
go func(x int) {
      :
}(1)  // note the function is called

```

注作者建议用后缀“Go”(或类似的词)来命名期望与 Go 一起运行的函数，以使这个用例清晰。

## 选择指令

Go 有一个 Select 语句，在 Java 中没有对应的语句。Select 语句用于处理通过通道接收的项或将项发送到通道。使用“选择”之前，请确保您了解频道。Select 语句看起来很像 Switch 语句:

```go
select {
  case <receiver>{, <receiver>}... = <- <channel>:
    <statements>
  case <identifier>, <var> := <- <channel>:
    <statements>
  case <channel> <- <expression>:
    <statements>
  default:
    <statements>
}

```

`<receiver>`是一个表达式(通常只是一个标识符),它指定一个变量来接收通道的值。Select 语句和每个 case 都创建了一个隐含的块，所以任何声明都隐藏了这样的名字，使其不包含作用域。

前两种情况在可以从渠道接收项目时触发。第三种情况是在可能向通道发送项目时触发(如果接收通道有空间)。所有病例都经过评估/测试。如果触发了任何案例，将随机选择并执行其中一个案例，并完成任何相关的赋值和/或语句。

第二种情况有一个`<var>`(通常命名为“ok”)，用于指示源通道是否关闭。通道关闭时将是`false`。

如果没有触发其他案例，则触发`default`案例。经常被省略。没有 default 子句的 Select 语句可以阻止等待接收或发送项目。

Select 语句经常在无限循环中执行，如下所示:

```go
for {
  select {
   :
  }
}

```

这允许从通道接收项目，并且只要它们被发送就进行处理(即，通道是打开的)。

可以接收两个不同通道的值的示例:

```go
var cchan chan int
var ichan chan int
var schan chan string
var scount, icount int
select {
case <- schan:                        // receive
      scount++                        // count receive
case <- ichan:                        // receive
      icount++                        // count receive
case cchan <- scount + icount:        // send current total
default:
      fmt.Println("no match")
}

```

<aside aria-label="Footnotes" class="FootnoteSection" epub:type="footnotes">Footnotes [1](#Fn1_source)

这些不是真正的常量(只存在于编译时)，而是不可变的值。

  [2](#Fn2_source)

[T2`https://en.wikipedia.org/wiki/Structured_programming`；](https://en.wikipedia.org/wiki/Structured_programming) [https:// en。维基百科。org/wiki/Structured _ program _ theory](https://en.wikipedia.org/wiki/Structured_program_theorem)

  [3](#Fn3_source)

[https:// en。维基百科。org/ wiki/ Goto](https://en.wikipedia.org/wiki/Goto)

 </aside>