# 十二、关键包的比较

本章总结了几个关键的(常用的和广泛使用的)Java 包。如果可行，任何等效的 Go 包或函数都会被注明。由于 Go 和 Java 在它们的可用库中不是一一对应的，如果在标准的 Go 库中存在 Go 等价类，那么让每个 Java API(方法)与 Go 等价类完全匹配是不切实际的。

## Java 语言包

*Java 标准版* (JSE)有很多类型和方法的捆绑包。可用的方法(也称为 API)数以千计。Go 也有一套标准包，其类型和功能多达数百种。在这些包和类型的集合之间，库行为有很大的重叠，但组织结构(行为所在的包、类型或功能)没有重叠。

仅仅列出(对比描述)所有的 JSE 包和它们包含的方法类型就需要上百页。本书不会试图这样做。相反，它将列出 JSE 包和类型的精选子集。对于其中的一些，它将把关键 Java 类型可用的方法与任何 Go 等价物进行比较。

JRE 在`java.lang`包中有一些关键类型。下面的列表描述了任何 Go 环境的等价物。

接口摘要

*   可追加–可以追加到该类型的实例中；由 Go 片隐式支持。

*   auto closeable–可以通过使用资源进行尝试来关闭；没有直接的 Go 对应物。

*   char sequence–字符序列(如字符串或 StringBuilder)；没有直接的 Go 对应物。

*   可克隆–实现`Object.clone()`；没有直接的 Go 对应物。

*   可比的<t>–可以支持一个`compareTo()`方法；没有直接的 Go 对应物；许多类型是隐式可比的。</t>

*   iterable<t>–可以迭代；一些 Go 类型:数组、切片、映射、通道。</t>

*   可读–可以将字符读入缓冲区；`io.Reader`为 UTF-8。

*   runnable——可以作为线程体运行；在 Go 中，任何函数都可以作为 goroutine 运行。

课程总结

*   布尔–布尔包装器；Go 系列不需要。

*   字节-字节包装器；Go 系列不需要。

*   字符-字符包装器；Go 系列不需要。

*   类<t>–类的运行时视图；用于反射的特征；Go 有反射包。</t>

*   class loader–在运行时加载/管理类；在 Go 中不需要，没有运行时类。

*   双层包装纸；Go 系列不需要。

*   枚举<e extends="" enum="">>–所有枚举类型的基本类型；在 Go 中不需要(`int`是大多数 Go 枚举的基础)。</e>

*   浮动——浮动包装；Go 系列不需要。

*   integer–Int 包装器；Go 系列不需要。

*   长长的包装纸；Go 系列不需要。

*   数学——一个有一套数学工具的类；Go 也有类似的`math`包。

*   模块——模块的运行时视图；没有直接的 Go 对应物。

*   number——数值包装类型的超类；没有直接的 Go 对应物。

*   object–所有对象类型的超类；没有直接的 Go 对应物；`interface{}`最接近。

*   包–包的运行时视图；没有直接的 Go 对应物。

*   进程——外部程序的运行时视图；一个 Go `exec`包有类似的。

*   ProcessBuilder 运行外部程序的助手；一个 Go `exec`包有类似的。

*   记录(新)-类似类的结构；Go 有`struct`型。

*   运行时——管理运行程序的实用程序；Go 有一个`runtime`包。

*   runtime permission——控制对类中函数的访问；Go 没有对等物。

*   security manager——控制对类中某个函数的访问的方法；Go 没有对等物。

*   短——短包装纸；Go 系列不需要。

*   StackTraceElement–描述一个调用堆栈元素；Go 也有类似的结构类型。

*   stack walker–抓取堆栈；没有直接的 Go 对应物；一个可以写。

*   strict Math——像一门数学课，有更多关于算法如何工作的规则；没有直接的 Go 对应物。

*   字符串–字符串类型；Go 有`string`类型和`strings`、`strconv`和`fmt`包。

*   StringBuffer，StringBuilder–一个可变的字符串类型；Go 有一个`strings.Builder`型。

*   系统-管理运行程序的实用程序；Go 有`runtime`、`time`和`io`包。

*   线程——操作系统线程；没有直接的 Go 对应物；go 有 goroutines。

*   thread group–相关线程的集合；没有直接的 Go 对应物。

*   thread local<t>–具有线程相关值的变量；没有直接的 Go 对应物；可以制造。</t>

*   throwable——可以被抛出的类型；Go 有恐慌。

*   void–不等同于 Go；Go 函数可以不返回任何值(相对于 void)。

*   数学。big integer–不定精度整数；Go 属于 math.Int 类型。

*   数学。BigDecimal–不确定精度的十进制浮点值；Go 有数学。Float(不过是二进制，不是十进制)。

*系统*类:

静态字段摘要

*   PrintStream err – STDERR – Go `os.Stderr`

*   输入流输入 – STDIN – 转到 `os.Stdin`

*   打印流出 – 标准输出 – 转到 `os.Stdout`

方法总结。如果没有提到，则没有直接的 Go 等价物。

*   array Copy(…)–复制一个数组；Go 运算符:数组[:]和`copy(..)`函数

*   clearProperty(字符串键)–空系统属性

*   console()–访问操作系统控制台

*   currentTimeMillis()–获取当前纪元时间；转到`time.Now()`

*   exit(…)–查看运行时类；转到`os.Exit(...)`

*   GC()–查看运行时类；转到`runtime.GC()`

*   getenv()–获取所有环境值；转到`os.Environ()`

*   getenv(…)–获取单个环境值

*   Get logger(…)–获取一个命名的记录器；转到`log`包

*   Get properties()–获取所有属性

*   getProperty(…)–获取一个命名的属性

*   getSecurityManager()–获取一个 JVM 安全管理器

*   identity hashcode(Object x)-获取对象的标识；在 Go 中，使用`&` x 运算符

*   line separator()–获取一个操作系统行分隔符(例如，NL，CR+NL)

*   load(…)，loadLibrary(…)–参见运行时类

*   nano time()–获取以纳秒为单位的运行时间；转到`time`包

*   run finalization()–查看运行时类

*   set err(…)-更改 STDERR

*   setIn（...） – 更改 STDIN

*   setOut(…)–更改标准输出

*   setProperties(属性属性)–设置许多系统属性

*   setProperty(字符串键，字符串值)-设置系统属性

*   setSecurityManager(…)–设置 JVM 安全管理器

属性有社区实现。参见 github.com/magiconair/properties 使用类似 Java 属性的文件格式的例子。

*运行时*类:

*   addShutdownHook(…)–在 JVM 的出口运行一个线程；没有直接的 Go 对应物；Go 可以陷 OS 信号；Go 可以捕捉恐慌。

*   available processors()–获取 CPU(核心)计数；去`runtime.NumCPU()`。

*   exec(...)–启动外部流程的方法系列；去`exec.Cmd`。

*   Exit(…)–用 cleanup 退出 JVM 没有 Go 对应物；去罐大约。通过使用`os.Exit()`。

*   free memory()–获取 JVM 可用的空闲内存；去`runtime.MemStats`。

*   GC()–运行垃圾收集；去`runtime.GC ()`。

*   Get runtime()–获取该类的 singleton 没有 Go 等价物。

*   halt(…)–不清理就退出 JVM 去`os.Exit(...)`。

*   load(…)，loadLibrary(…)–加载外部代码库；没有 Go 等价物。

*   maxMemory()–获取 JVM 可用的最大内存；去`runtime.MemStats`。

*   removeShutdownHook(…)–移除出口挂钩；没有 Go 等价物。

*   run finalization()–强制对象终结；没有 Go 等价物。

*   total memory()–获取 JVM 使用的内存；去`runtime.MemStats`。

*   version()–获取 JVM 版本；去`runtime.version()`。

由于 Go 是在构建时生成的一个完整的可执行文件，所以不需要加载系统库。

## Java IO 包

JRE 在`java.io`包中有一些关键类。下面的列表描述了任何 Go 环境的等价物。

界面摘要

*   可关闭–可以关闭。由 try with resources 使用；没有直接的 Go 对应物。

*   Data input–数据可以作为二进制编码值流读取。一些 Go 编码库提供了类似的功能。

*   Data output–数据可以作为二进制编码值流写入。一些 Go 编码库提供了类似的功能。

*   可外部化——可以使用非标准编码将数据读写到流中；没有直接的 Go 对应物。

*   file filter–选择与过滤器回调匹配的目录路径；没有直接的 Go 对应物。

*   filename filter–选择与过滤器回调匹配的文件名；没有直接的 Go 对应物。

*   可刷新–可以刷新(持久保存缓冲数据)；一些 Go `io`包接口提供了这个操作。

*   object input–可以读取 Java 序列化对象(DataInput 的超集)；没有直接的 Go 对应物。

*   object output–可以编写 Java 序列化对象(DataOutput 的超集)；没有直接的 Go 对应物。

*   serializable–通过默认编码将类型声明为可序列化；没有直接的 Go 对应物。

*   课程总结

*   BufferedInputStream–带缓冲区的输入流(字节数);Go `bufio`包提供了类似的支持。

*   BufferedOutputStream–带缓冲区的输出流(字节数);Go `bufio`包提供了类似的支持。

*   buffered reader–带缓冲区的输入编写器(在字符上);Go `bufio`包提供了类似的支持。

*   buffered writer–带缓冲区的输出编写器(在字符上);Go `bufio`包提供了类似的支持。

*   ByteArrayInputStream–从 byte[]读取；Go `io`包提供了类似的支持。

*   ByteArrayOutputStream–在 byte[]上写入；Go `io`包提供了类似的支持。

*   CharArrayReader–在 char[]上写入；Go `io`包提供了类似的支持。

*   CharArrayWriter–在 char[]上写入；Go `io`包提供了类似的支持。

*   控制台——标准输入、标准输出和标准错误的抽象。Go `io`包提供了类似的支持。

*   data inputstream–读取二进制编码值的流；一些 Go 编码库提供了类似的功能。

*   data output stream–写入二进制编码值的流；一些 Go 编码库提供了类似的功能。

*   文件-对文件(或目录)的访问；Go `io`和`os`包提供了类似的支持。

*   文件描述符-访问主机操作系统文件；Go `io`和`os`包提供了类似的支持。

*   file inputstream–从文件中读取字节；Go `io`和`os`包提供了类似的支持。

*   file output stream–将字节写入文件；Go `io`和`os`包提供了类似的支持。

*   file permission–访问文件权限；Go `io`和`os`包提供了类似的支持。

*   从文件中读取字符；Go `io`和`os`包提供了类似的支持。

*   FileWriter–将字符写入文件；Go `io`和`os`包提供了类似的支持。

*   InputStream–读取字节；Go `io`和`os`包提供了类似的支持。

*   InputStreamReader–将字节输入转换为字符输入；Go `io`和`os`包提供了类似的支持。

*   ObjectInputStream 读取序列化对象；没有直接的 Go 对应物。

*   object output stream–编写序列化对象；没有直接的 Go 对应物。

*   output stream–写入字节；Go `io`和`os`包提供了类似的支持。

*   output streamwriter–将字符转换为字节。

*   PrintStream–格式化的字节输出；Go `fmt`、`io`和`os`包提供了类似的支持。

*   PrintWriter–格式化字符输出；Go `fmt`、`io`和`os`包提供了类似的支持。

*   RandomAccessFile–支持查找的文件；Go `io`和`os`包提供了类似的支持。

*   阅读器-阅读字符；Go `fmt`、`io`和`os`包提供了类似的支持。

*   SequenceInputStream–连接输入流；Go `io`和`os`包提供了类似的支持。

*   stream tokenizer–标记流输入；Go `fmt`、`io`和`os`包提供了类似的支持。

*   string reader–从字符串中读取字符；Go `fmt`、`io`和`os`包提供了类似的支持。

*   作家——写人物；Go `io`和`os`包提供了类似的支持。

Java 还有一个 NIO(新 IO)包，提供更高级的文件(例如，监控文件更改)和目录服务。这本书不会涵盖他们。Go 库有一些功能可以与一些 NIO 类提供的功能相媲美。

## Java 文本包

JRE 在`java.text`包中有一些关键类。这个包提供了文本序列和消息格式的双向迭代。下面的列表描述了任何 Go 环境的等价物。一些 Go 扩展库和社区库提供了类似的支持。

界面摘要

*   AttributedCharacterIterator–对属性化文本序列进行双向迭代；没有直接的 Go 对应物。

*   character iterator——文本序列的双向迭代；没有直接的 Go 对应物；`utf8`和`utf16`包有一定的功能。

课堂总结。除非特别说明，Go 没有直接的对等词。

*   注释–类似注释的文本属性。

*   attributed String–带注释的字符串。

*   双向——提供双向遍历规则。

*   break iterator–迭代不同类型的断点(单词、行等)。).

*   choice format–帮助格式化具有不同计数和复数的消息。

*   CollationElementIterator–根据区域设置规则遍历中的字符。

*   collation Key–基于区域设置的排序规则的键。

*   collator–基于区域设置的排序规则的基类。

*   compact number format–使数字变小的十进制格式。

*   date format–格式化日期和时间；Go 有一个类似功能的`time`包。

*   decimal format–格式化十进制数字。

*   格式–各种格式类别的基本类别。

*   message format–用替换来格式化消息。

*   规格化器–规格化 Unicode 文本以帮助排序。

*   number format–数字格式化程序的基类。

*   RuleBasedCollator–规则表驱动程序整理器。

*   simple Date format–日期格式，具有可配置的日期和时间结构；Go 有一个类似功能的`time`包。

*   StringCharacterIterator–迭代字符串中的字符。

注意 Go 的`fmt`包可以用来完成各种格式类型的一些任务。此外，在 Java `String.format()`和 Go `fmt.Sprintf()`中可以做格式化程序做的很多事情。

## Java 时间包

JRE 在`java.time`包及其子包中有一些关键类。下面的列表描述了任何 Go 环境的等价物。

界面总结。Go 有`time`包可以解决其中的一些问题；大多是作为函数，而不是类型；这个函数只有一小部分出现在 Go 中。一些 Go 扩展库和社区库提供了类似的支持。

*   ChronoLocalDate–一些年表中的日期

*   ChronoLocalDateTime<d extends="" chronolocaldate="">–某些年表中的日期时间(时间戳)</d>

*   年表——一个日历系统(比如公历)

*   ChronoPeriod–一个时间段

*   ChronoZonedDateTime<d extends="" chronolocaldate="">>–某些年表中的时区日期时间(时间戳)</d>

*   纪元——一些年表中的有界范围(如公元前)

*   时间的；操作日期和时间的步骤如下

*   临时处理器

*   临时调整

*   临时账户

*   临时字段

*   风暴查询

*   临时的

课堂总结。Go 有如下一些行为的`time`包:主要是作为函数，而不是类型。

*   时钟-访问日期和时间

*   持续时间——一段时间；Go 具有持续时间类型

*   瞬间——一瞬间

*   local date–本地时区中的日期

*   local datetime–本地时区中的日期和时间(也称为时间戳)

*   local time–当地时区中的时间

*   月日-一个月中的一天

*   offset datetime–相对于 UTC 的日期和时间(即时间戳)偏移量

*   offset time–相对于 UTC 的时间偏移量

*   周期–以日历单位表示的持续时间

*   年–以年为单位的持续时间

*   year month–月份解析的瞬间

*   zoned DateTime–时区中的日期时间

*   datetime formatter–格式化日期时间

*   DateTimeFormatterBuilder–创建格式化程序

*   年表——年表(日历系统)的基础

*   年代学，年代学，等时学

*   日本年表，日本日期，日本

*   民国年表

*   ThaiBuddhistChronology，ThaiBuddhistDate

## Java Util 包

JRE 在`java.util`包及其子包中有一些关键类。子包处理对象集合、遗留日期和时间处理、并发(线程)处理以及对对象的并发访问。下面的列表描述了任何 Go 环境的等价物。

界面总结。大多数没有直接的 Go 等价物。这个函数的大部分是由 Go 内置类型提供的。一些 Go 扩展库和社区库提供了类似的支持。

*   集合<e>–e 类型的可迭代集合。</e>

*   比较器<t>–比较 t 类型的两个可比较对象。</t>

*   deque<e>–E 类双头队列；Go 切片已接近。</e>

*   枚举<e>–支持对 e 类型集合的正向迭代。</e>

*   event listener–形式化事件侦听器的类型(回调)。

*   format table–可以格式化。

*   迭代器<e>–支持对 e 类型集合的双向迭代。</e>

*   列表<e>–E 型可转位集合；Go 切片已接近。</e>

*   地图<k>，地图。条目<k>–具有键 K 的类型 V 的关联集合；Go 地图已经很接近了。</k></k>

*   队列<e>–E 类队列(FIFO)；Go 切片已接近。</e>

*   set<e>–一套 K 型；Go 地图已经很接近了。</e>

*   sorted Map<k>–带有排序关键字的地图。</k>

*   sorted Set<e>–用已排序的元素设置。</e>

课堂总结。大多数没有直接的 Go 等价物。

*   abstract collection<e>–以下是该类型的基本实现。</e>

*   抽象列表

*   摘要图〔t0〕

*   摘要队列〔t0〕

*   抽象序列列表

*   抽象集

*   array deque【<e>】阵列上的任何位置。</e>

*   ArrayList<e>–数组上的列表；Go 具有切片类型。</e>

*   数组——数组访问的助手。

*   Base64。解码器–解码 Base64 字符串；Go 有一个`base64`包。

*   Base64。编码器–对 Base64 字符串进行编码；Go 有一个`base64`包。

*   比特集–比特的集合；Go 有一个`bits`包。

*   日历–日历；Go 有一个`time`包。

*   集合——集合的助手。

*   货币——一种货币。

*   日期——一个日期；Go 有一个`time`包。

*   字典<k>–一种基本的地图类型；Go 有一个`map`型。</k>

*   枚举按钮<k extends="" enum="">，v></k>

*   枚举集〔t0〕

*   event list pro xy〔t0〕

*   事件对象

*   格式程序

*   公历-西历。

*   HashMap<k>–默认地图类型；Go 有一个地图类型。</k>

*   哈希集合<e>–默认集合类型。</e>

*   哈希表<k>–线程安全的哈希表。</k>

*   identity hashmap<k>–以对象标识为关键字的映射；Go 有一个 map[uintptr]类型。</k>

*   linked hashmap<k>–按照加法顺序迭代的地图。</k>

*   linked hashset<e>–按加法顺序迭代的集合。</e>

*   linked List<e>–由链表支持的列表。</e>

*   语言环境–定义区分语言环境的设置和行为。

*   对象–所有引用类型的帮助。

*   可选的<t>–空安全包装器。</t>

*   优先级队列<e>–按优先级排序的列表。</e>

*   properties–具有持久形式的键/值集合。

*   扫描仪-读取格式化输入；去`fmt`包。

*   simple time zone–一个时区实现。

*   堆栈<e>–按后进先出顺序处理的列表。</e>

*   String joiner–字符串助手。

*   string tokenizer–简单的字符串解析器；去`fmt`包。

*   定时器——每隔一段时间驱动事件(回调)。

*   TimerTask–间隔驱动事件(回调)。

*   时区–时区的基准。

*   树形图<k>–按关键字排序的地图。</k>

*   TreeSet<e>–按关键字排序的集合。</e>

*   UUID-UUID 型；可从第三方获得。

*   vector<e>–线程安全数组列表。</e>

*   weak hashmap<k>–不阻止键 GC 的映射。</k>

*   界面摘要

*   阻塞队列<e>–使用多个消费者线程进行排队。</e>

*   blocking Queue<e>–具有多个消费者线程的队列。</e>

*   可调用<v>–线程可以异步调用。</v>

*   concurrent map<k>–线程安全的高并发映射。</k>

*   ConcurrentNavigableMap<k>–线程安全的高并发映射。</k>

*   执行者——多线程的管理者。

*   ExecutorService–管理多个线程。

*   流动。处理器<t>–反应式编程流程处理器。</t>

*   流动。发布者<t>–反应式编程流程发布者。</t>

*   流动。订户<t>–反应式编程流程订户。</t>

*   流动。订阅–反应式编程流订阅。

*   未来<v>–可以在未来完成的异步任务。</v>

*   条件–带锁的外部化条件；去`sync.Cond`。

*   锁定-锁定对关键部分的访问；去`sync.Mutex`。

*   读写锁–多读，单写锁。

课堂总结。上述接口的部分或全部实现。名称通常描述该功能。许多没有直接的 Go 等价物。`LoadX`和`StoreX`函数提供了类似 Java `volatile`修饰符的行为。

*   抽象执行服务

*   array blockqueue

*   可完成的未来〔t0〕

*   并发哈希表

*   并发〔t0〕

*   并发队列〔t0〕

*   concurrentskiplistmap〔t0〕

*   concurrents kiplistset〔t0〕

*   CopyOnWriteArrayList

*   CopyOnWriteArraySet

*   CountDownLatch–等待计数；Go`sync.WaitGroup`也差不多。

*   cyclic barrier–允许多个线程到达同步点。

*   delay Queue<e extends="" delayed="">–延迟的队列(在特定时间使能)。</e>

*   交换器<v>–允许线程交换项目。</v>

*   executors–创建多线程的管理器。

*   ForkJoinPool–在多线程中分而治之。

*   未来<v>–可以在未来完成的异步任务。</v>

*   linkblockgdeque〔t0〕

*   linkeding queue〔t0〕

*   link transfer queue

*   phaser–线程同步；增强型 CyclicBarrier 或 CountDownLatch。

*   优先级阻塞队列

*   ScheduledThreadPoolExecutor

*   信号量——对关键会话的基本控制访问；Go 有一个`lock`包。

*   同步队列

*   线程池执行器

*   Atomic boolean——Atomic 允许线程间安全的读-修改-写循环；Go 有一个`async`包。

*   atomic integer–Go 有一个`async`包。

*   数组也能无锁

*   atomicintegerfieldupdater 更新程式

*   AtomicLong–Go 有一个`async`包。

*   原子克隆阵列

*   atomiclongfieldupdater 更新程式

*   AtomicMarkableReference

*   atomic reference<v>–Go 有一个`async`包。</v>

*   原子分离机阵列〔t0〕

*   原子参考场更新

*   AtomicStampedReference

*   双累加器——累加器/加法器支持线程间的安全读-修改-写循环。

*   double adder–Go 有一个`async`包。

*   long accumulator–Go 有一个`async`包。

*   long adder–Go 有一个`async`包。

*   锁定支持–锁定助手。

*   reentrant Lock–锁实现。

*   可重入读写锁

注意 Java 中的锁(和`synchronized`访问)是可重入的；同一个线程可以多次获取锁。在 Go 中，锁是不可重入的，试图重新获取锁的同一个 goroutine 会阻塞(死锁)自己。