# 十三、关键方法/函数比较

这个库调查并不打算作为程序员的参考，更多的是一个介绍。它将有一些有用的库函数的例子，应该可以帮助你在 Go 库中获得基本的能力。要更深入地理解提到的函数和任何省略的函数，请参阅在线 Go 软件包文档。该文档描述了每个包及其类型和它们提供的功能。也有一些例子。

Go 和 Java 库有很大的不同，因为 Java 是面向对象的，而 Go 不是。在 Java 中，许多函数是某个接收者类型的实例方法(隐含的`this`参数)。在 Go 中，它们通常是通用函数，将接收者类型作为第一个参数。这更像是 Java 中的`static`方法。例如，以将字符串转换为全大写的函数为例。

在 Java 中，这是

```go
var uc = "some string".toUpperCase();

```

在 Go 中，这是

```go
var uc = strings.ToUpper("some string")

```

这些函数的最大区别在于接收方传递给函数的方式。Java 实例方法可以被定义为

```go
public class String {
  public static String toUpperCase(String s) {
    :
  }
}

```

其访问方式如下

```go
var uc = String.toUpperCase("some string")

```

因此，很像 Go 的风格，但是 Java 设计者选择不这样做。

Java 有很多很多的库函数。下面的表格将总结几个常用的 Java 函数，跨越几个 Java 包和类型，列出它们的 Java 和 Go 等价物(其行为可能不总是与 Java 函数完全一样)。有关 Go 函数的更完整列表，请参见“Go 库调查”部分。

静态 Java 方法和 Go 包顶层函数都是以句点(" . ")，如表 [13-1](#Tab1) 至 [13-8](#Tab8) 所示。实例方法有一些变量(接收者)。

表 13-1

关键 Java 对象类方法

<colgroup><col class="tcol1 align-left"> <col class="tcol2 align-left"> <col class="tcol3 align-left"> <col class="tcol4 align-left"></colgroup> 
| 

Java 函数

 | 

Go 等效

 | 

Go 包

 | 

笔记

 |
| --- | --- | --- | --- |
| `o.toString()` | `o.String()` | `Many` | 如果由接收类型定义 |
| `o.wait()` | `c.Wait()` | `sync.Cond` |   |
| `o.notify()` | `c.Signal()` | `sync.Cond` |   |
| `o.notifyAll()` | `c.Broadcast()` | `sync.Cond` |   |
| `o.hashCode` |   |   | 没有 Go 等价物 |
| `o.equals(x)` | `x ==`<sup>T3】1T5`y`</sup> |   | Go 有一个运算符 |

表 13-2

关键 Java 运行时类方法

<colgroup><col class="tcol1 align-left"> <col class="tcol2 align-left"> <col class="tcol3 align-left"> <col class="tcol4 align-left"></colgroup> 
| 

Java 函数

 | 

Go 等效

 | 

Go 包

 | 

笔记

 |
| --- | --- | --- | --- |
| `add/removeShutdownHook` | `Signal` | `os` | Go 只监听信号 |
| `availableProcessors` | `NumCPU()` | `runtime` |   |
| `exec(...)` | `Exec(...)` | `runtime` | 许多变体 |
| `exit(n)` |   |   | 没有 Go 等价物 |
| `freeMemory()` | `ReadMemoryStats()` | `runtime` |   |
| `maxMemory()` | `ReadMemoryStats()` | `runtime` |   |
| `gc()` | `GC()` | `runtime` |   |
| `halt()` | `Exit(n)` | `runtime` |   |
| `runFinalization()` |   |   | 没有 Go 等价物 |
| `version()` | `Version()` | `runtime` |   |

表 13-3

关键的 Java 系统类方法(运行时的重复被省略)

<colgroup><col class="tcol1 align-left"> <col class="tcol2 align-left"> <col class="tcol3 align-left"> <col class="tcol4 align-left"></colgroup> 
| 

Java 函数

 | 

Go 等效

 | 

Go 包

 | 

笔记

 |
| --- | --- | --- | --- |
| `.arraycopy(a,...)` | `a[:]` |   | Go 运算符 |
| `.clear/set/getProperty``.getProperties()` |   |   | 没有 Go 等价物 |
| `.console` |   |   | 没有 Go 等价物 |
| `.currentTimeMillis` | `Now()` | `time` |   |
| `.getenv()` | `Environ()` | `os` |   |
| `.getLogger()` |   | `log` | Go 使用记录器方法 |
| `.get/SetSecurityManager()` |   |   | 没有 Go 等价物 |
| `.identityHashCode()` |   |   | 没有 Go 等价物 |
| `.lineSeparator` |   |   | Go 一般忽略 CR |
| `.nanoTime()` | `Now()` | `time` |   |
| `SetIn/Out/Err` |   |   | 可以设置 os.Stdin/out/err 值 |

表 13-4

关键 Java 字符串类方法

<colgroup><col class="tcol1 align-left"> <col class="tcol2 align-left"> <col class="tcol3 align-left"> <col class="tcol4 align-left"></colgroup> 
| 

Java 函数

 | 

Go 等效

 | 

Go 包

 | 

笔记

 |
| --- | --- | --- | --- |
| `s.substr(s,e)` | `s[s:e]` |   | Go 运算符 |
| `s.charAt(i)` | `s[i]` |   | Go 运算符 |
| `s.indexOf(s2)` | `.Index(s, o)` | `strings` |   |
| `s.lastIndexOf(s2)` |   | `strings` |   |
| `s.indexOf(c)` |   | `strings` |   |
| `s.lastIndexOf(c)` |   | `strings` |   |
| `s.toUpperCase()` | `.ToUpper(s)` | `strings` |   |
| `s.toLowerCase()` | `.ToLower(s)` | `strings` |   |
| `s.toCharArray()` | `[]byte(s)` |   | 转到转换 |
| `s.length()` | `len(s)` |   | 内置 |
| `s.compareTo(o)` | `s op o` |   | Go 运算符:< <= == != > >= |
| `s.startsWith(s2)` | `.HasPrefix(s1, s2)` | `strings` |   |
| `s.endsWith(s2)` | `.HasSuffix(s1, s2)` | `strings` |   |
| `s.contains(s2)` | `.Index(s,s2) >= 0` | `strings` |   |
| `s1 + s2` | `s1 + s2` |   | Go 运算符 |
| `.join(delim,s...)` | `.join(delim,s...)` | `strings` |   |
| `s.getBytes()` | `[]byte(s)` |   | 转到转换 |
| `s.matches(s1)` | `.matches(s, s1)` | `regex` |   |
| `s.repeat(n)` | `.Repeat(s, n)` | `strings` |   |
| `s.replace(c1,c2)` |   |   | 没有直接的对等物 |
| `s.replace(s1,s2)` | `.ReplaceAll(s, s1, s2)` | `strings` |   |
| `s.replaceAll(p,s2)` |   | `regex` |   |
| `s.replaceFirst(p,s2)` |   | `regex` |   |
|   | `.Split(s, s2)` | `strings` |   |
|   | `.Split(s, s2, n)` |   |   |
| `s.split(p)` |   |   | 没有直接的对等物 |
| `s.split(p,n)` | `.Split(s, n)` | `regex` |   |
| `s.trim()``s.strip()``s.stripLeading()``s.stripTrailing()` | `.TrimSpaces(s)` | `strings` |   |
| `s.substring(p)``s.substring(p,e)``s.substring(0,e)``s.substring(0,s.length())` | `s[p:]``s[p:e]``s[:e]``s[:]` |   | Go 运算符 |
| `.valueOf(x)``.format(f, ...)` | `.Sprintf("%v",x)``.Sprintf(f,...)` | `fmt` |   |

表 13-5

关键 Java StringBuilder 类方法

<colgroup><col class="tcol1 align-left"> <col class="tcol2 align-left"> <col class="tcol3 align-left"> <col class="tcol4 align-left"></colgroup> 
| 

Java 函数

 | 

Go 等效

 | 

Go 包

 | 

笔记

 |
| --- | --- | --- | --- |
| `sb.append(o)` | `b.Write``String(s)` | `strings.``Builder` |   |
| `sb.length()` | `b.Size()` | `strings.``Builder` |   |

表 13-6

关键 Java 列表接口方法

<colgroup><col class="tcol1 align-left"> <col class="tcol2 align-left"> <col class="tcol3 align-left"> <col class="tcol4 align-left"></colgroup> 
| 

Java 函数

 | 

Go 等效

 | 

Go 包

 | 

笔记

 |
| --- | --- | --- | --- |
| `l.add(x)` | `append(s,x)` |   | 内置 |
| `l.size()` | `len(l)` |   | 内置 |
| `l.get(i)` | `l[i]` |   | Go 运算符 |
| `l.set(I, x)` | `l[i] = x` |   | Go 运算符 |

表 13-7

关键的 Java Map 接口方法

<colgroup><col class="tcol1 align-left"> <col class="tcol2 align-left"> <col class="tcol3 align-left"> <col class="tcol4 align-left"></colgroup> 
| 

Java 函数

 | 

Go 等效

 | 

Go 包

 | 

笔记

 |
| --- | --- | --- | --- |
| `m.put(k,v)` | `m[k] = v` |   | Go 运算符 |
| `m.size()` | `len(m)` |   | 内置 |
| `m.get(k)` | `m[k]` |   | 内置缺失测试 |

表 13-8

关键 Java PrintWriter 类方法

<colgroup><col class="tcol1 align-left"> <col class="tcol2 align-left"> <col class="tcol3 align-left"> <col class="tcol4 align-left"> <col class="tcol5 align-left"></colgroup> 
| 

Java 函数

 | 

Java 类型

 | 

Go 等效

 | 

Go 包

 | 

笔记

 |
| --- | --- | --- | --- | --- |
| `pw.println(o)` | `Print``Writer` | `w.Write(o)``w.Write('\n')` | `io.Writer` |   |

<aside aria-label="Footnotes" class="FootnoteSection" epub:type="footnotes">Footnotes [1](#Fn1_source)

有些类型有一个等同于 equals 的方法。

 </aside>