# 十五、SQL 数据库访问

Java 在`java.sql`和`java.sqlx`包中提供了多级基于 SQL 的数据库访问，称为 *JDBC* (也称为 Java 数据库连接)。Go 与`sql`和`driver`软件包做了几乎相同的事情，但是功能更少。`sql`包提供了一个通用的功能框架，就像`java.sql`为 SQL 数据库提供的一样，而驱动程序包是一个*系统编程接口* (SPI)，允许使用可插拔的驱动程序。大多数 SQL 操作都是由驱动程序完成的。与 Java 不同，大多数 Go 驱动程序来自社区资源，而不是数据库供应商。

Java 有社区支持(例如，*Hibernate*<sup>[1](#Fn1)</sup>或 *Java 持久性架构* <sup>[2](#Fn2)</sup> (JPA)实现)用于*对象关系映射器* (ORMs) <sup>[3](#Fn3)</sup> 使保存持久对象(或*实体*；Java 中的类，Go 中的结构)要容易得多。许多 Java 开发人员使用这些 ORM，而不是 JDBC 提供的更基本的 CRUD<sup>[4](#Fn4)</sup>SQL<sup>[5](#Fn5)</sup>access。

以下软件包提供类似 JDBC 的低级访问。有一些社区包，比如 GORM，增加了对 Go 的 ORM 支持。

`sql`包很大，有几种类型和功能。

主要类型有

*   ColType–定义表格列的类型

*   conn–表示单个数据库连接

*   DB——代表一个连接池

*   row–表示单个返回的表格行(行的子情况)

*   rows–表示多个返回的表格行

*   扫描仪–访问行中各列的界面

*   stmt–表示一条准备好的 SQL 语句

*   tx–表示一个数据库事务

`sql`包具有以下功能:

*   func Drivers()[]string–获取注册的驱动程序名称

*   func 寄存器(名称字符串，驱动程序驱动程序。驱动程序)–注册驱动程序

*   ColumnType 提供了这些函数。它们是不言自明的:

*   func(ci * column type)database typename()string

*   func(ci * ColumnType)decimal size()(precision，scale int64，ok bool)

*   func(ci * ColumnType)Length()(Length int 64，ok bool)

*   func(ci * column type)name()string

*   func(ci * column type)null()(null，ok bool)

*   func(ci * ColumnType)ScanType()reflect。类型

Conn 提供连接级访问功能:

*   func (c *Conn) BeginTx(ctx 上下文。Context，opts *TxOptions) (*Tx，error)-启动事务

*   func (c *Conn) Close()错误–关闭连接；把它放回水池

*   func(c * Conn)exec context(CTX context。上下文，查询字符串，参数...接口{})(结果，错误)-执行非查询 SQL

*   func(c * Conn)ping context(CTX context。上下文)错误–查看连接是否可以工作

*   func(c * Conn)prepare context(CTX context。上下文，查询字符串)(*Stmt，error)-准备一条 SQL 语句

*   func (c *Conn) QueryContext(ctx 上下文。上下文，查询字符串，参数...interface{}) (*Rows，error)–执行可以返回许多行的 SQL 查询

*   func(c * Conn)QueryRowContext(CTX context。上下文，查询字符串，参数...interface { })* Row–执行将返回< = 1 行的 SQL 查询

DB 提供数据库级访问功能:

*   func Open(driverName，dataSourceName string) (*DB，error)–按名称打开数据库

*   c 驱动程序。连接器)*数据库–打开数据库

*   func (db *DB) Begin() (*Tx，error)-启动一个事务

*   func (db *DB) BeginTx(ctx 上下文。Context，opts *TxOptions) (*Tx，error)–用选项启动事务

*   func (db *DB) Close()错误–关闭数据库

*   func (db *DB) Conn(ctx 上下文。上下文)(*连接，错误)–获取到数据库的连接

*   func (db *DB) Driver()驱动程序。驱动程序–为数据库获取一个驱动程序

*   func (db *DB) Exec(查询字符串，参数...接口{})(结果，错误)–执行通用 SQL

*   func(DB * DB)exec context(CTX context。上下文，查询字符串，参数...接口{})(结果，错误)–执行通用 SQL

*   func (db *DB) Ping()错误–测试数据库可用

*   func(DB * DB)ping context(CTX context。上下文)错误–测试数据库可用

*   func (db *DB)准备(查询字符串)(*Stmt，error)-准备 SQL

*   func (db *DB) PrepareContext(ctx 上下文。上下文，查询字符串)(*Stmt，error)-准备 SQL

*   func (db *DB)查询(查询字符串，参数...接口{ })(*行，错误)–执行一般选择

*   func (db *DB) QueryContext(ctx 上下文。上下文，查询字符串，参数...接口{ })(*行，错误)–执行一般选择

*   func (db *DB) QueryRow(查询字符串，参数...interface { })* Row–执行查询

*   func(DB * DB)query rowcontext(CTX context。上下文，查询字符串，参数...interface { })* Row–执行查询

*   func(DB * DB)Stats()DBStats–获取各种数据库访问统计信息

行是单一选择结果行:

*   func(r * Row)Err()error–获取任何执行错误

*   功能(右*行)扫描(目的地...接口{})错误–将返回的数据复制到变量中

Rows 是一组选择结果行:

*   func (rs *Rows) Close()错误–表示现在已完成对行的处理

*   func(RS * Rows)column types()([]* ColumnType，error)-获取列元数据

*   func(RS * Rows)Columns()([]string，error)-获取列名

*   func(RS * Rows)Err()error–获取任何执行错误

*   func(RS * Rows)Next()bool–前进到下一行

*   func(RS * Rows)next resultset()bool–前进到下一个结果集

*   函数(rs *行)扫描(目标...接口{})错误–将返回的数据复制到变量中

Stmt 提供 SQL 语句级访问函数:

*   func (s *Stmt)关闭()错误

*   func (s *Stmt) Exec(args...接口{})(结果，错误)

*   func(s * Stmt)exec context(CTX context。上下文，参数...接口{})(结果，错误)

*   func (s *Stmt)查询(参数...接口{ })(*行，错误)

*   func (s *Stmt) QueryContext(ctx 上下文。上下文，参数...接口{ })(*行，错误)

*   func (s *Stmt) QueryRow(args...接口{ })*行

*   func(s * Stmt)QueryRowContext(CTX 上下文。上下文，参数...接口{ })*行

Tx 提供事务级访问功能。见前面类似的描述。

*   func (tx *Tx) Commit()错误–提交任何更改

*   func (tx *Tx) Exec(查询字符串，参数...接口{})(结果，错误)

*   func(Tx * Tx)exec context(CTX context。上下文，查询字符串，参数...接口{})(结果，错误)

*   func (tx *Tx)准备(查询字符串)(*Stmt，错误)

*   func (tx *Tx) PrepareContext(ctx 上下文。上下文，查询字符串)(*Stmt，错误)

*   func (tx *Tx)查询(查询字符串，参数...接口{ })(*行，错误)

*   func (tx *Tx) QueryContext(ctx 上下文。上下文，查询字符串，参数...接口{ })(*行，错误)

*   func (tx *Tx) QueryRow(查询字符串，参数...接口{ })*行

*   func (tx *Tx) QueryRowContext(ctx 上下文。上下文，查询字符串，参数...接口{ })*行

*   func (tx *Tx) Rollback()错误–回滚(取消)任何更改

*   func(Tx * Tx)Stmt(Stmt * Stmt)* Stmt–获取此事务中的一条语句

*   func (tx *Tx) StmtContext(ctx 上下文。上下文，stmt *Stmt) *Stmt

作为使用`sql`包的一个例子，清单 [15-1](#PC1) 展示了一个对简单表格进行 CRUDs 的程序。

```go
// Table row entity
type DBEntity struct {
     name  string
     value string
}

// Do in a DB context.
func DoInDB(driverName, datasourceParams string, f func(db *sql.DB) error) (err error) {
     db, err := sql.Open(driverName, datasourceParams)
     if err != nil {
          return
     }
     defer db.Close()
     err = f(db)
     return
}

// Do in a connection.
func DoInConn(db *sql.DB, ctx context.Context, f func(db *sql.DB, conn *sql.Conn, ctx context.Context) error) (err error) {
     conn, err := db.Conn(ctx)
     if err != nil {
          return
     }
     defer conn.Close()
     err = f(db, conn, ctx)
     return
}

// Do in a transaction.
func DoInTx(db *sql.DB, conn *sql.Conn, ctx context.Context, txOptions *sql.TxOptions, f func(tx *sql.Tx) error) (err error) {
     if txOptions == nil {
          txOptions = &sql.TxOptions{Isolation: sql.LevelSerializable}
     }
     tx, err := db.BeginTx(ctx, txOptions)
     if err != nil {
          return
     }

     err = f(tx)
     if err != nil {
          _ = tx.Rollback()
          return
     }
     err = tx.Commit()
     if err != nil {
          return
     }
     return
}

var ErrBadOperation = errors.New("bad operation")

// Execute a SQL statement.
func ExecuteSQL(tx *sql.Tx, ctx context.Context, sql string, params ...interface{}) (count int64, values []*DBEntity, err error) {
     lsql := strings.ToLower(sql)
     switch {

     // process query
     case strings.HasPrefix(lsql, "select "):
          rows, xerr := tx.QueryContext(ctx, sql, params...)
          if xerr != nil {
               err = xerr
               return
          }
          defer rows.Close()
          for rows.Next() {
               var name string
               var value string
               if err = rows.Scan(&name, &value); err != nil {
                    return
               }
               data := &DBEntity{name, value}
               values = append(values, data)
          }
          if xerr := rows.Err(); xerr != nil {
               err = xerr
               return
          }

     // process an update
     case strings.HasPrefix(lsql, "update "), strings.HasPrefix(lsql, "delete "), strings.HasPrefix(lsql, "insert "):
          result, xerr := tx.ExecContext(ctx, sql, params...)
          if xerr != nil {
               err = xerr
               return
          }
          count, xerr = result.RowsAffected()
          if xerr != nil {
               err = xerr
               return
          }

     default:
          err = ErrBadOperation  // INSERT and DELETE not demo’ed here
          return
     }
     return
}

Listing 15-1Sample DB Access (Part 1)

```

该库由该测试函数驱动(清单 [15-2](#PC2) )。

```go
func testDB() {
     values := make([]*DBEntity, 0, 10)
     values = append(values, &DBEntity{"Barry", "author"},
          &DBEntity{"Barry, Jr.", "reviewer"})

     err := DoInDB("postgres", "postgres://postgres:postgres@localhost:5432/postgres?sslmode=disable",
          func(db *sql.DB) (err error) {
               err = DoInConn(db, context.Background(), func(db *sql.DB, conn *sql.Conn,
                       ctx context.Context) (err error) {
                    err = createRows(db, conn, ctx, values)
                    if err != nil {
                         return
                    }

                    // must be done in separate transaction to see the change
                    err = queryRows(db, conn, ctx)
                    return
               })
               return
          })
     if err != nil {
          fmt.Printf("DB access failed: %v\n", err)
     }
}

// Create data rows.
func createRows(db *sql.DB, conn *sql.Conn, ctx context.Context, values []*DBEntity) (err error) {
     err = DoInTx(db, conn, ctx, nil, func(tx *sql.Tx) (err error) {
          // first remove any old rows
          count, _, err := ExecuteSQL(tx, ctx, `delete from xvalues`)
          if err != nil {
               return
          }
          fmt.Printf("deleted %d\n", count)
          // insert new rows
          for _, v := range values {
               count1, _, xerr := ExecuteSQL(tx, ctx, fmt.Sprintf(`insert into xvalues(name, value) values('%s', '%s')`, v.name, v.value))
               if xerr != nil || count1 != 1 {
                    err = xerr
                    return
               }
               fmt.Printf("inserted %q = %q\n", v.name, v.value)
          }
          // update a row
          v := &DBEntity{"Barry", "father"}
          _, _, xerr := ExecuteSQL(tx, ctx, fmt.Sprintf(`update xvalues set value='%s' where name='%s'`, v.value, v.name))
          if xerr != nil {
               err = xerr
               return
          }

          fmt.Printf("updated %q = %q\n", v.name, v.value)
          return
     })
     return
}

// Query and print all rows.
func queryRows(db *sql.DB, conn *sql.Conn, ctx context.Context) (err error) {
     err = DoInTx(db, conn, ctx, nil, func(tx *sqB.Tx) (err error) {
          _, xvalues, err := ExecuteSQL(tx, ctx, `select name, value from xvalues`)
          if err != nil {
               return
          }
          for _, v := range xvalues {
               fmt.Printf("queried %q = %q\n", v.name, v.value)
          }
          return
     })
     return
}

Listing 15-2Sample DB Access (Part 2)

```

注意确保数据库资源关闭的嵌套方法。释放资源是很重要的，特别是对于长时间运行的程序，比如服务器。顶层是对数据库连接池的访问。然后从池中访问单个连接。最后，有一个要在其中执行 SQL 语句的事务。在这个例子中，多个事务在一个连接中完成，这是很典型的。此外，多条语句通常在一个事务中完成。

该程序输出以下内容:

```go
deleted 2
inserted "Barry" = "author"
inserted "Barry, Jr." = "reviewer"
updated "Barry" = "father"queried "Barry, Jr." = "reviewer"
queried "Barry" = "father"

```

注意删除了两条记录，因为该输出来自程序的第二次执行。

当程序结束时，数据库中的数据如图 [15-1](#Fig1) 所示。

![img/516433_1_En_15_Fig1_HTML.jpg](img/516433_1_En_15_Fig1_HTML.jpg)

图 15-1

数据库示例结果(在 PostgreSQL 中)

<aside aria-label="Footnotes" class="FootnoteSection" epub:type="footnotes">Footnotes [1](#Fn1_source)

[`https://en.wikipedia.org/wiki/Hibernate_(framework)`](https://en.wikipedia.org/wiki/Hibernate_%2528framework%2529) 和 [`https://hibernate.org/`](https://hibernate.org/)

  [2](#Fn2_source)

[T2`https://en.wikipedia.org/wiki/Jakarta_Persistence`](https://en.wikipedia.org/wiki/Jakarta_Persistence)

  [3](#Fn3_source)

[T2`https://en.wikipedia.org/wiki/Object%E2%80%93relational_mapping`](https://en.wikipedia.org/wiki/Object%25E2%2580%2593relational_mapping)

  [4](#Fn4_source)

CRUD–创建、读取(又名查询)、更新、删除。

  [5](#Fn5_source)

结构化查询语言——关系数据库系统的标准 API。

 </aside>