# 二、Go 基础

第 1 章概述了 Go 编程语言和 Go 生态系统的主要组成部分。这一章包含了处理 Go 语言核心基础的方法。Go 是一种简单的编程语言，它提供了构建可伸缩软件系统的基本特性。与 C#和 Java 等其他编程语言不同，Go 在语言规范中提供了最少的特性，以保持其作为简单、最小语言的设计目标。尽管它是一种简单的语言，但 Go 提供了构建可靠而高效的软件系统所必需的语言。这一章中的方法涉及到编写函数、处理各种集合类型、错误处理以及用关键字`defer`、`panic`和`recover`实现的 Go 的独特特性，等等。

## 2-1.在 Go 中编写函数

### 问题

如何在函数中管理 Go 代码？

### 解决办法

关键字`func`用于声明函数。一个函数用一个名字、一个参数列表、一个可选的返回类型列表和一个编写函数逻辑的主体来声明。

### 它是如何工作的

Go 中的函数是一段可重用的代码，它将一系列代码语句组织成一个单元，可以从包中调用，如果函数被导出到其他包中，也可以从其他包中调用。因为函数是可重用的代码，所以可以多次调用这个表单。当您编写共享库包时，名称以大写字母开头的函数将被导出到其他包中。如果函数名以小写字母开头，它不会被导出到其他包中，但是您可以在同一个包中调用这个函数。

#### 声明函数

下面是在 Go 中编写函数的语法:

```go
func name(list of parameters)  (list of return types)
{
   function body
}

```

函数参数指定名称和类型。当调用者调用一个函数时，它提供函数参数的实参。在 Go 中，一个函数可以返回多个值。返回类型列表指定了函数返回值的类型。您在函数体中编写代码语句。清单 [2-1](#Par8) 显示了一个将两个整数值相加的示例函数。

```go
func Add(x, y int) int {
    return x + y
}
Listing 2-1.An Example Function That Adds Two Integer Values

```

声明了一个函数`Add`，它有两个类型为`integer`的参数，该函数返回一个整数值。使用`return`语句提供函数的返回值。

清单 [2-2](#Par11) 显示了调用这个`Add`函数的代码块。

```go
x, y := 20, 10
result := Add(x, y)
Listing 2-2.Code Block That Calls the Add Function

```

两个整数变量`x`和`y`被初始化，为调用`Add`函数提供参数。局部变量`result`用`Add`函数返回的返回值初始化。

清单 [2-3](#Par14) 展示了一个示例程序，它声明了两个函数并从一个`main`函数中调用它。

```go
package main

import (
    "fmt"
)

func Add(x, y int) int {
    return x + y
}

func Subtract(x, y int) int {
    return x - y
}

func main() {
    x, y := 20, 10

    result := Add(x, y)
    fmt.Println("[Add]:", result)

    result = Subtract(x, y)
    fmt.Println("[Subtract]:", result)
}

Listing 2-3.Example Program That Defines and Calls Functions

```

在这个程序中，声明了两个函数:`Add`和`Subtract`。这两个函数是从`main`函数中调用的。

运行该程序时，您应该会看到以下输出:

```go
[Add]: 30
[Subtract]: 10

```

#### 命名返回值

编写函数时，可以通过在函数顶部定义变量来命名返回值。清单 [2-4](#Par19) 显示了带有指定返回值的`Add`函数。

```go
func Add(x, y int) (result int) {
    result = x + y
    return

}
Listing 2-4.
Add Function with Named Return Values

```

`integer`类型的变量`result`在函数声明中为函数返回值指定。当您指定指定的返回值时，您可以将返回值赋给指定的变量，并且可以通过简单地指定`return`关键字来退出函数，而不需要随`return`语句一起提供返回值。

```go
    result = x + y
    return

```

这个`return`语句返回在函数声明中指定的命名返回值。这就是所谓的裸归。我不推荐这种方法，因为它会影响程序的可读性。

#### 返回多个值

Go 是一种在其语言设计中提供了很多实用主义的语言。在 Go 中，可以从一个函数返回多个值，这在很多实际场景中是一个很有用的特性。

清单 [2-5](#Par25) 展示了一个示例程序，它声明了一个具有两个返回值的函数，并从一个`main`函数中调用它。

```go
package main

import (
    "fmt"
)

func Swap(x, y string) (string, string) {
    return y, x
}

func main() {
    x, y := "Shiju", "Varghese"
    fmt.Println("Before Swap:", x, y)

    x, y = Swap(x, y)
    fmt.Println("After Swap:", x, y)
}

Listing 2-5.An Example Program That Uses a Function with Multiple Return Values

```

名为`Swap`的函数是用两个`string`类型的返回值声明的。`Swap`函数交换两个字符串值。我们从`main`函数中调用`Swap`函数。

运行该程序时，您应该会看到以下输出:

```go
Before Swap: Shiju Varghese
After Swap: Varghese Shiju

```

#### 可变函数

可变函数是接受可变数量参数的函数。当您不知道要传递给函数的参数数量时，这种类型的函数非常有用。`fmt`包的内置`Println`函数是可变函数的一个例子，它可以接受可变数量的参数。

清单 [2-6](#Par31) 显示了一个提供变量函数`Sum`的示例程序，它接受数量可变的`integer`类型的参数。

```go
package main

import (
    "fmt"
)

func Sum(nums ...int) int {
    total := 0
    for _, num := range nums {
        total += num
    }
    return total
}

func main() {
    // Providing four arguments
    total := Sum(1, 2, 3, 4)
    fmt.Println("The Sum is:", total)

    // Providing three arguments
    total = Sum(5, 7, 8)
    fmt.Println("The Sum is:", total)
}

Listing 2-6.Example Program with Variadic Function

```

表达式`. . .`用于指定参数表的可变长度。当调用者向`nums`参数提供值时，它可以提供可变数量的整数值参数。`Sum`函数提供了调用者提供的可变数量的参数的总和。该函数使用`range`构造迭代`nums`参数的值，以获得调用者提供的参数的总值。在`main`函数中，`Sum`函数被调用两次。每次都提供可变数量的参数。

运行该程序时，您应该会看到以下输出:

```go
The Sum is: 10
The Sum is: 20

```

当调用变量函数时，可以提供切片(动态数组)作为参数。你将在本章的后面学习切片。清单 [2-7](#Par36) 显示了通过提供一个切片作为参数来调用变量函数的代码块。

```go
// Providing a slice as an argument
nums := []int{1, 2, 3, 4, 5}
total = Sum(nums...)
fmt.Println("The Sum is:", total)
Listing 2-7.Code Block That Calls a Variadic Function with a Slice

```

当您提供切片作为参数时，您必须在切片值后提供表达式`...`。

#### 函数值、匿名函数和闭包

尽管 Go 是一种静态类型的语言，但 Go 的实用主义给开发人员带来了像动态类型语言一样的生产力。Go 中的函数为 Go 程序员提供了很大的灵活性。函数类似于值，这意味着您可以将函数值作为参数传递给其他返回值的函数。Go 还提供了对匿名函数和闭包的支持。匿名函数是没有函数名的函数定义。当您希望在不提供函数标识符的情况下内联形成函数时，这很有用。

清单 [2-8](#Par40) 显示了一个示例程序，它演示了将一个匿名函数作为参数传递给另一个函数，其中匿名函数封闭变量以形成闭包。

```go
package main

import (
    "fmt"
)

func SplitValues(f func(sum int) (int, int)) {
    x, y := f(35)
    fmt.Println(x, y)

    x, y = f(50)
    fmt.Println(x, y)
}

func main() {
    a, b := 5, 8
    fn := func(sum int) (int, int) {
        x := sum * a / b
        y := sum - x
        return x, y
    }

    // Passing function value as an argument to another function
    SplitValues(fn)

    // Calling the function value by providing argument
    x, y := fn(20)
    fmt.Println(x, y)
}

Listing 2-8.Example Program Demonstrating Passing Function as Value, Anonymous Function, and Closure

```

在`main`函数中，声明了一个匿名函数，并将匿名函数的值赋给一个名为`fn`的变量。

```go
    a, b := 5, 8
    fn := func(sum int) (int, int) {
        x := sum * a / b
        y := sum - x
        return x, y
    }

```

匿名函数在`main`函数中声明。在 Go 中，可以在函数内部编写函数。匿名函数使用任意逻辑将一个值拆分成两个值。为了形成任意逻辑，它访问在`main`函数的外部函数中声明的几个变量的值。

匿名函数被赋给变量`fn and`，将函数值传递给另一个名为`SplitValues`的函数。

```go
SplitValues(fn)

```

`SplitValues`函数接收一个函数作为参数。

```go
func SplitValues(f func(sum int) (int, int)) {
    x, y := f(35)
    fmt.Println(x, y)

    x, y = f(50)
    fmt.Println(x, y)
}    

```

在`SplitValues`函数中，作为参数传递的参数值被调用几次，以将值分成两个值。返回值被打印到控制台窗口。

让我们回到匿名函数。在`main`函数中，匿名函数的值用于两件事:通过将函数值作为参数传递来调用`SplitValues`函数，以及通过提供一个值作为参数来拆分整数值来直接调用函数值。

```go
// Passing function value as an argument to another function
    SplitValues(fn)

    // Calling the function value by providing argument
    x, y := fn(20)
    fmt.Println(x, y)

```

值得注意的是，匿名函数正在访问外部函数中声明的两个变量:

```go
a, b := 5, 8.

```

变量`a`和`b`在`main`函数中声明，但是匿名函数(内部函数)可以访问这些变量。当您通过将匿名函数的值作为参数传递来调用`SplitValues`函数时，匿名函数也可以访问变量`a`和`b`。匿名函数关闭`a`和`b`的值，使其成为闭包。不管从哪里调用匿名函数的值，它都可以访问外部函数中声明的变量`a`和`b`。

运行上述程序时，您应该会看到以下输出:

```go
21 14
31 19
12 8

```

## 2-2.使用数组

### 问题

您希望将元素集合存储到固定长度的数组类型中。

### 解决办法

Go 的数组类型允许您存储单一类型的固定大小的元素集合。

### 它是如何工作的

数组是由单一类型的元素集合组成的数据结构。数组是固定大小的数据结构，通过指定长度和元素类型来声明。

#### 声明和初始化数组

下面是声明数组的代码块:

```go
var x [5]int

```

变量`x`被声明为由五个`int`类型的元素组成的数组。数组`x`允许你存储`integer values.`的五个元素，你通过指定从 0 开始的索引来赋值给一个数组。下面是为数组`x`的第一个元素赋值的表达式:

```go
x[0]=5

```

表达式`x[4]=25`为数组`x`的最后一个元素(第五个元素)赋值。

您还可以使用数组文字来声明和初始化数组，如下所示:

```go
y := [5]int {5,10,15,20,25}

```

当使用数组文字初始化数组时，可以为特定元素提供值，如下所示:

```go
langs := [4]string{0: "Go", 3: "Julia"}

```

一个`string`类型的数组被声明为大小为 4，但是只为第一个元素(索引 0)和最后一个元素(索引 3)提供值。您将获得没有初始化的元素的默认值。对于字符串类型，它是空字符串；对于整数类型，它是 0；对于布尔类型，它是 false。如果你试图返回`langs[1]`的值，你将得到一个空字符串。您可以像往常一样随时为其余元素提供值:

```go
langs[1] = "Rust"
langs[2] = "Scala"

```

当使用数组文字声明和初始化数组时，可以在多行语句中提供初始化表达式，如下所示:

```go
y := [5]int {
   5,
  10,
  15,
  20,
  25,
}

```

在多行语句中初始化数组元素时，必须在所有元素后提供逗号，包括最后一个元素。当您修改代码时，这使可用性成为可能。因为每个元素后面都有一个逗号，所以您可以轻松地删除或注释元素初始化，或者在任何位置添加新元素，包括最后一个位置。

当您声明数组时，您总是指定数组的长度，但是当您声明和初始化数组时，您可以使用表达式`…`来代替指定长度，如下所示:

```go
z := [...] { 5,10,15,20,25}

```

这里，数组的长度由初始化表达式中提供的元素数量决定。

#### 遍历数组

因为数组是一种集合类型，所以您可能希望迭代数组的元素。下面是使用普通的`for`循环迭代数组元素的代码块:

```go
langs := [4]string{"Go", "Rust", "Scala","Julia"}
for i := 0; i < len(langs); i++ {
        fmt.Printf("langs[%d]:%s \n", i, langs[i])
    }

```

在这里，我们迭代`langs`数组的元素，并通过指定索引值简单地打印每个元素的值。`len`函数获取集合类型的值的长度。

Note

Go 语言只有一个循环结构，那就是`for`循环。与许多其他语言不同，Go 不支持`while`循环结构。如果你想要一个类似于`while`的循环结构，你可以使用`for`循环(例如`for i< 1000{}`)。

Go 有一个`range`构造，可以让您迭代各种集合类型中的元素。Go 程序员通常使用`range`构造来迭代数据结构的元素，比如数组、切片和映射。下面是迭代数组元素的代码块:

```go
for k, v := range langs {
        fmt.Printf("langs[%d]:%s \n", k, v)
 }

```

数组上的`range`构造为集合中的每个元素提供了索引和值。在我们的示例代码块中，变量`k`获取索引，变量`v`获取元素的值。如果您不想使用您在左侧声明的任何变量的值，您可以通过使用空白标识符(`_`)来忽略它，如下所示:

```go
for _, v := range langs {
        fmt.Printf(v)
   }

```

在这个`range`块中，使用了元素的值，但没有使用索引，因此使用一个空白标识符(`_`)来代替索引变量，以避免编译错误。如果一个变量被声明但从未被使用过，Go 编译器会显示一个错误。

#### 示例程序

清单 [2-9](#Par86) 显示了探索数组类型的示例程序。

```go
package main

import (
    "fmt"
)

func main() {
    // Declare arrays
    var x [5]int
    // Assign values at specific index
    x[0] = 5
    x[4] = 25
    fmt.Println("Value of x:", x)

    x[1] = 10
    x[2] = 15
    x[3] = 20
    fmt.Println("Value of x:", x)

    // Declare and initialize array with array literal
    y := [5]int{10, 20, 30, 40, 50}
    fmt.Println("Value of y:", y)

    // Array literal with ...
    z := [...]int{10, 20, 30, 40, 50}
    fmt.Println("Value of z:", z)
    fmt.Println("Length of z:", len(z))

    // Initialize values at specific index with array literal
    langs := [4]string{0: "Go", 3: "Julia"}
    fmt.Println("Value of langs:", langs)
    // Assign values to remaining positions
    langs[1] = "Rust"
    langs[2] = "Scala"

    // Iterate over the elements of array
    fmt.Println("Value of langs:", langs)
    fmt.Println("\nIterate over arrays\n")
    for i := 0; i < len(langs); i++ {
        fmt.Printf("langs[%d]:%s \n", i, langs[i])
    }
    fmt.Println("\n")

    // Iterate over the elements of array using range
    for k, v := range langs {
        fmt.Printf("langs[%d]:%s \n", k, v)
    }
}

Listing 2-9.Example Program on Arrays

```

运行该程序时，您应该会看到以下输出:

```go
Value of x: [5 0 0 0 25]
Value of x: [5 10 15 20 25]
Value of y: [10 20 30 40 50]
Value of z: [10 20 30 40 50]
Length of z: 5
Value of langs: [Go   Julia]
Value of langs: [Go Rust Scala Julia]

Iterate over arrays

langs[0]:Go
langs[1]:Rust
langs[2]:Scala
langs[3]:Julia

langs[0]:Go
langs[1]:Rust
langs[2]:Scala
langs[3]:Julia

```

## 2-3.使用切片处理动态数组

### 问题

您希望将数据集合存储到动态数组中，因为您在声明数组时不知道它的大小。

### 解决办法

Go 的切片类型允许您存储单一类型元素的动态长度。

### 它是如何工作的

当您声明用于存储元素集合的数据结构时，您可能不知道它的大小。例如，假设您想从数据库表或 NoSQL 集合中查询数据，并将数据放入一个变量中。在这种情况下，您不能通过提供大小来声明数组，因为数组的大小会根据数据库表中包含的数据随时变化。切片是建立在 Go 的数组类型之上的数据结构，它允许您存储单一类型元素的动态长度。在您的 Go 应用程序中，数组的使用可能是有限的，您可能会经常使用切片，因为它们提供了一个灵活的、可扩展的数据结构。

切片数据结构具有长度和容量。长度是切片引用的元素数量。容量是切片中分配有空间的元素数量。切片长度不能超过容量值，因为这是可以达到的最大值长度。切片的长度和容量可以分别通过使用 len 和 cap 函数来确定。由于存储片的动态特性，当存储片增长时，存储片的长度和容量可以随时变化。

#### 声明零切片

声明一个`slice`类似于声明一个数组，但是当声明切片时，不需要指定大小，因为它是一个动态数组。下面是声明一个`nil`片的代码块:

```go
var x []int

```

切片`x`被声明为整数的`nil`切片。此时，切片的长度和容量为零。虽然`x`的长度现在为零，但是您可以在以后修改长度并初始化值，因为片是动态数组。Go 提供了一个函数`append`,该函数可用于在以后放大任何片(nil 或非 nil)。

#### 使用 make 函数初始化切片

在赋值之前，必须初始化切片。在前面的声明中，片`x`被声明，但是它没有被初始化，所以如果你试图给它赋值，这将导致运行时错误。Go 内置的`make`函数用于初始化切片。当使用`make`函数声明切片时，`length`和`capacity`作为参数提供。

下面是使用指定了`length`和`capacity`的`make`函数创建切片的代码块:

```go
y:= make ([]int, 3,5)

```

使用`make`函数，用为 3 的`length`和为 5 的`capacity`声明并初始化一个片`y`。当`make`函数的参数中省略了`capacity`参数时，`capacity`的值默认为`length`的指定值。

```go
y:= make ([]int, 3)  

```

用 3 的`length`和 3 的`capacity`声明并初始化片 y。因为没有提供`capacity`的值，所以默认为`length`的值。

可以像数组一样给片`y`赋值:

```go
y[0] = 10
y[1] = 20
y[2] = 30

```

#### 使用切片文字创建切片

除了使用`make`函数创建切片之外，还可以使用切片文字创建切片，这类似于数组文字。下面是使用切片文字创建切片的代码块:

```go
z:= []int {10,20,30}

```

用为 3 的`length`和为 3 的`capacity`声明并初始化切片`z`。初始化这些值时，可以为特定的索引提供值，如下所示:

```go
z:= []int {0:10, 2:30}

```

创建一个切片`z`，并用 3 的`length`和 3 的`capacity`进行初始化。当您使用这种方法创建切片时，`length`由您指定的最高索引值决定，因此您也可以通过简单地提供最高索引来创建切片，如下所示:

```go
z:= []int {2:0}

```

通过初始化索引 2 的零值来创建切片`z`，因此该切片的`capacity`和`length`将是 3。

通过使用切片文字，您还可以创建一个空切片:

```go
z:= []int{}

```

切片`z`是用零个值元素创建的。当您希望从函数中返回空集合时，空切片非常有用。假设您提供了一个从数据库表中查询数据的函数，并通过填充表中的数据返回一个切片。如果表格不包含任何数据，您可以在这里返回一个空的切片。请注意，零切片和空切片是不同的。如果`z`是一个空片，代码表达式`z == nil`返回`false`，如果是一个零片，表达式`z == nil`返回`true`。

#### 使用复制和附加功能放大切片

因为切片是动态数组，所以可以随时放大它们。当您想要增加切片的`capacity`时，一种方法是创建一个新的更大的切片，并将原始切片的元素复制到新创建的切片中。Go 内置的`copy`函数用于将数据从一个片复制到另一个片。清单 [2-10](#Par115) 显示了一个使用`copy`函数增加切片大小的示例程序。

```go
package main

import (
    "fmt"
)

func main() {
    x := []int{10, 20, 30}
    fmt.Printf("[Slice:x] Length is %d Capacity is %d\n", len(x), cap(x))
    // Create a bigger slice
    y := make([]int, 5, 10)
    copy(y, x)
    fmt.Printf("[Slice:y] Length is %d Capacity is %d\n", len(y), cap(y))
    fmt.Println("Slice y after copying:", y)
    y[3] = 40
    y[4] = 50
    fmt.Println("Slice y after adding elements:", y)
}

Listing 2-10.Program to Enlarge a Slice Using the copy Function

```

运行该程序时，您应该会看到以下输出:

```go
 [Slice:x] Length is 3 Capacity is 3
[Slice:y] Length is 5 Capacity is 10
Slice y after copying: [10 20 30 0 0]
Slice y after adding elements: [10 20 30 40 50]

```

创建一个切片`x`，其`length`为 3，`capacity`为 3。为了增加`capacity`并向切片添加更多元素，创建了一个新的切片`y`，其`length`为 5，`capacity`为 10。然后，`copy`函数将数据从片`x`复制到目标片`y`。

您还可以通过使用 Go 内置的`append`函数将数据追加到现有切片的末尾来放大切片。如有必要，`append`功能会自动增加`slice`的大小，并返回更新后的`slice`和新添加的数据。清单 [2-11](#Par120) 显示了一个使用`append`函数增加切片的示例程序。

```go
package main

import (
    "fmt"
)

func main() {
    x := make([]int, 2, 5)
    x[0] = 10
    x[1] = 20recipes for arrays
    fmt.Println("Slice x:", x)
    fmt.Printf("Length is %d Capacity is %d\n", len(x), cap(x))
    // Create a bigger slice
    x = append(x, 30, 40, 50)
    fmt.Println("Slice x after appending data:", x)

    fmt.Printf("Length is %d Capacity is %d\n", len(x), cap(x))

    x = append(x, 60, 70, 80)
    fmt.Println("Slice x after appending data for the second time:", x)
    fmt.Printf("Length is %d Capacity is %d\n", len(x), cap(x))

}

Listing 2-11.Program That Enlarges a Slice Using the append Function

```

运行该程序时，您应该会看到以下输出:

```go
Slice x: [10 20]
Length is 2 Capacity is 5
Slice x after appending data: [10 20 30 40 50]
Length is 5 Capacity is 5
Slice x after appending data for the second time: [10 20 30 40 50 60 70 80]
Length is 8 Capacity is 10

```

创建一个切片`x`，其中`length`为 2，`capacity`为 5。然后，三个以上的数据元素被附加到`slice`。这次`length`和`capacity`都是 5。然后，将另外三个数据元素追加到切片中。这次你试图将切片的`length`增加到 8，但是切片的`capacity`是 5。如有必要,`append`功能可以自动增大`capacity`。这里增加到 10。

您可以将数据附加到一个 nil 片上，在那里它会分配一个新的底层数组，如清单 [2-12](#Par125) 所示。

```go
package main

import "fmt"

func main() {
     // Declare a nil slice
     var x []int
     fmt.Println(x, len(x), cap(x))
     x = append(x, 10, 20, 30)
     fmt.Println("Slice x after appending data:", x)
}

Listing 2-12.Appending Data to a Nil Slice

```

运行该程序时，您应该会看到以下输出:

```go
[] 0 0
Slice x after appending data: [10 20 30]

```

#### 遍历切片

迭代切片元素的惯用方法是使用`range`构造。清单 [2-13](#Par129) 展示了一个迭代切片元素的示例程序。

```go
package main

import (
    "fmt"
)

func main() {
    x := []int{10, 20, 30, 40, 50}
    for k, v := range x {
        fmt.Printf("x[%d]: %d\n", k, v)
    }
}

Listing 2-13.Program to Iterate Over the Elements of a Slice

```

运行该程序时，您应该会看到以下输出:

```go
x[0]: 10
x[1]: 20
x[2]: 30
x[3]: 40
x[4]: 50

```

片上的`range`构造为集合中的每个元素提供了索引和值。在我们的示例程序中，变量`k`获取索引，变量`v`获取数据元素的值。

## 2-4.使用映射持久化键/值对

### 问题

您希望将键/值对的集合保存到类似于哈希表的集合类型中。

### 解决办法

Go 的 map 类型允许您将键/值对的集合存储到类似于散列表的结构中。

### 它是如何工作的

Go 的 map 类型是一种数据结构，它提供了哈希表的实现(在 Java 中称为 HashMap)。哈希表实现允许您将数据元素作为键和值来保存。哈希表提供了对数据元素的快速查找，因为您可以通过提供键来轻松地检索值。

#### 声明和初始化地图

以下是地图类型的定义:

```go
map[KeyType]ValueType

```

这里`KeyType`是键的类型，`ValueType`是值的类型。下面是声明地图的代码块:

```go
var chapts  map[int]string

```

用`int`作为键的类型和`string`作为值的类型来声明映射`chapts`。此时，映射图`chapts`的值是`nil`，因为映射图没有初始化。试图将值写入`nil`映射将导致运行时错误。在向映射写入值之前，需要初始化映射。内置的`make`函数用于初始化地图，如下图所示:

```go
chapts = make(map[int] string)

```

使用`make`函数初始化地图`chapts`。让我们向地图添加一些数据值:

```go
chapts[1]="Beginning Go"
chapts[2]="Go Fundamentals"
chapts[3]="Structs and Interfaces"

```

需要注意的是，不能向映射中添加重复的键。

您还可以使用映射文字来声明和初始化映射，如下所示:

```go
langs := map[string]string{
             "EL": "Greek",
             "EN": "English",
             "ES": "Spanish",
             "FR": "French",
             "HI": "Hindi",
      }

```

映射`langs`是用`string`作为键和值的类型来声明的，值是使用映射文字来初始化的。

#### 使用地图

映射提供了对数据结构中数据元素的快速查找。通过提供如下所示的键，您可以轻松地检索元素的值:

```go
lan, ok := langs["EN"]

```

通过提供一个键在 map 上执行的查找返回两个值:元素的值和一个指示查找是否成功的布尔值。变量`lan`获取键`"EN"`的元素值，变量`ok`获取布尔值:`true`如果键`"EN"`有值，而`false`如果键不存在。Go 为编写可用于编写查找语句的`if`语句提供了方便的语法:

```go
if lan, ok := langs["EN"]; ok {
       fmt.Println(lan)
}

```

当把一个`if`语句写成单行上的多个语句时，语句之间用分号(；)并且最后一个表达式应该有一个布尔值。

要从地图中移除项目，请通过提供键来使用内置函数`delete`。`delete`函数从 map 中删除给定键的元素，并且不返回任何内容。下面的代码块为键`"EL"`从`langs`映射中删除了一个元素。

```go
delete(langs,"EL")

```

这将删除键`"EL"`的一个元素。如果指定的键不存在，它不会做任何事情。

像其他集合类型一样，`range`构造通常用于迭代 map 的元素。清单 [2-14](#Par157) 展示了一个演示地图上各种操作的示例程序。

```go
package main

import (
    "fmt"
)

func main() {
    // Declares a nil map
    var chapts map[int]string

    // Initialize map with make function
    chapts = make(map[int]string)

    // Add data as key/value pairs
    chapts[1] = "Beginning Go"
    chapts[2] = "Go Fundamentals"
    chapts[3] = "Structs and Interfaces"

    // Iterate over the elements of map using range
    for k, v := range chapts {
        fmt.Printf("Key: %d Value: %s\n", k, v)
    }

    // Declare and initialize map using map literal
    langs := map[string]string{
        "EL": "Greek",
        "EN": "English",
        "ES": "Spanish",
        "FR": "French",
        "HI": "Hindi",
    }

    // Delete an element
    delete(langs, "EL")

    // Lookout an element with key
    if lan, ok := langs["EL"]; ok {
        fmt.Println(lan)
    } else {
        fmt.Println("\nKey doesn't exist")
    }
}

Listing 2-14.Various operations on maps

```

您应该会看到类似如下的输出:

```go
Key: 3 Value: Structs and Interfaces
Key: 1 Value: Beginning Go
Key: 2 Value: Go Fundamentals

Key doesn't exist

```

#### 地图的迭代顺序

当您使用`range`构造对地图进行迭代时，迭代顺序并未指定，因此不能保证一次迭代得到相同的结果，因为 Go 会随机化地图迭代顺序。如果您想要以特定的顺序迭代地图，您必须维护一个数据结构来指定该顺序。清单 [2-15](#Par161) 显示了一个示例程序，它遍历一个带有顺序的地图。为了指定顺序，这个例子维护了一个片来存储映射的排序键。

```go
package main

import (
       "fmt"
       "sort"
)

func main() {
       // Initialize map with make function
       chapts := make(map[int]string)

       // Add data as key/value pairs
       chapts[1] = "Beginning Go"
       chapts[2] = "Go Fundamentals"
       chapts[3] = "Structs and Interfaces"

       // Slice for specifying the order of the map
       var keys []int
       // Appending keys of the map
       for k := range chapts {
              keys = append(keys, k)
       }
       // Ints sorts a slice of ints in increasing order.
       sort.Ints(keys)
       // Iterate over the map with an order
       for _, k := range keys {
              fmt.Println("Key:", k, "Value:", chapts[k])
       }
}

Listing 2-15.Iterate over a Map With an Order

```

您应该会看到以下输出:

```go
Key: 1 Value: Structs and Interfaces
Key: 2 Value: Go Fundamentals
Key: 3 Value: Beginning Go

```

因为您指定了顺序，所以所有迭代的输出顺序都是相同的。

## 2-5.在函数中编写清理代码

### 问题

您希望在函数中编写清理逻辑，以便在周围的函数返回后执行清理操作。

### 解决办法

Go 提供了一个`defer`语句，允许你在函数中编写清理逻辑。

### 它是如何工作的

函数中的`defer`语句将函数调用或 case 语句推送到保存的调用列表中。您可以在一个函数中添加多个`defer`语句。这些来自保存列表的延迟函数调用在周围函数返回后执行。`defer`语句通常用于在函数内部编写清理逻辑，以释放您在其中创建的资源。例如，假设您在一个函数中打开了一个数据库连接对象，您可以在函数返回后安排关闭该连接对象以清理该连接对象的资源。`defer`语句通常用于`close`、`disconnect`和`unlock`语句，与`open`、`connect`或`lock`语句相对。`defer`语句确保函数调用的延迟列表在所有情况下都被调用，即使发生异常也是如此。

列表 [2-16](#Par169) 显示了一个代码块，该代码块使用`defer`语句来关闭一个为读取而打开的文件对象。

```go
import (
    "io/ioutil"
    "os"
)

func ReadFile(filename string) ([]byte, error) {
    f, err := os.Open(filename)
    if err != nil {
        return nil, err
    }
    defer f.Close()

    return ioutil.ReadAll(f)
}

Listing 2-16.
Defer Statement Used to Close a File Object

```

我们打开一个文件对象`f`来读取它的内容。为了确保对象`f`正在释放它的资源，我们将代码语句`f.Close()`添加到函数调用的延迟列表中。释放资源的`defer`语句通常是在资源被创建且没有任何错误之后编写的。我们把`defer f.Close()`写在对象`f has been successfully created.`之后

```go
    f, err := os.Open(filename)
    if err != nil {
        return nil, err
    }
    defer f.Close()

```

使用`defer`编写清理逻辑类似于在 C#和 Java 等其他编程语言中使用`finally`块。在`try/catch/finally`块中，您在`finally`块中为已经在`try`块中创建的资源编写清理逻辑。Go 的`defer`比传统编程语言的`finally`块更强大。例如，结合`defer`和`recover`语句，您可以从一个混乱的函数中重新获得控制。我们将在本章的下一节介绍`panic`和`recover`。

## 2-6.使用 Panic 停止控制的执行流

### 问题

当您的程序出现严重错误时，您希望停止函数中的执行控制流，并开始恐慌。

### 解决办法

Go 提供了一个内置的`panic`函数，停止一个程序的正常执行，开始死机。

### 它是如何工作的

当 Go 运行时在执行过程中检测到任何未处理的错误时，它会死机并停止执行。因此，所有运行时错误都会导致程序崩溃。通过显式调用内置的`panic`函数，可以创建同样的情况；它停止正常执行并开始死机。在继续执行几乎不可能的情况下，通常会调用`panic`函数。例如，如果您试图连接到一个数据库，但是无法连接，那么继续执行程序就没有任何意义，因为您的应用程序依赖于数据库。在这里你可以调用`panic`函数来停止正常执行并使你的程序死机。`panic`函数接受任何类型的值作为参数。当函数内部发生异常时，它会停止函数的正常执行，执行该函数中所有延迟的函数调用，然后调用方函数会得到一个异常函数。在停止执行之前，执行所有的延迟函数是很重要的。Go 运行时确保在所有情况下都执行 defer 语句，包括紧急情况。

清单 [2-17](#Par177) 显示了当试图打开一个文件导致错误时调用`panic`的代码块；它通过提供一个错误对象作为参数来调用`panic`。

```go
import (
    "io/ioutil"
    "os"
)

func ReadFile(filename string) ([]byte, error) {
    f, err := os.Open(filename)
    if err != nil {
        panic (err)  // calls panic

    }
    defer f.Close()

    return ioutil.ReadAll(f)
}

Listing 2-17.Using panic to Panic a Function

```

函数`ReadFile`试图打开一个文件来读取其内容。如果`Open`函数出错，就会调用`panic`函数来启动一个应急函数。当你编写真实世界的应用程序时，你很少会调用`panic`函数；您的目标应该是处理所有错误以避免出现恐慌情况，记录错误消息，并向最终用户显示正确的错误消息。

## 2-7.使用 Recover 恢复死机功能

### 问题

你想重新获得对恐慌功能的控制。

### 解决办法

Go 提供了一个内置的`recover`函数，让你重新获得对一个死机函数的控制；因此，它仅用于延迟函数。在延迟函数中使用了`recover`函数，以恢复死机函数的正常执行。

### 它是如何工作的

当函数死机时，该函数中所有延迟的函数调用都会在正常执行停止之前执行。在这里，对延迟函数中的`recover`的调用获得了赋予`panic`的值，并重新获得了对正常执行的控制。简而言之，即使在紧急情况下，您也可以使用`recover`恢复正常执行。

清单 [2-18](#Par183) 展示了一个使用`recover`进行紧急恢复的例子。

```go
package main

import (
    "fmt"
)

func panicRecover() {

    defer fmt.Println("Deferred call - 1")
    defer func() {
        fmt.Println("Deferred call - 2")
        if e := recover(); e != nil {
            // e is the value passed to panic()
            fmt.Println("Recover with: ", e)
        }
    }()
    panic("Just panicking for the sake of example")
    fmt.Println("This will never be called")
}

func main() {
    fmt.Println("Starting to panic")
    panicRecover()
    fmt.Println("Program regains control after the panic recovery")
}

Listing 2-18.Example that demonstrates recover

```

这个示例程序演示了如何使用`recover`函数恢复一个死机函数的正常执行。在函数`panicRecover`中，增加了两个延迟函数。在这两个延迟的函数调用中，第二个是匿名函数，在这个函数中，调用`recover`来恢复执行，即使在出现紧急情况之后。理解您可以在函数中添加任意数量的延迟函数调用是很重要的。延迟函数的执行顺序是最后添加的，按顺序是第一个。例如，`panic`通过提供一个字符串值作为参数来显式调用。这个值可以通过调用`recover`函数来检索。当调用`panic`函数时，控制流向延迟函数，其中从第二个延迟函数调用`recover`函数(当执行延迟函数调用时，这将首先被调用)。当调用`recover`时，它接收给`panic`的值并恢复正常执行，程序正常运行。

运行该程序时，您应该会看到以下输出:

```go
Starting to panic
Deferred call - 2
Recover with:  Just panicking for the sake of example
Deferred call - 1
Program regains control after the panic recovery

```

该结果还说明了延迟函数的执行顺序。最后添加的延迟函数在第一次延迟函数调用之前执行。

## 2-8.执行错误处理

### 问题

您希望在 Go 应用程序中执行错误处理。

### 解决办法

Go 提供了一个内置的`error`类型，用于通知函数中的错误。Go 函数可以返回多个值。这可以通过返回一个`error`值和其他返回值来实现函数中的异常处理，因此调用函数可以检查函数是否提供了一个错误值。

### 它是如何工作的

与许多其他编程语言不同，Go 不提供`try/catch`块来处理异常。取而代之，您可以使用内置的`error`类型向调用者函数发出异常信号。如果你能研究一下标准库包的功能，你会对如何处理 Go 中的异常有更好的理解。标准库包的大多数函数返回多个值，包括一个`error`值。在函数中返回一个`error`值的惯用方法是在`return`语句中提供的其他值之后提供`error`值。因此，在`return`语句中，`error`值将是最后一个参数。在清单 [2-14](#Par157) 中，您调用了标准库包`os`的`Open`函数来打开一个文件对象。

```go
f, err := os.Open(filename)
    if err != nil {
        return nil, err
    }

```

`Open`函数返回两个值:一个文件对象和一个`error`值。检查返回的`error`值，以确定打开文件时是否出现任何异常。如果`error`值返回一个非空值，这意味着发生了一个错误。

下面是`os`包中`Open`函数的源代码:

```go
// Open opens the named file for reading.  If successful, methods on
// the returned file can be used for reading; the associated file
// descriptor has mode O_RDONLY.
// If there is an error, it will be of type *PathError.
func Open(name string) (*File, error) {
       return OpenFile(name, O_RDONLY, 0)
}

```

正如标准库包通过返回一个`error`值来使用异常处理一样，您可以在 Go 代码中采用相同的方法。清单 [2-19](#Par196) 显示了一个返回`error`值的示例函数。

```go
func Login(user User) (User, error) {
              var u User
       err = C.Find(bson.M{"email": user.Email}).One(u)
       if err != nil {
              return nil, err
       }
       err = bcrypt.CompareHashAndPassword(u.HashPassword, []byte(user.Password))
       if err != nil {
              return nil, err
       }
       return u, nil
}
Listing 2-19.Example Function That Provides error Value

```

`Login`函数返回两个值，包括一个`error`值。下面是调用`Login`函数并验证该函数是否返回任何非空值`error`的代码块:

```go
if user, err := repo.Login(loginUser); err != nil {
    fmt.Println(err)
}
// Implementation here if error is nil

```

在这个代码块中，调用者函数检查返回的`error`值；如果`error`值返回一个非空值，则表明该函数返回一个错误。如果返回的`error`值为`nil,`，则表明函数调用成功，没有任何错误。当`fmt.Println`函数获得一个`error`值作为参数时，它通过调用其`Error() string`方法格式化`error`值。`error`值的`Error`方法返回字符串形式的错误信息。调用函数可以与`Error`方法一起使用，以字符串形式获取错误消息。

```go
Message :=  err.Error()

```

当您返回`error`值时，您可以向调用函数提供描述性的`error`值。通过使用`errors`包的`New`功能，您可以提供描述性的`error`值，如下所示:

```go
func Login(user User) (User, error) {
        var u User
       err = C.Find(bson.M{"email": user.Email}).One(u)
       if err != nil {
              return nil, errors.New("Email doesn't exists")
       }
       // Validate password
       err = bcrypt.CompareHashAndPassword(u.HashPassword, []byte(user.Password))
       if err != nil {
              return nil, errors.New("Invalid password")
       }
       return u, nil
}

```

`errors.New`函数返回一个`error`值，用于向调用函数提供描述性的`error`值。`fmt`包的`Errorf`函数允许您使用`fmt`包的格式化功能来创建描述性的`error`值，如下所示:

```go
func Login(user User) (User, error) {
        var u User
       err = C.Find(bson.M{"email": user.Email}).One(u)
       if err != nil {
               errObj:= fmt.Errorf("User %s doesn't exists. Error:%s, user.Email, err.Error())
              return nil, errObj
       }
       // Validate password
       err = bcrypt.CompareHashAndPassword(u.HashPassword, []byte(user.Password))
       if err != nil {
              errObj:= fmt.Errorf("Invalid password for the user:%s. Error:%s, user.Email, err.Error())
              return nil, errObj
       }
       return u, nil
}

```

前面的代码块使用`fmt.Errorf`函数来使用`fmt`包的格式化特性来创建描述性的`error`值。

Go 中的函数是一段可重用的代码，它将一系列代码语句组织成一个单元。关键字`func`用于声明函数。如果函数的名称以大写字母开头，那么这些函数会被导出到其他包中。Go 函数的一个独特特性是它们可以返回多个值。

Go 提供了三种类型的数据结构来处理数据集合:数组、切片和映射。数组是固定长度的类型，包含单一类型的元素序列。通过指定长度和类型来声明数组。切片类似于数组，但是它的大小可以随时变化，所以您不必指定切片的长度。使用内置的`make`函数或切片文字初始化切片。切片可以使用两个内置函数进行修改:`append`和`copy`。映射是哈希表的一种实现，它提供了一个无序的键/值对集合。使用内置的`make`函数或使用地图文字来初始化地图。

Go 提供了`defer,`，可以用来在函数中写清理逻辑。一个`defer`语句将一个函数调用推送到一个保存的列表上，该列表在周围的函数返回后执行。`Panic`是一个内置函数，可以让你停止正常执行，并开始一个函数的死机。`Recover`是一个内置函数，可恢复对恐慌功能的控制。`Recover`仅用于延迟函数内部。

Go 使用一种不同且独特的方法在 Go 代码中实现异常处理。因为 Go 函数可以返回多个值，所以`return`语句提供了一个`error`值，以及其他返回值。这样，调用函数可以检查返回的`error`值，以确定是否有错误。