# 三、套接字级编程

世界上有很多种网络。这些网络从非常古老的网络(如串行链路)到由铜线和光纤构成的广域网，再到各种类型的无线网络，既用于计算机，也用于电信设备(如电话)。这些网络显然在物理链路层有所不同，但在许多情况下，它们在 OSI 堆栈的更高层也有所不同。

这些年来，IP 和 TCP/UDP 的“互联网堆栈”已经趋同。例如，蓝牙定义了物理层和协议层，但是在物理层和协议层之上是 IP 栈，因此相同的互联网编程技术可以在许多蓝牙设备上使用。类似地，开发物联网(IoT)无线技术，如 LoRaWAN 和 6LoWPAN，也包括 IP 堆栈。

IP 提供 OSI 堆栈的第 3 层网络，而 TCP 和 UDP 处理第 4 层。即使在互联网世界中，这些也不是最终的结论:SCTP(流控制传输协议)来自电信世界，挑战 TCP 和 UDP，而在星际空间中提供互联网服务需要新的、正在开发的协议，如 DTN(延迟容忍网络)。然而，IP、TCP 和 UDP 作为主要的网络技术在现在以及至少在未来相当长的一段时间内占据主导地位。Go 完全支持这种风格的编程

本章展示了如何使用 Go 进行 TCP 和 UDP 编程，以及如何为其他协议使用原始套接字。

## TCP/IP 协议栈

OSI 模型是使用一个委员会过程设计的，在这个过程中，标准被建立，然后被实施。OSI 标准的一些部分是模糊的，一些部分不容易实现，一些部分还没有实现。

TCP/IP 协议是通过长期运行的 DARPA 项目设计的。这是通过 RFC(征求意见)后的实施来实现的。TCP/IP 是主要的 UNIX 网络协议。TCP/IP 代表传输控制协议/互联网协议。

TCP/IP 协议栈比 OSI 协议栈短，如图 [3-1](#Fig1) 所示。

![A436770_1_En_3_Fig1_HTML.gif](img/A436770_1_En_3_Fig1_HTML.gif)

图 3-1。

TCP/IP stack versus the OSI

TCP 是面向连接的协议，而 UDP(用户数据报协议)是无连接的协议。

### IP 数据报

IP 层提供了一个无连接和不可靠的传输系统。它认为每个数据报独立于其他数据报。数据报之间的任何关联必须由较高层提供。

IP 层提供包含其自身报头的校验和。报头包括源地址和目的地址。

IP 层处理通过互联网的路由。它还负责将较大的数据报分解成较小的数据报进行传输，并在另一端重新组装。

### 用户数据报协议(User Datagram Protocol)

UDP 也是无连接和不可靠的。它添加到 IP 中的是数据报内容和端口号的校验和。这些用来给出一个客户机-服务器模型，稍后你会看到。

### 三氯苯酚

TCP 提供逻辑，在 IP 之上给出可靠的面向连接的协议。它提供了两个进程可以用来通信的虚电路。它还使用端口号来识别主机上的服务。

## 互联网地址

为了使用服务，您必须能够找到它。互联网对计算机等设备使用地址方案，以便对它们进行定位。这种编址方案最初是在只有少数几台相连的计算机时设计的，使用 32 位无符号整数，非常宽松地允许多达 2^32 地址。这些就是所谓的 IPv4 地址。近年来，连接的(或者至少是可直接寻址的)设备的数量有超过这个数字的危险，并且正在逐步过渡到 IPv6。这种转变是不完整的，例如在谷歌( [`https://www.google.com/intl/en/ipv6/statistics.html`](https://www.google.com/intl/en/ipv6/statistics.html) )的图表中显示的。遗憾的是，在我看来，很少有澳大利亚 IP 提供商支持 IPv6。

### IPv4 地址

该地址是一个 32 位整数，给出了 IP 地址。该地址可解析为单个设备上的网络接口卡。地址通常写成十进制的四个字节，中间用点`.`隔开，如`127.0.0.1`或`66.102.11.104`。

任何设备的 IP 地址通常由两部分组成:设备所在网络的地址，以及设备在该网络中的地址。从前，网络地址和内部地址之间的划分很简单，是基于 IP 地址中使用的字节。

*   在 A 类网络中，第一个字节标识网络，后三个字节标识设备。只有 128 个 A 类网络，由互联网领域的早期参与者拥有，如 IBM、通用电气公司和麻省理工学院 [<sup>1</sup>](#Fn1) 。
*   B 类网络使用前两个字节标识网络，后两个字节标识子网内的设备。这允许一个子网上最多有 2^16 (65，536)台设备。
*   C 类网络使用前三个字节标识网络，最后一个字节标识网络中的设备。这允许多达 2^8(实际上是 254，而不是 256，因为底部和顶部地址是保留的)设备。

如果你想要，比如说，一个网络上有 400 台计算机，这个方案就不太管用。254 太小，而 65，536 (-2)太大。用二进制算术术语来说，你要 512 (-2)左右。这可以通过使用 23 位网络地址和 9 位设备地址来实现。同样，如果您想要多达 1024 (-2)个设备，您可以使用 22 位网络地址和 10 位设备地址。

给定一个设备的 IP 地址，并且知道网络地址使用了多少位 N，给出了一个相对简单的过程来提取网络地址和该网络内的设备地址。形成一个“网络掩码”，它是一个 32 位二进制数，前 N 位全为 1，其余全为 0。例如，如果网络地址使用 16 位，掩码为`11111111111111110000000000000000`。用二进制有点不方便，所以一般用十进制字节。16 位网络地址的网络掩码为`255.255.0.0`，24 位网络地址的网络掩码为`255.255.255.0`，23 位网络地址的网络掩码为`255.255.254.0`，22 位网络地址的网络掩码为`255.255.252.0`。

然后查找设备的网络，按位`AND`使用网络掩码查找其 IP 地址，而子网内的设备地址则按位`AND`使用 IP 地址查找掩码的补码。例如，IP 地址`192.168.1.3`的二进制值是`11000000101010000000000100000011`(使用 [IP 地址子网掩码计算器](https://www.fukatani.org/%E2%88%BChi-lo/cgi-bin/fk-ip_calc.cgi))。如果使用 16 位网络掩码，网络为`1100000010101000 0000000000000000`(或`192.168.0.0`)，而设备地址为`0000000000000000  0000000100000011`(或`0.0.1.3`)。

### IPv6 地址

互联网的发展远远超出了最初的预期。最初慷慨的 32 位寻址方案即将耗尽。有一些令人不快的解决方法，如 NAT(网络地址转换)寻址，但最终我们将不得不切换到更宽的地址空间。IPv6 使用 128 位地址。用偶数字节来表示这样的地址变得很麻烦，所以使用十六进制数字，分成四个数字，并用冒号`:`隔开。典型的地址可能是`FE80:CD00:0000:0CDE:1257:0000:211E:729C`。

这些地址不好记！DNS 将变得更加重要。减少一些地址是有技巧的，比如前导零和重复数字。比如“localhost”就是`0:0:0:0:0:0:0:1`，可以简称为`::1`。

每个地址分为三部分:第一部分是用于互联网路由的网络地址，是地址的前 64 位。下一部分是 16 位网络掩码。这用于将网络划分为子网。它可以给出从一个子网(全 0)到 65，535 个子网(全 1)的任何值。最后一部分是器件组件，48 位。上述地址将是网络的`FE80:CD00:0000:0CDE`、子网的`1257`和设备的`0000:211E:729C`。

## IP 地址类型

最后，我们可以开始使用一些 Go 语言网络包。包`net`定义了许多类型、功能和在 Go 网络编程中的使用方法。类型`IP`被定义为一个字节数组:

```go
type IP []byte

```

有几个函数可以操作类型为`IP`的变量，但是在实践中你可能只使用其中的一部分。例如，函数`ParseIP(String)`将采用带点的 IPv4 地址或冒号的 IPv6 地址，而 IP 方法`String()`将返回一个字符串。请注意，您可能无法回到开始时的状态:字符串形式的`0:0:0:0:0:0:0:1`是`::1`。

说明这个过程的一个程序是`IP.go`:

```go
/* IP
 */

package main

import (
        "fmt"
        "net"
        "os"
)

func main() {
        if len(os.Args) != 2 {
                fmt.Fprintf(os.Stderr, "Usage: %s ip-addr\n", os.Args[0])
                os.Exit(1)
        }
        name := os.Args[1]

        addr := net.ParseIP(name)
        if addr == nil {
                fmt.Println("Invalid address")
        } else {
                fmt.Println("The address is ", addr.String())
        }
        os.Exit(0)
}

```

例如，这可以如下运行:

```go
go run IP.go  127.0.0.1

```

以下是回应:

```go
The address is 127.0.0.1

```

或者它可以运行为:

```go
go run IP.go 0:0:0:0:0:0:0:1

```

有了这样的回应:

```go
The address is ::1

```

### IPMask 类型

IP 地址通常分为网络地址、子网和设备部分。网络地址和子网构成了设备部分的前缀。掩码是一个全二进制的 IP 地址，以匹配前缀长度，后跟全零。

为了处理屏蔽操作，可以使用以下类型:

```go
type IPMask []byte

```

创建网络掩码最简单的函数是使用 CIDR 表示法，即 1 后面跟 0，最大位数为:

```go
func CIDRMask(ones, bits int) IPMask

```

然后，IP 地址的方法可以使用掩码来查找该 IP 地址的网络:

```go
func (ip IP) Mask(mask IPMask) IP

```

下面这个叫做`Mask.go`的程序就是一个例子:

```go
/* Mask
 */

package main

import (
        "fmt"
        "net"
        "os"
        "strconv"
)

func main() {
        if len(os.Args) != 4 {
                fmt.Fprintf(os.Stderr, "Usage: %s dotted-ip-addr ones bits\n", os.Args[0])
                os.Exit(1)
        }
        dotAddr := os.Args[1]
        ones, _ := strconv.Atoi(os.Args[2])
        bits, _ := strconv.Atoi(os.Args[3])

        addr := net.ParseIP(dotAddr)
        if addr == nil {
                fmt.Println("Invalid address")
                os.Exit(1)
        }
        mask := net.CIDRMask(ones, bits)
        network := addr.Mask(mask)
        fmt.Println("Address is ", addr.String(),
                "\nMask length is ", bits,
                "\nLeading ones count is ", ones,
                "\nMask is (hex) ", mask.String(),
                "\nNetwork is ", network.String())
        os.Exit(0)
}

```

这可以编译成`Mask`并运行如下:

```go
Mask <ip-address> <ones> <zeroes>

```

也可以直接运行，如下所示:

```go
go run Mask.go <ip-address> <ones> <zeroes>

```

对于`/24`网络上的 IPv4 地址`103.232.159.187`,我们得到如下结果:

```go
go run Mask.go 103.232.159.187 24 32
Address is  103.232.159.187
Mask length is  32
Leading ones count is  24
Mask is (hex)  ffffff00
Network is  103.232.159.0

```

对于一个 IPv6 地址`fda3:97c:1eb:fff0:5444:903a:33f0:3a6b`，其中网络组件是`fda3:97c:1eb`，子网是`fff0`，设备部分是`5444:903a:33f0:3a6b`，我们得到如下:

```go
go run Mask.go fda3:97c:1eb:fff0:5444:903a:33f0:3a6b 52 128
Address is  fda3:97c:1eb:fff0:5444:903a:33f0:3a6b
Mask length is  128
Leading ones count is  52
Mask is (hex)  fffffffffffff0000000000000000000
Network is  fda3:97c:1eb:f000::      

```

IPv4 网络掩码通常以 4 字节点符号表示，如`255.255.255.0`表示`/24`网络。有一个函数可以从这样一个 4 字节的 IPv4 地址创建一个掩码:

```go
func IPv4Mask(a, b, c, d byte) IPMask

```

此外，有一种 IP 方法可以返回 IPv4 的默认掩码:

```go
func (ip IP) DefaultMask() IPMask

```

注意掩码的字符串形式是一个十六进制数，例如`ffffff00`代表`/24`掩码。

以下名为`IPv4Mask.go`的程序说明了这些:

```go
/* IPv4Mask
 */

package main

import (
        "fmt"
        "net"
        "os"
)

func main() {
        if len(os.Args) != 2 {
                fmt.Fprintf(os.Stderr, "Usage: %s dotted-ip-addr\n", os.Args[0])
                os.Exit(1)
        }
        dotAddr := os.Args[1]

        addr := net.ParseIP(dotAddr)
        if addr == nil {
                fmt.Println("Invalid address")
                os.Exit(1)
        }
        mask := addr.DefaultMask()
        network := addr.Mask(mask)

        ones, bits := mask.Size()
        fmt.Println("Address is ", addr.String(),
                "\nDefault mask length is ", bits,
                "\nLeading ones count is ", ones,
                "\nMask is (hex) ", mask.String(),
                "\nNetwork is ", network.String())
        os.Exit(0)
}

```

例如，运行以下命令:

```go
go run Mask.go 192.168.1.3

```

在我的家庭网络中给出以下结果:

```go
Address is  192.168.1.3
Default mask length is  32
Leading ones count is  24
Mask is (hex)  ffffff00
Network is  192.168.1.0

```

### IPAddr 类型

net 包中的许多其他函数和方法返回一个指向`IPAddr`的指针。这只是一个包含 IP(和 IPv6 地址可能需要的区域)的结构。

```go
type IPAddr {
    IP IP
    Zone string
}

```

这种类型的主要用途是在 IP 主机名上执行 DNS 查找。对于具有多个网络接口的不明确 IPv6 地址，可能需要该区域。

```go
func ResolveIPAddr(net, addr string) (*IPAddr, error)

```

其中`net`是`ip`、`ip4`或`ip6`中的一个。这表现在名为`ResolveIP.go`的节目中:

```go
/* ResolveIP
 */

package main

import (
        "fmt"
        "net"
        "os"
)

func main() {
        if len(os.Args) != 2 {
                fmt.Fprintf(os.Stderr, "Usage: %s hostname\n", os.Args[0])
                fmt.Println("Usage: ", os.Args[0], "hostname")
                os.Exit(1)
        }
        name := os.Args[1]

        addr, err := net.ResolveIPAddr("ip", name)
        if err != nil {
                fmt.Println("Resolution error", err.Error())
                os.Exit(1)
        }
        fmt.Println("Resolved address is ", addr.String())
        os.Exit(0)
}

```

运行这个:

```go
go run ResolveIP.go www.google.com

```

返回以下内容:

```go
Resolved address is 172.217.25.164

```

如果网络类型的第一个参数`ResolveIPAddr()`被给定为`ip6`而不是`ip`，我得到这样的结果:

```go
Resolved address is  2404:6800:4006:801::2004

```

你可能会得到不同的结果，这取决于从你的地址来看谷歌似乎住在哪里。

### 主机查找

`ResolveIPAddr`函数将对主机名执行 DNS 查找，并返回一个 IP 地址。它如何做到这一点取决于操作系统及其配置。例如，Linux/UNIX 系统可能使用`/etc/resolv.conf`或`/etc/hosts`，搜索顺序设置在`/etc/nsswitch.conf`中。

一些主机可能有多个 IP 地址，通常来自多个网络接口卡。他们也可能有多个主机名，充当别名。`LookupHost`函数将返回一片地址。

```go
func LookupHost(name string) (cname string, addrs []string, err error)

```

其中一个地址将被标记为“标准”主机名。如果您想找到规范名称，请使用:

。

```go
func LookupCNAME(name string) (cname string, err error)

```

.

对于 [`www.google.com`](http://www.google.com) ，它打印 IPv4 和 IPv6 地址:

```go
172.217.25.164
2404:6800:4006:806::2004

```

这显示在以下名为`LookupHost.go`的程序中:

```go
/* LookupHost
 */

package main

import (
        "fmt"
        "net"
        "os"
)

func main() {
        if len(os.Args) != 2 {
                fmt.Fprintf(os.Stderr, "Usage: %s hostname\n", os.Args[0])
                os.Exit(1)
        }
        name := os.Args[1]

        addrs, err := net.LookupHost(name)
        if err != nil {
                fmt.Println("Error: ", err.Error())
                os.Exit(2)
        }

        for _, s := range addrs {
                fmt.Println(s)
        }
        os.Exit(0)
}

```

注意，这个函数返回字符串，而不是 IP 地址值。运行时:

```go
go run LookupHost.go

```

它打印出类似这样的内容:

```go
172.217.25.132
2404:6800:4006:807::2004

```

## 服务

服务在主机上运行。它们通常具有很长的生命周期，被设计用来等待请求并对请求做出响应。服务有很多种，他们向客户提供服务的方式也有很多种。互联网世界中的许多服务都基于两种通信方法——TCP 和 UDP——尽管还有其他通信协议，如 SCTP，正准备接管。许多其他类型的服务，如点对点、远程过程调用、通信代理等，都是建立在 TCP 和 UDP 之上的。

### 港口

服务存在于主机上。我们可以使用 IP 地址来定位主机。但是在每台计算机上可能有许多服务，需要一种简单的方法来区分它们。TCP、UDP、SCTP 和其他协议使用的方法是使用端口号。这是一个介于 1 和 65，535 之间的无符号整数，每个服务将自己与这些端口号中的一个或多个相关联。

有许多“标准”端口。Telnet 通常使用 TCP 协议的端口 23。DNS 通过 TCP 或 UDP 使用端口 53。FTP 使用端口 21 和 20，一个用于命令，另一个用于数据传输。HTTP 一般使用端口 80，但也经常使用端口 8000、8080、8088，都是用 TCP。X Window 系统通常使用 TCP 和 UDP 上的 6000-6007 端口。

在 UNIX 系统上，常用端口列在文件`/etc/services`中。Go 具有在所有系统上查找端口的功能:

```go
func LookupPort(network, service string) (port int, err error)

```

网络参数是一个字符串，如`"tcp"`或`"udp"`，而服务是一个字符串，如`"telnet"`或`"domain"`(用于 DNS)。

使用这个的程序是`LookupPort.go`:

```go
/* LookupPort
 */

package main

import (
        "fmt"
        "net"
        "os"
)

func main() {
        if len(os.Args) != 3 {

                fmt.Fprintf(os.Stderr,
                        "Usage: %s network-type service\n",
                        os.Args[0])
                os.Exit(1)
        }
        networkType := os.Args[1]
        service := os.Args[2]

        port, err := net.LookupPort(networkType, service)
        if err != nil {
                fmt.Println("Error: ", err.Error())
                os.Exit(2)
        }

        fmt.Println("Service port ", port)
        os.Exit(0)
}

```

例如，运行`LookupPort tcp telnet`打印服务端口 23。

### TCPAddr 类型

`TCPAddr`类型是包含 IP、端口和区域的结构。需要该区域来区分可能不明确的 IPv6 链路本地地址和站点本地地址，因为不同的网络接口卡(NIC)可能具有相同的 IPv6 地址。

```go
type TCPAddr struct {
    IP   IP
    Port int
    Zone string
}

```

创建一个`TCPAddr`的函数是`ResolveTCPAddr`:

```go
func ResolveTCPAddr(net, addr string) (*TCPAddr, error)

```

其中`net`是`tcp`、`tcp4`或`tcp6`中的一个，`addr`是由主机名或 IP 地址组成的字符串，后跟`:`后的端口号，如`www.google.com:80`或`127.0.0.1:22`。如果地址是 IPv6 地址，其中已经有冒号，那么主机部分必须用方括号括起来，例如`[::1]:23`。另一种特殊情况通常用于服务器，其中主机地址为零，因此 TCP 地址实际上只是端口名，如 HTTP 服务器的`:80`所示。

## TCP 套接字

当您知道如何通过网络和端口 id 访问服务时，接下来该怎么办呢？如果您是一个客户端，您需要一个 API 来允许您连接到一个服务，然后向该服务发送消息并从该服务读取回复。

如果您是服务器，您需要能够绑定到一个端口并监听它。当消息进来时，您需要能够阅读它并写回给客户端。

`net.` `TCPConn`是 Go 类型，允许客户端和服务器之间的全双工通信。感兴趣的两种主要方法如下:

```go
func (c *TCPConn) Write(b []byte) (n int, err error)
func (c *TCPConn) Read(b []byte) (n int, err error)   

```

客户端和服务器都使用`TCPConn`来读取和写入消息。

注意，`TCPConn`实现了`io.Reader`和`io.Writer`接口，因此任何使用读取器或写入器的方法都可以应用于`TCPConn`。

### TCP 客户端

一旦客户端为服务建立了 TCP 地址，它就“拨号”该服务。如果成功，拨号盘返回一个`TCPConn`用于通信。客户端和服务器就此交换消息。通常，客户端使用`TCPConn`向服务器写入请求，并从`TCPConn`读取响应。这种情况一直持续到任一端(或两端)关闭连接。客户端使用以下函数建立 TCP 连接:

```go
func DialTCP(net string, laddr, raddr *TCPAddr) (c *TCPConn, err error)

```

其中`laddr`是本地地址，通常设置为`nil`,`raddr`是服务的远程地址。`net`字符串是`"tcp4"`、`"tcp6"`或`"tcp"`中的一种，这取决于您是想要 TCPv4 连接、TCPv6 连接还是不在乎。

一个简单的例子可以由客户机提供给 web (HTTP)服务器。我们将在后面的章节中更详细地讨论 HTTP 客户端和服务器，所以现在我们保持简单。

客户端可能发送的消息之一是`HEAD`消息。这将向服务器查询有关该服务器和该服务器上的文档的信息。服务器返回信息，但不返回文档本身。发送查询 HTTP 服务器的请求可能如下:

```go
"HEAD / HTTP/1.0\r\n\r\n"

```

这要求提供关于根文档和服务器的信息。典型的回答可能是:

```go
HTTP/1.1 200 OK
Server: nginx/1.10.0 (Ubuntu)
Date: Tue, 28 Feb 2017 10:33:01 GMT
Content-Type: text/html
Content-Length: 2152
Last-Modified: Mon, 13 Oct 2008 02:38:03 GMT
Connection: close
ETag: "48f2b48b-868"
Accept-Ranges: bytes

```

我们首先给程序(`GetHeadInfo.go`)建立一个 TCP 地址的连接，发送请求字符串，然后读取并打印响应。编译后，可以按如下方式调用它:

```go
GetHeadInfo www.google.com:80

```

程序是`GetHeadInfo.go`:

```go
/* GetHeadInfo
 */
package main

import (
        "fmt"
        "io/ioutil"
        "net"
        "os"
)

func main() {
        if len(os.Args) != 2 {
                fmt.Fprintf(os.Stderr, "Usage: %s host:port ", os.Args[0])
                os.Exit(1)
        }
        service := os.Args[1]

        tcpAddr, err := net.ResolveTCPAddr

("tcp4", service)
        checkError(err)

        conn, err := net.DialTCP("tcp", nil, tcpAddr)
        checkError(err)

        _, err = conn.Write([]byte("HEAD / HTTP/1.0\r\n\r\n"))
        checkError(err)

        result, err := ioutil.ReadAll(conn)
        checkError(err)

        fmt.Println(string(result))

        os.Exit(0)
}

func checkError(err error) {
        if err != nil {
                fmt.Fprintf(os.Stderr, "Fatal error: %s", err.Error())
                os.Exit(1)
        }
}

```

要注意的第一点是正在进行的几乎过量的错误检查。这对于网络程序来说是正常的:失败的机会远远大于独立程序。客户端、服务器或中间的任何路由器和交换机上的硬件可能会出现故障；通信可能被防火墙阻止；网络负载可能会导致超时；当客户端与服务器通信时，服务器可能会崩溃。将执行以下检查:

1.  指定的地址中可能有语法错误。
2.  连接到远程服务的尝试可能会失败。例如，请求的服务可能没有运行，或者可能没有这样的主机连接到网络。
3.  虽然已经建立了连接，但是如果连接突然中断或者网络超时，对服务的写入可能会失败。
4.  类似地，读取可能会失败。

从服务器读取需要注释。在这种情况下，我们基本上从服务器读取一个响应。这将由连接上的文件结尾终止。但是，它可能由几个 TCP 包组成，所以我们需要一直读取，直到文件结束。`io/ioutil`函数`ReadAll`将处理这些问题并返回完整的响应。(感谢`golang-nuts`邮件列表上的罗杰·佩佩。)

这涉及到一些语言问题。首先，大多数函数返回一个 dual 值，可能的错误作为第二个值。如果没有错误发生，那么这将是`nil`。在 C 语言中，如果可能的话，通过返回特殊值，如`NULL`、或-1、或零，可以获得相同的行为。在 Java 中，同样的错误检查是通过抛出和捕获异常来管理的，这会使代码看起来非常混乱。

### 日间服务员

我们可以建立的最简单的服务是日间服务。这是 RFC 867 定义的标准互联网服务，TCP 和 UDP 的默认端口都是 13。不幸的是，随着(合理的)对安全的偏执的增加，几乎没有任何网站再运行日间服务器了。没关系；我们可以自己造。(对于那些感兴趣的人，如果你在你的系统上安装了`inetd`，你通常会得到一个日间服务器。)

服务器在一个端口上注册并监听该端口。然后它阻塞一个“接受”操作，等待客户端连接。当客户端连接时，accept 调用返回，并带有一个连接对象。日间服务非常简单，只需将当前时间写入客户端，关闭连接，然后继续等待下一个客户端。

相关电话如下:

```go
func ListenTCP(net string, laddr *TCPAddr) (l *TCPListener, err error)
func (l *TCPListener) Accept() (c Conn, err error)

```

参数`net`可以设置为字符串`"tcp"`、`"tcp4"`或`"tcp6"`中的一个。如果要监听所有网络接口，IP 地址应设置为零；如果只想监听某个网络接口，则应设置为该接口的 IP 地址。如果端口设置为零，那么操作系统将为您选择一个端口。否则，你可以自己选择。请注意，在 UNIX 系统上，您不能监听低于 1024 的端口，除非您是系统管理员、root 用户，而低于 128 的端口是由 IETF 标准化的。示例程序选择端口 1200 没有任何特殊原因。TCP 地址给定为`:1200`—所有接口，端口 1200。

程序是`DaytimeServer.go`:

```go
/* DaytimeServer
 */
package main

import (
        "fmt"
        "net"
        "os"
        "time"
)

func main() {

        service := ":1200"
        tcpAddr, err := net.ResolveTCPAddr("tcp", service)
        checkError(err)

        listener, err := net.ListenTCP("tcp", tcpAddr)
        checkError(err)

        for {
                conn, err := listener.Accept()
                if err != nil {
                        continue
                }

                daytime := time.Now().String()

                conn.Write([]byte(daytime)) // don't care about return value
                conn.Close()                // we're finished with this client
        }
}

func checkError(err error) {
        if err != nil {
                fmt.Fprintf(os.Stderr, "Fatal error: %s", err.Error())
                os.Exit(1)
        }
}

```

如果你运行这个服务器，它将只是在那里等待，不做太多。当客户端连接到它时，它将通过向它发送日间字符串来响应，然后返回等待下一个客户端。

请注意，与客户端相比，服务器中的错误处理发生了变化。服务器应该永远运行，这样，如果客户机出现任何错误，服务器就会忽略该客户机并继续运行。否则，客户端可能会试图破坏与服务器的连接并使其崩溃！

我们还没有建立客户。这很容易，只需更改以前的客户端以省略初始写入。或者，只需打开到主机 telnet 连接:

```go
telnet localhost 1200

```

这将产生如下输出:

```go
$telnet localhost 1200
Trying 127.0.0.1...
Connected to localhost.
Escape character is '^]'.
2017-01-02 20:13:21.934698384 +1100 AEDTConnection closed by foreign host.

```

其中`2017-01-02 20:13:21.934698384 +1100 AEDT`是服务器的输出。

### 多线程服务器

`echo`是另一个简单的 IETF 服务。`SimpleEchoServer.go`程序只是读取客户端输入的内容并将其发送回来:

```go
/* SimpleEchoServer
 */
package main

import (
        "fmt"
        "net"
        "os"
)

func main() {

        service := ":1201"
        tcpAddr, err := net.ResolveTCPAddr("tcp4", service)
        checkError(err)

        listener, err := net.ListenTCP("tcp", tcpAddr)
        checkError(err)

        for {
                conn, err := listener.Accept()
                if err != nil {
                        continue
                }
                handleClient(conn)
                conn.Close() // we're finished
        }
}

func handleClient(conn net.Conn) {
        var buf [512]byte
        for {
                n, err := conn.Read(buf[0:])
                if err != nil {
                        return
                }
                fmt.Println(string(buf[0:]))
                _, err2 := conn.Write(buf[0:n])
                if err2 != nil {
                        return
                }
        }
}

func checkError(err error) {
        if err != nil {
                fmt.Fprintf(os.Stderr, "Fatal error: %s", err.Error())
                os.Exit(1)
        }
}

```

虽然它可以工作，但是这个服务器有一个重要的问题:它是单线程的。当一个客户端打开了一个连接时，没有其他客户端可以连接。其他客户端被阻止，可能会超时。幸运的是，这很容易通过使客户端处理程序成为一个`go`例程来解决。我们还将关闭连接移到了处理程序中，因为它现在属于那里。这个程序叫做`ThreadedEchoServer.` `go`:

```go
/* ThreadedEchoServer
 */
package main

import (
        "fmt"
        "net"
        "os"
)

func main() {

        service := ":1201"
        tcpAddr, err := net.ResolveTCPAddr("tcp", service)
        checkError(err)

        listener, err := net.ListenTCP("tcp", tcpAddr)
        checkError(err)

        for {
                conn, err := listener.Accept()
                if err != nil {
                        continue
                }
                // run as a goroutine
                go handleClient(conn)
        }
}

func handleClient(conn net.Conn) {
        // close connection on exit
        defer conn.Close()

        var buf [512]byte
        for {
                // read up to 512 bytes
                n, err := conn.Read(buf[0:])
                if err != nil {
                        return
                }
                fmt.Println(string(buf[0:]))
                // write the n bytes read
                _, err2 := conn.Write(buf[0:n])
                if err2 != nil {
                        return
                }
        }
}

func checkError(err error) {
        if err != nil {
                fmt.Fprintf(os.Stderr, "Fatal error: %s", err.Error())
                os.Exit(1)
        }

}

```

## 控制 TCP 连接

### 超时

如果客户端响应不够快，即没有及时向服务器写入请求，服务器可能希望客户端超时。这应该是一段很长的时间(几分钟)，因为用户可能会慢慢来。相反，客户端可能希望服务器超时(在短得多的时间后)。两者都是这样做的:

```go
func (c *IPConn) SetDeadline(t time.Time) error

```

这是在套接字上的任何读取或写入之前完成的。

### 活着

客户端可能希望保持与服务器的连接，即使它没有要发送的内容。它可以使用这个:

```go
func (c *TCPConn) SetKeepAlive(keepalive bool) error

```

还有其他几种连接控制方法，记录在`net`包中。

## UDP 数据报

在无连接协议中，每条消息都包含有关其来源和目的地的信息。没有使用长期套接字建立的“会话”。UDP 客户端和服务器使用数据报，数据报是包含源和目的信息的单独消息。除非客户端或服务器维护状态，否则这些消息不会维护状态。不保证消息会到达，或者可能会无序到达。

对于客户端来说，最常见的情况是发送一条消息并希望收到回复。对于服务器来说，最常见的情况是接收一条消息，然后向客户端发送一个或多个回复。但是，在对等的情况下，服务器可能只是将消息转发给其他对等方。

Go 的 TCP 和 UDP 处理之间的主要区别是如何处理来自多个客户端的数据包，没有 TCP 会话的缓冲来管理事情。需要的主要调用如下:

```go
func ResolveUDPAddr(net, addr string) (*UDPAddr, error)
func DialUDP(net string, laddr, raddr *UDPAddr) (c *UDPConn, err error)
func ListenUDP(net string, laddr *UDPAddr) (c *UDPConn, err error)
func (c *UDPConn) ReadFromUDP(b []byte) (n int, addr *UDPAddr, err error
func (c *UDPConn) WriteToUDP(b []byte, addr *UDPAddr) (n int, err error)

```

UDP 时间服务的客户端不需要做很多改变；将`UDPDaytimeClient.go`程序中的`...TCP...`调用改为`...UDP...`调用即可:

```go
/* UDPDaytimeClient
 */
package main

import (
        "fmt"
        "net"
        "os"
)

func main() {
        if len(os.Args) != 2 {
                fmt.Fprintf(os.Stderr, "Usage: %s host:port", os.Args[0])
                os.Exit(1)
        }
        service := os.Args[1]

        udpAddr, err := net.ResolveUDPAddr("udp", service)
        checkError(err)

        conn, err := net.DialUDP("udp", nil, udpAddr)
        checkError(err)

        _, err = conn.Write([]byte("anything"))
        checkError(err)

        var buf [512]byte
        n, err := conn.Read(buf[0:])
        checkError(err)

        fmt.Println(string(buf[0:n]))

        os.Exit(0)
}

func checkError(err error) {
        if err != nil {
                fmt.Fprintf(os.Stderr, "Fatal error ", err.Error())
                os.Exit(1)

        }
}

```

而服务器必须在程序`UDPDaytimeServer.go`中再做一些改变:

```go
/* UDPDaytimeServer
 */
package main

import (
        "fmt"
        "net"
        "os"
        "time"
)

func main() {

        service := ":1200"
        udpAddr, err := net.ResolveUDPAddr("udp", service)
        checkError(err)

        conn, err := net.ListenUDP("udp", udpAddr)
        checkError(err)

        for {
                handleClient(conn)

        }
}

func handleClient(conn *net.UDPConn) {

        var buf [512]byte

        _, addr, err := conn.ReadFromUDP(buf[0:])
        if err != nil {
                return
        }

        daytime := time.Now().String()

        conn.WriteToUDP([]byte(daytime), addr)
}

func checkError(err error) {
        if err != nil {
                fmt.Fprintf(os.Stderr, "Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

服务器按如下方式运行:

```go
go run UDPDaytimeServer.go

```

同一主机上的客户端运行如下:

```go
go run UDPDaytimeClient.go localhost:1200

```

输出将是这样的:

```go
2017-03-01 21:37:03.988603994 +1100 AEDT

```

## 服务器监听多个套接字

一个服务器可能试图监听多个客户机，不只是在一个端口上，而是在许多端口上。在这种情况下，它必须在端口之间使用某种轮询机制。

在 C 中，`select()`调用让内核完成这项工作。该调用需要许多文件描述符。该过程被暂停。当其中一个上的 I/O 就绪时，唤醒完成，该过程可以继续。这比忙轮询便宜。在 Go 中，您可以通过为每个端口使用不同的`go`例程来完成相同的任务。当较低级别的`select()`发现 I/O 已经为线程准备好时，线程将变得可运行。

## Conn、PacketConn 和侦听器类型

到目前为止，我们已经区分了 TCP 的 API 和 UDP 的 API，例如使用分别返回`TCPConn`和`UDPConn`的`DialTCP`和`DialUDP`。`Conn`类型是一个接口，`TCPConn`和`UDPConn`都实现了这个接口。在很大程度上，您可以处理这个接口，而不是这两种类型。

您可以使用一个单独的功能来代替 TCP 和 UDP 的单独拨号功能:

```go
func Dial(net, laddr, raddr string) (c Conn, err error)

```

`net`可以是`tcp`、`tcp4`(仅 IPv4)、`tcp6`(仅 IPv6)、`udp`、`udp4`(仅 IPv4)、`udp6`(仅 IPv6)、`ip`、`ip4`(仅 IPv4)和`ip6`(仅 IPv6)中的任何一个，以及几个特定于 UNIX 的，例如用于 UNIX 套接字的`unix`。它将返回一个适当的`Conn`接口实现。注意，这个函数采用一个字符串而不是地址作为`raddr`参数，这样使用它的程序可以避免首先计算出地址类型。

使用此功能可以对程序进行微小的更改。例如，从网页获取`HEAD`信息的早期程序可以重写为`IPGetHeadInfo.` `go`:

```go
/* IPGetHeadInfo
 */
package main

import (
        "bytes"
        "fmt"
        "io"
        "net"
        "os"
)

func main() {
        if len(os.Args) != 2 {
                fmt.Fprintf(os.Stderr, "Usage: %s host:port", os.Args[0])
                os.Exit(1)
        }
        service := os.Args[1]

        conn, err := net.Dial("tcp", service)
        checkError(err)

        _, err = conn.Write([]byte("HEAD / HTTP/1.0\r\n\r\n"))

        checkError(err)

        result, err := readFully(conn)
        checkError(err)

        fmt.Println(string(result))

        os.Exit(0)
}

func checkError(err error) {
        if err != nil {
                fmt.Fprintf(os.Stderr, "Fatal error: %s", err.Error())
                os.Exit(1)
        }
}

func readFully(conn net.Conn) ([]byte, error) {
        defer conn.Close()

        result := bytes.NewBuffer(nil)
        var buf [512]byte
        for {
                n, err := conn.Read(buf[0:])
                result.Write(buf[0:n])
                if err != nil {
                        if err == io.EOF {
                                break
                        }
                        return nil, err
                }
        }
        return result.Bytes(), nil

}

```

这可以在我自己的机器上运行，如下所示:

```go
go run  IPGetHeadInfo.go  localhost:80

```

它打印了关于在端口 80 上运行的服务器的以下信息:

```go
HTTP/1.1 200 OK
Server: nginx/1.10.0 (Ubuntu)
Date: Wed, 01 Mar 2017 10:42:39 GMT
Content-Type: text/html
Content-Length: 2152
Last-Modified: Mon, 13 Oct 2008 02:38:03 GMT
Connection: close
ETag: "48f2b48b-868"
Accept-Ranges: bytes

```

使用此函数可以类似地简化服务器的编写:

```go
func Listen(net, laddr string) (l Listener, err error)

```

这将返回一个实现了`Listener`接口的对象。这个接口有一个方法:

```go
func (l Listener) Accept() (c Conn, err error)

```

这将允许构建服务器。利用这一点，前面给出的多线程`Echo`服务器变成了`ThreadedIPEchoServer.` `go`:

```go
/* ThreadedIPEchoServer
 */
package main

import (
        "fmt"
        "net"
        "os"
)

func main() {

        service := ":1200"
        listener, err := net.Listen("tcp", service)
        checkError(err)

        for {
                conn, err := listener.Accept()
                if err != nil {
                        continue
                }
                go handleClient(conn)
        }
}

func handleClient(conn net.Conn) {
        defer conn.Close()

        var buf [512]byte
        for {
                n, err := conn.Read(buf[0:])
                if err != nil {
                        return
                }
                _, err2 := conn.Write(buf[0:n])
                if err2 != nil {
                        return
                }
        }
}

func checkError(err error) {
        if err != nil {
                fmt.Fprintf(os.Stderr, "Fatal error: %s", err.Error())
                os.Exit(1)
        }
}

```

如果你想写一个 UDP 服务器，有一个名为`PacketConn`的接口和一个返回实现的方法:

```go
func ListenPacket(net, laddr string) (c PacketConn, err error)

```

这个接口有处理包读写的主要方法`ReadFrom`和`WriteTo`。

Go `net`包推荐使用这些接口类型，而不是具体的接口类型。但是通过使用它们，你会失去特定的方法，比如`TCPConn`的`SetKeepAlive`和`UDPConn`的`SetReadBuffer`，除非你进行类型转换。这是你的选择。

## 原始套接字和 IPConn 类型

本节涵盖了大多数程序员不太可能需要的高级材料。它处理原始套接字，允许程序员构建自己的 IP 协议，或者使用 TCP 或 UDP 以外的协议。

TCP 和 UDP 并不是唯一建立在 IP 层之上的协议。网站 [`http://www.iana.org/assignments/protocol-numbers`](http://www.iana.org/assignments/protocol-numbers) 列出了其中的大约 140 个(这个列表通常可以在 UNIX 系统的文件`/etc/protocols`中找到)。在这个列表中，TCP 和 UDP 分别只排在第 6 和第 17 位。

Go 允许您构建所谓的原始套接字，使您能够使用这些其他协议之一进行通信，甚至构建自己的协议。但是它提供了最低限度的支持:它将连接主机，并在主机之间读写数据包。在下一章，我们将着眼于在 TCP 之上设计和实现你自己的协议；本节考虑的是同一类型的问题，但是是在 IP 层。

为了简单起见，我们使用最简单的例子:如何向主机发送 IPv4 ping 消息。Ping 使用 ICMP 协议中的`echo`命令。这是一个面向字节的协议，客户端向另一台主机发送字节流，主机进行回复。ICMP 数据包有效负载的格式如下:

*   第一个字节是 8，代表回应消息。
*   第二个字节是零。
*   第三和第四个字节是整个消息的校验和。
*   第五和第六个字节是一个任意的标识符。
*   第七个和第八个字节是一个任意的序列号。
*   数据包的其余部分是用户数据。

可以使用`Conn.Write`方法发送数据包，该方法用这个有效载荷准备数据包。收到的回复包括 IPv4 报头，占用 20 个字节。(例如，参见维基百科关于因特网控制消息协议 ICMP 的文章。)

下面这个名为`Ping.` `go`的程序将准备一个 IP 连接，向一个主机发送 ping 请求，并得到回复。您可能需要 root 访问权限才能成功运行它:

```go
/* Ping
 */
package main

import (
        "bytes"
        "fmt"
        "io"
        "net"
        "os"
)

// change this to my own IP address or set to 0.0.0.0
const myIPAddress = "192.168.1.2"
const ipv4HeaderSize = 20

func main() {
        if len(os.Args) != 2 {
                fmt.Println("Usage: ", os.Args[0], "host")
                os.Exit(1)
        }

        localAddr, err := net.ResolveIPAddr("ip4", myIPAddress)
        if err != nil {
                fmt.Println("Resolution error", err.Error())
                os.Exit(1)
        }

        remoteAddr, err := net.ResolveIPAddr("ip4", os.Args[1])
        if err != nil {
                fmt.Println("Resolution error", err.Error())

                os.Exit(1)
        }

        conn, err := net.DialIP("ip4:icmp", localAddr, remoteAddr)
        checkError(err)

        var msg [512]byte
        msg[0] = 8  // echo
        msg[1] = 0  // code 0
        msg[2] = 0  // checksum, fix later
        msg[3] = 0  // checksum, fix later
        msg[4] = 0  // identifier[0]
        msg[5] = 13 // identifier[1] (arbitrary)
        msg[6] = 0  // sequence[0]
        msg[7] = 37 // sequence[1] (arbitrary)
        len := 8

        // now fix checksum bytes
        check := checkSum(msg[0:len])
        msg[2] = byte(check >> 8)
        msg[3] = byte(check & 255)

        // send the message
        _, err = conn.Write(msg[0:len])
        checkError(err)

        fmt.Print("Message sent:    ")
        for n := 0; n < 8; n++ {
                fmt.Print(" ", msg[n])
        }
        fmt.Println()

        // receive a reply
        size, err2 := conn.Read(msg[0:])
        checkError(err2)

        fmt.Print("Message received:")
        for n := ipv4HeaderSize; n < size; n++ {
                fmt.Print(" ", msg[n])
        }
        fmt.Println()
        os.Exit(0)

}

func checkSum(msg []byte) uint16 {
        sum := 0

        // assume even for now
        for n := 0; n < len(msg); n += 2 {
                sum += int(msg[n])*256 + int(msg[n+1])
        }
        sum = (sum >> 16) + (sum & 0xffff)
        sum += (sum >> 16)
        var answer uint16 = uint16(^sum)
        return answer
}

func checkError(err error) {
        if err != nil {
                fmt.Fprintf(os.Stderr, "Fatal error: %s", err.Error())
                os.Exit(1)
        }
}

func readFully(conn net.Conn) ([]byte, error) {
        defer conn.Close()

        result := bytes.NewBuffer(nil)
        var buf [512]byte
        for {
                n, err := conn.Read(buf[0:])
                result.Write(buf[0:n])
                if err != nil {
                        if err == io.EOF {
                                break
                        }
                        return nil, err
                }
        }
        return result.Bytes(), nil
}

```

它使用目标地址作为参数来运行。接收到的消息与发送的消息的区别仅在于第一个类型字节以及第三和第四个校验和字节，如下所示:

```go
Message sent:        8 0 247 205 0 13 0 37
Message received: 0 0 255 205 0 13 0 37

```

## 结论

本章考虑了 IP、TCP 和 UDP 级别的编程。如果您想要实现自己的协议或者为现有协议构建客户端或服务器，这通常是必要的。

Footnotes [1](#Fn1_source)

最近，麻省理工学院将他们的 A 类网络归还给了游泳池。 [`http://www.iana.org/assignments/ipv4-address-space/ipv4-address-space.xml`](http://www.iana.org/assignments/ipv4-address-space/ipv4-address-space.xml) 。