# 四、数据序列化

客户端和服务器需要通过消息交换信息。TCP 和 UDP 提供了实现这一点的传输机制。这两个过程还需要有一个合适的协议，这样消息交换才能有意义地进行。

消息作为字节序列在网络上发送，除了作为线性字节流之外，它没有任何结构。我们将在下一章讨论消息的各种可能性以及定义它们的协议。在这一章中，我们集中讨论消息的一个组成部分——被传输的数据。

程序通常会构建复杂的数据结构来保存当前的程序状态。在与远程客户端或服务进行对话时，程序将试图通过网络传输这种数据结构——也就是说，在应用程序自己的地址空间之外。

## 结构数据

编程语言使用结构化数据，如下所示:

*   记录/结构
*   变体记录
*   数组:固定大小或可变大小
*   字符串:固定大小或可变大小
*   表格:记录数组
*   非线性结构，例如
    *   循环链表
    *   二叉树
    *   引用其他对象的对象

IP、TCP 或 UDP 数据包都不知道这些数据类型的含义。它们所能包含的只是一个字节序列。因此，应用程序必须将任何数据序列化为字节流才能写入数据，并在读取数据时将字节流反序列化为合适的数据结构。这两个操作称为编组和解组，分别为 [<sup>1</sup>](#Fn1) 。

例如，考虑发送以下两列可变长度字符串的可变长度表:

<colgroup><col> <col></colgroup> 
| 图像读取器设备（figure-reader electronic device 的缩写） | 节目编排者 |
| 黎平 | 分析家 |
| 一定的 | 经理 |

这可以通过各种方式来实现。例如，假设已知数据将是两列表中未知数量的行。那么编组的表单可以是:

```go
3                // 3 rows, 2 columns assumed
4 fred           // 4 char string,col 1
10 programmer    // 10 char string,col 2
6 liping         // 6 char string, col 1
7 analyst        // 7 char string, col 2
8 sureerat       // 8 char string, col 1
7 manager        // 7 char string, col 2

```

可变长度的东西也可以用一个“非法”的值来表示它们的长度，比如用`\0`来表示字符串。前面的表也可以用行数来写，但是每个字符串都以`\0`结束(换行符是为了可读性，不是序列化的一部分):

```go
3
fred\0        
programmer\0
liping\0
analyst\0
sureerat\0
manager\0

```

或者，可以知道数据是一个三行的固定表，包含两列长度分别为 8 和 10 的字符串。那么表的序列化可以是(换行符也不是序列化的一部分):

```go
fred\0\0\0\0
programmer
liping\0\0
analyst\0\0\0
sureerat
manager\0\0\0

```

这些格式都可以，但是消息交换协议必须指定使用哪种格式，或者允许在运行时确定使用哪种格式。

## 双方协议

上一节概述了数据序列化的问题。实际上，细节可能要复杂得多。例如，考虑第一种可能性，将一个表编组到流中:

```go
3
4 fred
10 programmer
6 liping
7 analyst
8 sureerat
7 manager

```

许多问题出现了。例如，表可能有多少行——也就是说，我们需要多大的整数来描述行大小？如果它小于或等于 255，那么单个字节就可以了，但是如果它大于 255，那么就需要一个短整型、整型或长整型。每个字符串的长度也会出现类似的问题。对于字符本身，它们属于哪个字符集？7 位 ASCII 码？16 位 Unicode？字符集的问题将在后面的章节中详细讨论。

这种序列化是不透明的或隐式的。如果数据是用这种格式编排的，那么在序列化的数据中就没有说明应该如何对其进行解组。为了正确地解组数据，解组方必须确切地知道数据是如何序列化的。例如，如果行数被封送为 8 位整数，但被解组为 16 位整数，那么当接收方试图将 3 和 4 解组为 16 位整数时，将会出现不正确的结果，并且接收程序稍后几乎肯定会失败。

早期一个众所周知的序列化方法是 Sun 的 RPC 使用的 XDR(外部数据表示)，后来被称为 ONC(开放网络计算)。RFC 1832 定义了 XDR，了解这一规格有多精确很有意义。尽管如此，XDR 本质上是类型不安全的，因为序列化数据不包含类型信息。它在 ONC 中使用的正确性主要是由编译器为编组和解组生成代码来保证的。

Go 不包含对不透明序列化数据的编组和解组的显式支持。Go 中的 RPC 包不使用 XDR，而是使用 Gob 序列化，这将在本章后面描述。

## 自描述数据

自描述数据携带数据的类型信息。例如，以前的数据可能会被编码如下:

```go
table
   uint8 3
   uint 2
string
   uint8 4
   []byte fred
string
   uint8 10
   []byte programmer
string
   uint8 6
   []byte liping
string
   uint8 7
   []byte analyst
string
   uint8 8
   []byte sureerat
string
   uint8 7
   []byte manager

```

当然，真正的编码通常不会像示例中那样麻烦和冗长:小整数将被用作类型标记，并且整个数据将被打包在尽可能小的字节数组中。(不过，XML 提供了一个反例。)但是，原理是封送拆收器将在序列化数据中生成这样的类型信息。解组器将知道类型生成规则，并将能够使用它们来重建正确的数据结构。

## ASN.1

抽象语法符号一(ASN.1)最初是在 1984 年为电信行业设计的。ASN.1 是一个复杂的标准，它的一个子集在包`asn1`中被 Go 支持。它从复杂的数据结构中构建自描述的序列化数据。它在当前网络系统中的主要用途是作为 X.509 证书的编码，在认证系统中大量使用。Go 中的支持基于读写 X.509 证书所需的内容。

两个函数允许我们编组和解组数据:

```go
func Marshal(val interface{}) ([]byte, error)
func Unmarshal(val interface{}, b []byte) (rest []byte, err error)

```

第一个函数将数据值封送到一个序列化的字节数组中，第二个函数将它解组。但是，类型接口的第一个参数值得进一步研究。给定一个类型的变量，我们可以通过传递它的值来封送它。为了解组它，我们需要一个命名类型的变量来匹配序列化的数据。具体细节将在后面讨论。但是我们还需要确保变量被分配给该类型的内存，这样实际上就有现有的内存供解组写入值。

我们在`ASN1.go`中用一个简单的例子来说明一个整数的编组和解组。我们可以将一个整数值传递给`marshal`以返回一个字节数组，并将该数组解组为一个整数变量，如下所示:

```go
/* ASN1
 */

package main

import (
        "encoding/asn1"
        "fmt"
        "os"
)

func main() {
        val := 13
        fmt.Println("Before marshal/unmarshal: ", val)
        mdata, err := asn1.Marshal(val)
        checkError(err)

        var n int
        _, err1 := asn1.Unmarshal(mdata, &n)
        checkError(err1)

        fmt.Println("After marshal/unmarshal: ", n)
}

func checkError(err error) {
        if err != nil {
                fmt.Fprintf(os.Stderr, "Fatal error: %s", err.Error())
                os.Exit(1)
        }
}

```

该程序运行如下:

```go
go run ASN1.go

```

解组值，当然是`13`。

一旦我们超越了这一点，事情就变得更难了。为了管理更复杂的数据类型，我们必须更仔细地研究 ASN.1 支持的数据结构，以及 ASN.1 在 Go 中是如何支持的。

任何序列化方法都能够处理某些数据类型，而不能处理其他一些数据类型。因此，为了确定任何序列化(如 ASN.1)的适用性，您必须查看可能支持的数据类型和您希望在应用程序中使用的数据类型。以下 ASN.1 类型摘自 [`http://www.obj-sys.com/asn1tutorial/node4.html`](http://www.obj-sys.com/asn1tutorial/node4.html) 。

简单类型如下:

*   `BOOLEAN`:双态变量值
*   `INTEGER`:模型整型变量值
*   `BIT STRING`:模拟任意长度的二进制数据
*   `OCTET STRING`:模拟长度为 8 的倍数的二进制数据
*   `NULL`:表示序列元素的有效缺失
*   `OBJECT IDENTIFIER`:命名信息对象
*   `REAL`:模拟实变量值
*   `ENUMERATED`:用至少三种状态模拟变量值
*   `CHARACTER STRING`:模拟指定字符集中的字符串值

字符串可以来自某些字符集:

*   `NumericString` : 0，1，2，3，4，5，6，7，8，9，空格
*   `PrintableString`:大小写字母、数字、空格、撇号、左/右括号、加号、逗号、连字符、句号、实线、冒号、等号和问号
*   `TeletexString (T61String)`:CCITT T61 中的 Teletex 字符集，空格，删除
*   `VideotexString`:CCITT 的 T.100 和 T.101 中的可视图文字符集，空格，删除
*   `VisibleString`(输入字符串):打印国际 ASCII 字符集，空格
*   `IA5String`:国际字母表 5(国际 ASCII)
*   `GraphicString 25`:所有注册的 G 组，空格`GraphicString`
*   除此之外，还有其他字符串类型，特别是`UTF8String`

最后，还有结构化类型:

*   `SEQUENCE`:对不同类型变量的有序集合建模
*   `SEQUENCE OF`:对同一类型变量的有序集合进行建模
*   对不同类型的变量的无序集合建模
*   `SET OF`:对同一类型变量的无序集合建模
*   `CHOICE`:指定不同类型的集合，从中选择一种类型
*   `SELECTION`:从指定的`CHOICE`类型中选择一个组件类型
*   `ANY`:允许应用程序指定类型

Note

`ANY`是不推荐使用的 ASN.1 结构化类型。已经换成 X.680 开放式了。

并不是所有这些都是 Go 支持的。Go 并不支持所有可能的值。Go `asn1`包文档中给出的规则如下:

*   ASN.1 `INTEGER`可以写入到`int`或`int64`中。如果编码值不符合 Go 类型，`Unmarshal`返回一个解析错误。
*   ASN.1 `BIT STRING`可以写入到`BitString`中。
*   ASN.1 `OCTET STRING`可以写入到`[]byte`中。
*   ASN.1 `OBJECT IDENTIFIER`可以写入到`ObjectIdentifier`中。
*   ASN.1 `ENUMERATED`可以写入到`Enumerated`中。
*   ASN.1 `UTCTIME`或`GENERALIZEDTIME`可以被写入`*time.Time`。
*   ASN.1 `PrintableString`或`IA5String`可以写入字符串。
*   任何上述 ASN.1 值都可以写入一个`interface{}`。存储在接口中的值具有相应的 Go 类型。对于整数，该类型是`int64`。
*   如果可以将`x`写入片的元素类型，则可以将 ASN.1 `SEQUENCE OF x`或`SET OF x`写入片。
*   如果序列中的每个元素都可以写入结构中的相应元素，则 ASN.1 `SEQUENCE`或`SET`可以写入 Go 结构。

Go 对 ASN.1 进行了真正的限制，比如 ASN.1 允许任意大小的整数，而 Go 实现最多只允许有符号的 64 位整数。另一方面，Go 区分有符号和无符号类型，而 ASN.1 不区分。例如，如果值`uint64`对于`int64`来说太大，传输可能会失败。

同理，ASN.1 允许几种不同的字符集，而 Go 包声明只支持`PrintableString`和`IA5String` (ASCII)。ASN.1 现在有了 Unicode UTF8 字符串类型，Go 也支持这种类型，但目前还没有文档。

我们已经看到，像整数这样的值可以很容易地进行编组和解组。其他基本类型如布尔型和实数型也可以类似地处理。完全由 ASCII 字符或 UTF8 字符组成的字符串可以被编组和解组。只要字符串仅由 ASCII 或 UTF8 字符组成，此代码就有效:

```go
s := "hello"
mdata, _ := asn1.Marshal(s)

var newstr string
asn1.Unmarshal(mdata, &newstr)

```

ASN.1 还包括一些不在这个列表中的“有用类型”，比如 UTC 时间。Go 支持此 UTC 时间类型。这意味着您可以以一种其他数据值不可能的方式传递时间值。ASN.1 不支持指针，但是 Go 有专门的代码来管理指向时间值的指针。函数`Now()`返回`*time.Time`。特殊的代码对此进行整理，并且可以将其解组到一个指向`time.Time`对象的指针变量中。因此，这段代码是有效的:

```go
t := time.Now()
mdata, err := asn1.Marshal(t)

var newtime = new(time.Time)
_, err1 := asn1.Unmarshal(newtime, mdata)

```

`LocalTime`和`new`都处理指向一个`*time.Time`的指针，而 Go 处理这个特例。程序`ASN1basic.go`说明了这些:

```go
/* ASN.1 Basic
 */

package main

import (
        "encoding/asn1"
        "fmt"
        "os"
        "time"
)

func main() {

        t := time.Now()
        fmt.Println("Before marshalling: ", t.String())

        mdata, err := asn1.Marshal(t)
        checkError(err)
        fmt.Println("Marshalled ok")

        var newtime = new(time.Time)
        _, err1 := asn1.Unmarshal(mdata, newtime)
        checkError(err1)

        fmt.Println("After marshal/unmarshal: ", newtime.String())

        s := "hello \u00bc"
        fmt.Println("Before marshalling: ", s)

        mdata2, err := asn1.Marshal(s)
        checkError(err)
        fmt.Println("Marshalled ok")

        var newstr string
        _, err2 := asn1.Unmarshal(mdata2, &newstr)
        checkError(err2)

        fmt.Println("After marshal/unmarshal: ", newstr)

}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

当它如下运行时:

```go
go run ASN1basic.go

```

它打印出类似这样的内容:

```go
Before marshalling:  2017-03-02 22:31:16.878943019 +1100 AEDT
Marshalled ok
After marshal/unmarshal:  2017-03-02 22:31:16 +1100 AEDT
Before marshalling:  hello ¼
Marshalled ok
After marshal/unmarshal:  hello ¼

```

一般来说，您可能想要封送和解封送结构。除了时间这个特例，Go 会很乐意处理结构，但是不会处理指向结构的指针。像`new`这样的操作会创建指针，所以您必须在编组/解组它们之前解引用它们。Go 通常会在需要时解引用指针，但在这种情况下不会，所以您必须显式地解引用它们。这两者都适用于一种类型`T`:

```go
// using variables
var t1 T
t1 = ...
mdata1, _ := asn1.Marshal(t)

var newT1 T
asn1.Unmarshal(&newT1, mdata1)

// using pointers
var t2 = new(T)
*t2 = ...
mdata2, _ := asn1.Marshal(*t2)

var newT2 = new(T)
asn1.Unmarshal(newT2, mdata2)

```

指针和变量的任何合适的组合都可以。这里我们没有给出完整的例子，因为应用这些规则应该足够简单。

结构中的所有字段都必须是可导出的，也就是说，字段名必须以大写字母开头。Go 使用反射包来编组/解组结构，因此它必须能够检查所有字段。无法封送此类型:

```go
type T struct {
    Field1 int
    field2 int // not exportable
}

```

ASN.1 只处理数据类型。它不考虑结构字段的名称。因此，下面的类型 T1 可以被编组/解组到类型`T2`中，因为对应的字段是相同的类型:

```go
type T1 struct {
    F1 int
    F2 string
}

type T2 struct {
    FF1 int
    FF2 string
}

```

不仅每个字段的类型必须匹配，数量也必须匹配。这两种类型不起作用:

```go
type T1 struct {
    F1 int
}

type T2 struct {
    F1 int
    F2 string // too many fields
}

```

我们没有给出完整的代码示例，因为我们不会用到这些特性。

ASN.1 说明了实现序列化方法的人可以做出的许多选择。可以通过使用更多的代码对指针进行特殊处理，例如强制名称匹配。字符串的顺序和数量将取决于序列化规范的细节、它所允许的灵活性以及利用这种灵活性所需的编码工作。值得注意的是，其他序列化格式会做出不同的选择，不同语言的实现也会强制执行不同的规则。

### ASN.1 日间客户端和服务器

现在(最后)让我们转向使用 ASN.1 跨网络传输数据。

我们可以使用上一章的技术编写一个 TCP 服务器，以 ASN.1 `Time`的形式提供当前时间。一个服务器是`ASNDaytimeServer.go`:

```go
/* ASN1 DaytimeServer
 */
package main

import (
        "encoding/asn1"
        "fmt"
        "net"Calibri
        "os"
        "time"
)

func main() {

        service := ":1200"
        tcpAddr, err := net.ResolveTCPAddr("tcp", service)
        checkError(err)

        listener, err := net.ListenTCP("tcp", tcpAddr)
        checkError(err)

        for {
                conn, err := listener.Accept()
                if err != nil {
                        continue
                }

                daytime := time.Now()
                // Ignore return network errors.
                mdata, _ := asn1.Marshal(daytime)
                conn.Write(mdata)
                conn.Close() // we're finished
        }
}

func checkError(err error) {
        if err != nil {
                fmt.Fprintf(os.Stderr, "Fatal error: %s", err.Error())
                os.Exit(1)
        }
}

```

这可以编译成一个可执行文件，比如`ASN1DaytimeServer`，并且不带参数运行。它将等待连接，然后将时间作为 ASN.1 字符串发送给客户端。

一个客户是`ASNDaytimeClient.go`:

```go
/* ASN.1 DaytimeClient
 */
package main

import (
        "bytes"
        "encoding/asn1"
        "fmt"
        "io"
        "net"
        "os"
        "time"
)

func main() {
        if len(os.Args) != 2 {
                fmt.Fprintf(os.Stderr, "Usage: %s host:port", os.Args[0])
                os.Exit(1)
        }
        service := os.Args[1]

        conn, err := net.Dial("tcp", service)

        checkError(err)

        result, err := readFully(conn)
        checkError(err)

        var newtime time.Time
        _, err1 := asn1.Unmarshal(result, &newtime)
        checkError(err1)

        fmt.Println("After marshal/unmarshal: ", newtime.String())

        os.Exit(0)
}

func checkError(err error) {
        if err != nil {
                fmt.Fprintf(os.Stderr, "Fatal error: %s", err.Error())
                os.Exit(1)
        }
}

func readFully(conn net.Conn) ([]byte, error) {
        defer conn.Close()

        result := bytes.NewBuffer(nil)
        var buf [512]byte
        for {
                n, err := conn.Read(buf[0:])
                result.Write(buf[0:n])
                if err != nil {
                        if err == io.EOF {
                                break
                        }
                        return nil, err
                }
        }
        return result.Bytes(), nil
}

```

它连接到以类似于`localhost:1200`的形式给出的服务，读取 TCP 包，并将 ASN.1 内容解码成一个字符串，然后打印出来。

注意，这两者——客户机或服务器——都与上一章中基于文本的客户机和服务器不兼容。这个客户端和服务器正在交换 ASN.1 编码的数据值，而不是文本字符串。

## 数据

JSON 代表 JavaScript 对象符号。它被设计成一种在 JavaScript 系统之间传递数据的轻量级方法。它使用基于文本的格式，非常通用，已经成为许多编程语言的通用序列化方法。

JSON 序列化对象、数组和基本值。基本值包括字符串、数字、布尔值和空值。数组是以逗号分隔的值列表，可以表示数组、向量、列表或各种编程语言的序列。它们由方括号`[ ... ]`分隔。对象由花括号`{ ... }`中的“字段:值”对列表表示。

例如，前面给出的雇员表可以写成一个雇员对象数组:

```go
[
   {"Name": "fred", "Occupation": "programmer"},
   {"Name": "liping", "Occupation": "analyst"},
   {"Name": "sureerat", "Occupation": "manager"}
]

```

对于复杂的数据类型(如日期、数字类型之间没有区别、没有递归类型等)没有特殊的支持。JSON 是一种非常简单的语言，但是仍然非常有用。它基于文本的格式使它易于使用和调试，尽管它有字符串处理的开销。

根据 Go JSON 包规范，编组使用以下类型相关的默认编码:

*   布尔值编码为 JSON 布尔值。
*   浮点和整数值编码为 JSON 数字。
*   字符串值编码为 JSON 字符串，每个无效的 UTF-8 序列由 Unicode 替换字符 U+FFFD 的编码替换。
*   数组和切片值编码为 JSON 数组，除了`[]byte`编码为 Base64 编码的字符串。
*   结构值编码为 JSON 对象。每个结构字段都成为对象的成员。默认情况下，对象的键名是转换为小写的结构字段名。如果 struct 字段有标记，则该标记将被用作名称。
*   映射值编码为 JSON 对象。映射的键类型必须是字符串；对象键直接用作贴图键。
*   指针值编码为所指向的值。(注意:这允许树，但不允许图！).nil 指针编码为空 JSON 对象。
*   接口值编码为接口中包含的值。nil 接口值编码为空 JSON 对象。
*   通道、复杂和函数值不能在 JSON 中编码。试图对这样的值进行编码会导致`Marshal`返回`InvalidTypeError`。
*   JSON 不能表示循环数据结构，并且`Marshal`不处理它们。将循环结构传递给`Marshal`将导致无限递归。

将 JSON 序列化数据存储到文件`person.json`中的程序是`SaveJSON.go`:

```go
/* SaveJSON
 */

package main

import (
        "encoding/json"
        "fmt"
        "os"
)

type Person struct {
        Name  Name
        Email []Email
}

type Name struct {
        Family   string
        Personal string
}

type Email struct {
        Kind    string
        Address string
}

func main() {
        person := Person{
                Name: Name{Family: "Newmarch", Personal: "Jan"},
                Email: []Email{Email{Kind: "home", Address: "jan@newmarch.name"},
                        Email{Kind: "work", Address: "j.newmarch@boxhill.edu.au"}}}

        saveJSON("person.json", person)
}

func saveJSON(fileName string, key interface{}) {
        outFile, err := os.Create(fileName)
        checkError(err)
        encoder := json.NewEncoder(outFile)
        err = encoder.Encode(key)
        checkError(err)
        outFile.Close()
}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }

}

```

要将其加载回内存，请使用`LoadJSON.go`:

```go
/* LoadJSON
 */

package main

import (
        "encoding/json"
        "fmt"
        "os"
)

type Person struct {
        Name  Name
        Email []Email
}

type Name struct {
        Family   string
        Personal string
}

type Email struct {
        Kind    string
        Address string
}

func (p Person) String() string {
        s := p.Name.Personal + " " + p.Name.Family
        for _, v := range p.Email {
                s += "\n" + v.Kind + ": " + v.Address
        }
        return s
}
func main() {
        var person Person
        loadJSON("person.json", &person)

        fmt.Println("Person", person.String())
}

func loadJSON(fileName string, key interface{}) {
        inFile, err := os.Open(fileName)
        checkError(err)
        decoder := json.NewDecoder(inFile)
        err = decoder.Decode(key)
        checkError(err)
        inFile.Close()
}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)

        }
}

```

序列化形式(格式良好):

```go
{"Name":{"Family":"Newmarch",
         "Personal":"Jan"},
 "Email":[{"Kind":"home","Address":"jan@newmarch.name"},
          {"Kind":"work","Address":"j.newmarch@boxhill.edu.au"}
         ]
}

```

### 客户端和服务器

一个客户端发送一个人的数据并读回 10 次是`JSONEchoClient.go`:

```go
/* JSON EchoClient
 */
package main

import (
        "bytes"
        "encoding/json"
        "fmt"
        "io"
        "net"
        "os"
)

type Person struct {
        Name  Name
        Email []Email
}

type Name struct {
        Family   string
        Personal string
}

type Email struct {
        Kind    string
        Address string

}

func (p Person) String() string {
        s := p.Name.Personal + " " + p.Name.Family
        for _, v := range p.Email {
                s += "\n" + v.Kind + ": " + v.Address
        }
        return s
}

func main() {
        person := Person{
                Name: Name{Family: "Newmarch", Personal: "Jan"},
                Email: []Email{Email{Kind: "home", Address: "jan@newmarch.name"},
                        Email{Kind: "work", Address: "j.newmarch@boxhill.edu.au"}}}

        if len(os.Args) != 2 {
                fmt.Println("Usage: ", os.Args[0], "host:port")
                os.Exit(1)
        }
        service := os.Args[1]

        conn, err := net.Dial("tcp", service)
        checkError(err)

        encoder := json.NewEncoder(conn)
        decoder := json.NewDecoder(conn)

        for n := 0; n < 10; n++ {
                encoder.Encode(person)
                var newPerson Person
                decoder.Decode(&newPerson)
                fmt.Println(newPerson.String())
        }

        os.Exit(0)
}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

func readFully(conn net.Conn) ([]byte, error) {
        defer conn.Close()

        result := bytes.NewBuffer(nil)
        var buf [512]byte
        for {
                n, err := conn.Read(buf[0:])
                result.Write(buf[0:n])
                if err != nil {
                        if err == io.EOF {
                                break
                        }
                        return nil, err

                }
        }
        return result.Bytes(), nil
}

```

对应的服务器是`JSONEchoServer.go`:

```go
/* JSON EchoServer
 */
package main

import (
        "encoding/json"
        "fmt"
        "net"
        "os"
)

type Person struct {
        Name  Name
        Email []Email
}

type Name struct {
        Family   string
        Personal string
}

type Email struct {
        Kind    string
        Address string
}

func (p Person) String() string {
        s := p.Name.Personal + " " + p.Name.Family
        for _, v := range p.Email {
                s += "\n" + v.Kind + ": " + v.Address
        }
        return s
}

func main() {

        service := "0.0.0.0:1200"
        tcpAddr, err := net.ResolveTCPAddr("tcp", service)
        checkError(err)

        listener, err := net.ListenTCP("tcp", tcpAddr)
        checkError(err)

        for {
                conn, err := listener.Accept()
                if err != nil {
                        continue
                }

                encoder := json.NewEncoder(conn)
                decoder := json.NewDecoder(conn)

                for n := 0; n < 10; n++ {
                        var person Person
                        decoder.Decode(&person)
                        fmt.Println(person.String())
                        encoder.Encode(person)
                }
                conn.Close() // we're finished
        }
}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

## Gob 包

Gob 是一种特定于 Go 的序列化技术。它是专门为编码 Go 数据类型而设计的，目前还没有其他语言的实质性支持。它支持除通道、函数和接口之外的所有 Go 数据类型。它支持所有类型和大小的整数、字符串和布尔值、结构、数组和切片。目前，它在环形结构方面存在一些问题，但随着时间的推移会有所改善。

Gob 将类型信息编码成它的序列化形式。这比 X.509 序列化中的类型信息要广泛得多，但比 XML 文档中包含的类型信息要有效得多。对于每一段数据，类型信息只包括一次，但包括例如结构字段的名称。

这种类型信息的包含使得 Gob 编组和解组对于编组器和解组器之间的变化或差异相当健壮。例如，此结构:

```go
 struct T {
     a int
     b int
}

```

可以被编组，然后解组到不同的结构中，其中字段的顺序已经改变:

```go
 struct T {
     b int
     a int
}

```

它还可以处理丢失的字段(值被忽略)或额外的字段(字段保持不变)。它可以处理指针类型，因此前面的结构可以被解组到这个结构中:

```go
 struct T {
     *a int
     **b int
}

```

在某种程度上，它可以处理类型强制，使`int`字段可以扩展为`int64`，但不能处理不兼容的类型，如`int`和`uint`。

要使用 Gob 编组数据值，首先需要创建一个`Encoder`。它将一个`Writer`作为参数，并将对这个写流进行编组。编码器有一个名为`Encode`的方法，它将值封送到流中。可以对多段数据多次调用此方法。但是，每种数据类型的类型信息只写一次。

您使用一个`Decoder`来解组序列化的数据流。这需要一个`Reader`，每次读取都返回一个解组的数据值。

将 Gob 序列化数据存储到文件`person.go`中的程序是`SaveGob.go`:

```go
/* SaveGob
 */

package main

import (
        "encoding/gob"
        "fmt"
        "os"
)

type Person struct {
        Name  Name
        Email []Email
}

type Name struct {
        Family   string
        Personal string
}

type Email struct {
        Kind    string
        Address string

}

func main() {
        person := Person{
                Name: Name{Family: "Newmarch", Personal: "Jan"},
                Email: []Email{Email{Kind: "home", Address: "jan@newmarch.name"},
                        Email{Kind: "work", Address: "j.newmarch@boxhill.edu.au"}}}

        saveGob("person.gob", person)
}

func saveGob(fileName string, key interface{}) {
        outFile, err := os.Create(fileName)
        checkError(err)
        encoder := gob.NewEncoder(outFile)
        err = encoder.Encode(key)
        checkError(err)
        outFile.Close()
}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

要将其加载回内存，请使用`LoadGob.go`:

```go
/* LoadGob
 */

package main

import (
        "encoding/gob"
        "fmt"
        "os"
)

type Person struct {
        Name  Name
        Email []Email
}

type Name struct {
        Family   string
        Personal string
}

type Email struct {
        Kind    string
        Address string
}

func (p Person) String() string {
        s := p.Name.Personal + " " + p.Name.Family
        for _, v := range p.Email {
                s += "\n" + v.Kind + ": " + v.Address
        }
        return s
}
func main() {
        var person Person

        loadGob("person.gob", &person)

        fmt.Println("Person", person.String())
}

func loadGob(fileName string, key interface{}) {
        inFile, err := os.Open(fileName)
        checkError(err)
        decoder := gob.NewDecoder(inFile)
        err = decoder.Decode(key)
        checkError(err)
        inFile.Close()
}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

### 客户端和服务器

一个客户端发送一个人的数据并读回 10 次是`GobEchoClient.go`:

```go
/* Gob EchoClient
 */
package main

import (
        "bytes"
        "encoding/gob"
        "fmt"
        "io"
        "net"
        "os"
)

type Person struct {
        Name  Name
        Email []Email
}

type Name struct {
        Family   string
        Personal string
}

type Email struct {
        Kind    string
        Address string

}

func (p Person) String() string {
        s := p.Name.Personal + " " + p.Name.Family
        for _, v := range p.Email {
                s += "\n" + v.Kind + ": " + v.Address
        }
        return s
}

func main() {
        person := Person{
                Name: Name{Family: "Newmarch", Personal: "Jan"},
                Email: []Email{Email{Kind: "home", Address: "jan@newmarch.name"},
                        Email{Kind: "work", Address: "j.newmarch@boxhill.edu.au"}}}

        if len(os.Args) != 2 {
                fmt.Println("Usage: ", os.Args[0], "host:port")
                os.Exit(1)
        }
        service := os.Args[1]

        conn, err := net.Dial("tcp", service)
        checkError(err)

        encoder := gob.NewEncoder(conn)
        decoder := gob.NewDecoder(conn)

        for n := 0; n < 10; n++ {
                encoder.Encode(person)
                var newPerson Person
                decoder.Decode(&newPerson)
                fmt.Println(newPerson.String())
        }

        os.Exit(0)
}

func checkError(err error) {

        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

func readFully(conn net.Conn) ([]byte, error) {
        defer conn.Close()

        result := bytes.NewBuffer(nil)
        var buf [512]byte
        for {
                n, err := conn.Read(buf[0:])
                result.Write(buf[0:n])
                if err != nil {
                        if err == io.EOF {
                                break
                        }
                        return nil, err
                }
        }
        return result.Bytes(), nil
}

```

对应的服务器是`GobEchoServer.go`:

```go
/* Gob EchoServer
 */
package main

import (
        "encoding/gob"
        "fmt"
        "net"
        "os"
)

type Person struct {
        Name  Name
        Email []Email
}

type Name struct {
        Family   string
        Personal string
}

type Email struct {
        Kind    string
        Address string

}

func (p Person) String() string {
        s := p.Name.Personal + " " + p.Name.Family
        for _, v := range p.Email {
                s += "\n" + v.Kind + ": " + v.Address
        }
        return s
}

func main() {

        service := "0.0.0.0:1200"
        tcpAddr, err := net.ResolveTCPAddr("tcp", service)
        checkError(err)

        listener, err := net.ListenTCP("tcp", tcpAddr)
        checkError(err)

        for {
                conn, err := listener.Accept()
                if err != nil {
                        continue
                }

                encoder := gob.NewEncoder(conn)
                decoder := gob.NewDecoder(conn)

                for n := 0; n < 10; n++ {
                        var person Person
                        decoder.Decode(&person)
                        fmt.Println(person.String())
                        encoder.Encode(person)
                }
                conn.Close() // we're finished
        }
}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

## 将二进制数据编码为字符串

从前，传输 8 位数据是有问题的。它通常通过嘈杂的串行线路传输，很容易被破坏。另一方面，7 位数据可以更可靠地传输，因为第 8 位可以用作校验位。例如，在“偶数奇偶校验”方案中，校验位将被设置为 1 或 0，以使一个字节中的 1 为偶数。这允许检测每个字节中单个位的错误。

ASCII 是 7 位字符集。已经开发了许多比简单的奇偶校验更复杂的方案，但是这些方案涉及将 8 位二进制数据转换成 7 位 ASCII 格式。本质上，8 位数据以某种方式扩展到 7 位字节。

在 HTTP 响应和请求中传输的二进制数据通常被转换成 ASCII 形式。这使得用一个简单的文本阅读器检查 HTTP 消息变得很容易，而不用担心奇怪的 8 位字节会对您的显示产生什么影响！

一种常见的格式是 Base64。Go 支持许多二进制到文本的格式，包括 Base64。

Base64 编码和解码有两个主要函数:

```go
func NewEncoder(enc *Encoding, w io.Writer) io.WriteCloser
func NewDecoder(enc *Encoding, r io.Reader) io.Reader

```

一个简单的编码和解码八个二进制数字的程序是:

```go
/**
 * Base64
 */

package main

import (
        "encoding/base64"
        "fmt"
)

func main() {

        eightBitData := []byte{1, 2, 3, 4, 5, 6, 7, 8}

        enc := base64.StdEncoding.EncodeToString(eightBitData)
        dec, _ := base64.StdEncoding.DecodeString(enc)

        fmt.Println("Original data ", eightBitData)
        fmt.Println("Encoded string ", enc)
        fmt.Println("Decoded data ", dec)
}

```

输出如下所示:

```go
Original data  [1 2 3 4 5 6 7 8]
Encoded string  AQIDBAUGBwg=
Decoded data  [1 2 3 4 5 6 7 8]

```

## 协议缓冲区

到目前为止，考虑的序列化方法分为多种类型:

*   ASN.1 使用数据中的二进制标签对不同类型进行编码。从这个意义上说，ASN.1 编码的数据结构是一种自描述结构。
*   JSON 同样是一种自描述格式，使用 JavaScript 数据结构的规则:列表、字典等。
*   Gob 同样将类型信息编码成它的编码形式。这比 JSON 格式要详细得多。

另一类序列化技术依赖于要编码的数据类型的外部规范。有几个主要的，比如 ONC RPC 使用的编码。

ONC RPC 是一种旧的编码，面向 C 语言。最近的一个来自谷歌，被称为协议缓冲区。Go 标准库中不支持这一点，但 Google Protocol Buffers 开发小组( [`https://developers.google.com/protocol-buffers/`](https://developers.google.com/protocol-buffers/) )支持这一点，而且这一点在 Google 内部显然非常流行。出于这个原因，我们包括了一个关于协议缓冲区的部分，尽管在本书的其余部分我们通常处理 Go 标准库。

协议缓冲区是数据的二进制编码，旨在支持多种语言的数据类型。它们依赖于数据结构的外部规范，该规范用于编码数据(用源语言)以及将编码的数据解码回目标语言。(注:协议缓冲区于 2016 年 7 月过渡到版本 3。它与版本 2 不兼容。版本 2 将会被长期支持，但最终会被淘汰。参见 [`https://github.com/google/protobuf/releases/tag/v3.0.0`](https://github.com/google/protobuf/releases/tag/v3.0.0) 的协议缓冲区 3.0.0 版)。

要序列化的数据结构称为消息。每个消息中支持的数据类型包括:

*   数字(整数或浮点数)
*   布尔运算
*   弦乐(UTF 语-8)
*   原始字节
*   地图
*   其他消息，允许构建复杂的数据结构

消息的所有字段都是可选的(这是从`proto2`开始的变化，在那里字段是必需的或可选的)。字段可以代表关键字 repeated 的列表或数组，也可以代表使用关键字 map 的映射。每个字段都有一个类型，后跟一个名称，再后跟一个标记索引值。完整的语言指南称为“协议缓冲区语言指南”(参见 [`https://developers.google.com/protocol-buffers/docs/proto`](https://developers.google.com/protocol-buffers/docs/proto) )。

消息的定义与可能的目标语言无关。协议缓冲区版本 3 的语法中的`Person`类型的一个版本是`personv3.proto`。请注意，该文件包含每种类型的特定标记(1，2)。

```go
syntax = "proto3";
package person;

message Person {
        message Name {
                string family = 1;
                string personal = 2;
        }

        message Email {
                string kind = 1;
                string address = 2;
        }

        Name  name = 1;
        repeated Email email = 2;
}

```

### 安装和编译协议缓冲区

使用名为`protoc`的程序编译协议缓冲区。这不太可能安装在您的系统上。版本 3 是 2016 年 7 月才发布的，所以存储库中的副本很可能是版本 2。

从协议缓冲区 v3.0.0 页面安装最新版本。以 64 位 Linux 为例，从 GitHub 下载`protoc-3.0.0-linux-x86_64.zip`并解压到合适的地方(它包括二进制`bin/protoc`，应该放在你的`PATH`的某个地方)。

安装通用二进制文件。您还需要“后端”来生成 Go 文件。为此，从 GitHub 获取它:

```go
go get -u github.com/golang/protobuf/protoc-gen-go      

```

您几乎已经准备好编译一个`.proto`文件了。前面的例子`personv3.proto`声明了包`person`。在你的`GOPATH`中，你应该有一个名为`src`的目录。创建一个名为`src/person`的子目录。然后编译`personv3.proto`如下:

```go
protoc --go_out=src/person personv3.proto

```

这应该会创建`src/person/personv3.pb.go`文件。

### 编译后的 personv3.pb.go 文件

编译后的文件将声明许多类型和这些类型上的方法。这些类型如下:

```go
type Person struct {
        Name  *Person_Name    `protobuf:"bytes,1,opt,name=name" json:"name,omitempty"`
        Email []*Person_Email `protobuf:"bytes,2,rep,name=email" json:"email,omitempty"`
}

type Person_Name struct {
        Family   string `protobuf:"bytes,1,opt,name=family" json:"family,omitempty"`
        Personal string `protobuf:"bytes,2,opt,name=personal" json:"personal,omitempty"`
}

type Person_Email struct {
        Kind    string `protobuf:"bytes,1,opt,name=kind" json:"kind,omitempty"`
        Address string `protobuf:"bytes,2,opt,name=address" json:"address,omitempty"`
}

```

它们在名为`person`的包中。(注意:字符串等简单类型直接编码。在协议缓冲区 v2 中，使用了指针。对于复合类型，需要一个指针，如在 v2 中。)

### 使用编译的代码

JSON 示例中使用的代码和这个示例中使用的代码基本上没有区别，除了必须监视所用结构的指针。一个简单的程序就是`ProtocolBuffer.go`来编组和解组一个`Person`:

编组前后的输出应该是一个`Person`，并且应该是相同的:

```go
/* ProtocolBuffer
 */

package main

import (
        "fmt"
        "github.com/golang/protobuf/proto"
        "os"
        "person"

)

func main() {
        name := person.Person_Name{
                Family:   "newmarch",
                Personal: "jan"}

        email1 := person.Person_Email{
                Kind:    "home",
                Address: "jan@newmarch.name"}
        email2 := person.Person_Email{
                Kind:    "work",
                Address: "j.newmarch@boxhill.edu.au"}

        emails := []*person.Person_Email{&email1, &email2}
        p := person.Person{
                Name:  &name,
                Email: emails,
        }
        fmt.Println(p)

        data, err := proto.Marshal(&p)
        checkError(err)
        newP := person.Person{}
        err = proto.Unmarshal(data, &newP)
        checkError(err)
        fmt.Println(newP)
}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

编组前后的输出应该是一个人，通过运行以下命令应该是相同的:

```go
   go run ProtocolBuffer.go

{family:"newmarch" personal:"jan"  [kind:"home" address:"jan@newmarch.name"  kind:"work" address:"j.newmarch@boxhill.edu.au" ]}
{family:"newmarch" personal:"jan"  [kind:"home" address:"jan@newmarch.name"  kind:"work" address:"j.newmarch@boxhill.edu.au" ]}

```

我们还没有对编组对象做太多的工作。然而，它可以被保存到文件中或通过网络发送，并由任何支持的语言解组:C++、C#、Java、Python 以及 Go。

## 结论

本章讨论了序列化数据类型的一般属性，并展示了一些常见的编码。还有很多，包括 XML(包含在 Go 库中)、CBOR(JSON 的二进制形式)和 YAML(类似于 XML)，以及许多特定于语言的语言，如 Java 对象序列化和 Python 的 Pickle。那些不在 Go 标准包中的可能经常在 GitHub 上找到。

Footnotes [1](#Fn1_source)

我将序列化和编组视为同义词。关于这一点有各种各样的观点，有些观点比其他观点更具体。例如，参见“[序列化和封送处理的区别是什么？](http://stackoverflow.com/questions/770474/what-is-the-difference-between-serialization-and-marshaling)