# 八、HTTP

万维网是一个主要的分布式系统，拥有数百万用户。通过运行 HTTP 服务器，站点可以成为 web 主机。虽然 web 客户端通常是拥有浏览器的用户，但是还有许多其他的“用户代理”，例如 web 蜘蛛、web 应用程序客户端等等。

Web 建立在 HTTP(超文本传输协议)之上，而 HTTP 又是 TCP 之上的一层。HTTP 已经经历了四个公开的版本。版本 1.1(第三个版本)是最常用的，但是预计将会快速过渡到 HTTP/2，并且这现在占当前流量的 10%以上。

本章是 HTTP 的概述，接着是管理 HTTP 连接的 Go APIs。

## URL 和资源

URL 指定资源的位置。资源通常是静态文件，如 HTML 文档、图像或声音文件。但越来越多的是，它可能是一个动态生成的对象，可能基于存储在数据库中的信息。

当用户代理请求资源时，返回的不是资源本身，而是该资源的某种表示。例如，如果资源是一个静态文件，那么发送给用户代理的就是该文件的副本。

多个 URL 可能指向同一个资源，HTTP 服务器将为每个 URL 返回适当的资源表示。例如，一家公司可能使用同一产品的不同 URL 在内部和外部提供产品信息。产品的内部表示可能包括产品的内部联系人等信息，而外部表示可能包括销售产品的商店的位置。

这种资源视图意味着 HTTP 协议可以相当简单直接，而 HTTP 服务器可以任意复杂。HTTP 必须将来自用户代理的请求传递给服务器并返回一个字节流，而服务器可能必须对请求进行大量的处理。

### I18n

互联网的日益国际化带来了一些复杂的问题。主机名可能以国际化的形式给出，称为 IDN(国际化域名)。为了保持与不理解 Unicode 的遗留实现(如旧的电子邮件服务器)的兼容性，非 ASCII 域名被映射到称为 punycode 的 ASCII 表示中。例如，域名日本語。jp 具有 punycode 值`xn—wgv71a119e.jp`。从非 ASCII 域到 punycode 值的转换不是由 Go net 库自动执行的(从 Go 1.7 开始)，但是有一个名为`golang.org/x/net/idna`的扩展包可以在 Unicode 和它的 punycode 值之间进行转换。在“弄清 IDNA 一派胡言的故事”( [`https://github.com/golang/go/issues/13835`](https://github.com/golang/go/issues/13835) )上正在进行一场关于这个话题的讨论。

国际化域名开启了所谓的同形异义词攻击的可能性。许多 Unicode 字符具有相似的外观，例如俄语 o (U+043E)、希腊语 o (U+03BF)和英语 o (U+006F)。使用同形异义词的域名，如`google.com`(带有两个俄语 o，可能会引起混乱。已知有多种防御措施，比如总是显示 punycode(这里是`xn—ggle-55da.com`，使用 [Punycode 转换器](https://www.punycoder.com/))。

URI/URL 中的路径处理起来更复杂，因为它指的是相对于可能在特定本地化环境中运行的 HTTP 服务器的路径，编码可能不是 UTF-8，甚至不是 Unicode。IRI(国际化资源标识符)通过首先将任何本地化字符串转换为 UTF-8，然后对任何非 ASCII 字节进行百分比转义来管理这一点。名为“多语言网址介绍”( [`https://www.w3.org/International/articles/idn-and-iri/`](https://www.w3.org/International/articles/idn-and-iri/) `)`)的 W3C 页面有更多信息。从其他编码到 UTF-8 的转换在第 [6](06.html) 章中有介绍，而 Go 在`Queryescape/Queryunescape`的`net` /url 和`PathEscape/PathUnescape`的 Go 1.8 中有函数来做百分比转换。

### HTTP 特征

HTTP 是一种无状态、无连接、可靠的协议。在最简单的形式中，来自用户代理的每个请求都被可靠地处理，然后连接被中断。

在 HTTP 的最早版本中，每个请求都涉及一个单独的 TCP 连接，所以如果需要很多资源(比如嵌入在 HTML 页面中的图像)，那么就必须在很短的时间内建立和拆除很多 TCP 连接。

HTTP 1.1 在 HTTP 中加入了很多优化，在简单的结构上增加了复杂性，但却创造了更高效可靠的协议。为了进一步提高效率，HTTP/2 采用了二进制形式。

### 版本

HTTP 有四个版本:

*   0.9 版(1991 年):完全过时
*   版本 1.0 (1996):几乎过时
*   版本 1.1 (1999):目前最流行的版本
*   版本 2 (2015):最新版本

每个版本都必须理解早期版本的请求和响应。

### HTTP 0.9

请求格式:

```go
Request = Simple-Request

Simple-Request = "GET" SP Request-URI CRLF

```

#### 响应格式

响应的形式如下:

```go
Response = Simple-Response

Simple-Response = [Entity-Body]

```

### HTTP 1.0

这个版本为请求和响应添加了更多的信息。而不是“增长”0.9 格式，它只是留在新版本旁边。

#### 请求格式

从客户端到服务器的请求格式是:

```go
Request = Simple-Request | Full-Request

Simple-Request = "GET" SP Request-URI CRLF

Full-Request = Request-Line
                       *(General-Header
                          | Request-Header
                          | Entity-Header)
                      CRLF
                       [Entity-Body]

```

一个`Simple-Request`是一个 HTTP/0.9 请求，必须由一个`Simple-Response`回复。

一个`Request-Line`有这样的格式:

```go
Request-Line = Method SP Request-URI SP HTTP-Version CRLF

```

在哪里

```go
Method = "GET" | "HEAD" | POST |
         extension-method

```

这里有一个例子:

```go
GET http://jan.newmarch.name/index.html HTTP/1.0

```

#### 响应格式

响应的形式如下:

```go
Response = Simple-Response | Full-Response

Simple-Response = [Entity-Body]

Full-Response = Status-Line
                *(General-Header
                   | Response-Header
                   | Entity-Header)
                CRLF
                [Entity-Body]

```

`Status-Line`给出关于请求命运的信息:

```go
Status-Line = HTTP-Version SP Status-Code SP Reason-Phrase CRLF

```

这里有一个例子:

```go
HTTP/1.0 200 OK

```

状态行中的状态代码如下:

```go
Status-Code =     "200" ; OK
                | "201" ; Created
                | "202" ; Accepted
                | "204" ; No Content
                | "301" ; Moved permanently
                | "302" ; Moved temporarily
                | "304" ; Not modified
                | "400" ; Bad request
                | "401" ; Unauthorized
                | "403" ; Forbidden
                | "404" ; Not found
                | "500" ; Internal server error
                | "501" ; Not implemented
                | "502" ; Bad gateway
                | "503" | Service unavailable
                | extension-code

```

`General-` `Header`通常是日期，而`Response-Header`是位置、服务器或认证字段。

`Entity-` `Header`包含了关于`Entity-Body`的有用信息，可以跟随:

```go
Entity-Header = Allow
                | Content-Encoding
                | Content-Length
                | Content-Type
                | Expires
                | Last-Modified
                | extension-header

```

例如，(其中字段类型在`//`之后给出):

```go
HTTP/1.1 200 OK                             // status line
Date: Fri, 29 Aug 2003 00:59:56 GMT         // general header
Server: Apache/2.0.40 (Unix)                // response header
Content-Length: 1595                        // entity header

Content-Type: text/html; charset=ISO-8859-1 // entity header

```

### HTTP 1.1

HTTP 1.1 修复了 HTTP 1.0 的许多问题，但也因此变得更加复杂。这个版本是通过扩展或细化 HTTP 1.0 的可用选项来实现的。例如:

*   还有更多命令如`TRACE`和`CONNECT`
*   HTTP 1.1 收紧了请求 URL 的规则，以允许代理处理。如果请求是通过代理定向的，URL 应该是绝对 URL，如:

    ```go
              GET http://www.w3.org/index.html HTTP/1.1

    ```

    否则应该使用绝对路径，并且应该包括一个`Host`头字段，如:

    ```go
              GET /index.html HTTP/1.1
              Host: www.w3.org

    ```

*   有更多的属性，如`If-Modified-Since`，也供代理使用

这些变化包括

*   主机名标识(允许虚拟主机)
*   内容协商(多种语言)
*   持久连接(减少 TCP 开销；这个很复杂)
*   分块传输
*   字节范围(文档的请求部分)
*   代理支持

### HTTP/2

所有早期版本的 HTTP 都是基于文本的。HTTP/2 最大的不同在于它是一种二进制格式。为了确保向后兼容，这不能通过向旧服务器发送二进制消息来查看它做了什么来管理。取而代之的是发送一个带有额外属性的 HTTP 1.1 消息，本质上是询问服务器是否想切换到 HTTP/2。如果它不理解额外的字段，它会用一个普通的 HTTP 1.1 响应进行回复，会话继续使用 HTTP 1.1。

否则，服务器可以响应它愿意改变，并且会话可以使用 HTTP/2 继续。

0.9 协议用了一页。在大约 20 页中描述了 1.0 协议，并且包括 0.9 协议。1.1 协议需要 120 页，是对 1.0 的实质性扩展，而 HTTP/2 需要大约 96 页。HTTP/2 规范只是对 HTTP 1.1 规范的补充。

## 简单用户代理

浏览器等用户代理发出请求并得到响应。这涉及到 Go 类型和相关的方法调用。

### 响应类型

响应类型如下:

```go
type Response struct {
    Status     string // e.g. "200 OK"
    StatusCode int    // e.g. 200
    Proto      string // e.g. "HTTP/1.0"
    ProtoMajor int    // e.g. 1
    ProtoMinor int    // e.g. 0

    Header map[string][]string

    Body io.ReadCloser

    ContentLength int64

    TransferEncoding []string

    Close bool

    Trailer map[string][]string

    Request *Request // the original request

    TLS *tls.ConnectionState // info about the TLS connection or nil
}

```

### 头部方法

我们通过例子来检验这种数据结构。每个 HTTP 请求类型在`net/http`包中都有自己的 Go 函数。最简单的请求来自一个名为`HEAD`的用户代理，它请求关于资源及其 HTTP 服务器的信息。该功能可用于进行查询:

```go
func Head(url string) (r *Response, err error)

```

响应的状态在响应字段`Status`中，而字段`Header`是 HTTP 响应中报头字段的映射。一个名为`Head.go`的程序发出这个请求并显示结果如下:

```go
/* Head
 */

package main

import (
        "fmt"
        "net/http"
        "os"
)

func main() {
        if len(os.Args) != 2 {
                fmt.Println("Usage: ", os.Args[0], "host:port")
                os.Exit(1)
        }
        url := os.Args[1]

        response, err := http.Head(url)
        if err != nil {
                fmt.Println(err.Error())
                os.Exit(2)
        }

        fmt.Println(response.Status)
        for k, v := range response.Header {
                fmt.Println(k+":", v)
        }

        os.Exit(0)
}

```

针对资源运行时，如下所示:

```go
go run Head.go http://www.golang.com/

```

它会打印出这样的内容:

```go
200 OK
Date: [Fri, 06 Jan 2017 11:20:37 GMT]
Server: [Google Frontend]
Content-Length: [7902]
Alt-Svc: [quic=":443"; ma=2592000; v="35,34"]
Strict-Transport-Security: [max-age=31536000; preload]
Content-Type: [text/html; charset=utf-8]
X-Cloud-Trace-Context: [6e28ebc86bb1026ae7b784c891d0117c]

```

响应来自我们控制之外的服务器，它可能会在途中经过其他服务器。显示的字段可能会有所不同，当然字段的值也会有所不同。

### GET 方法

通常，我们想要检索一个资源的表示，而不仅仅是获取关于它的信息。`GET`请求将做到这一点，并且可以使用以下方法来完成:

```go
func Get(url string) (r *Response, finalURL string, err error)

```

响应的内容在类型为`io.ReadCloser`的响应字段`Body`中。我们可以用程序`Get.go`将内容打印到屏幕上:

```go
/* Get
 */

package main

import (
        "fmt"
        "net/http"
        "net/http/httputil"
        "os"
        "strings

"
)

func main() {
        if len(os.Args) != 2 {
                fmt.Println("Usage: ", os.Args[0], "host:port")
                os.Exit(1)
        }
        url := os.Args[1]

        response, err := http.Get(url)
        if err != nil {
                fmt.Println(err.Error())
                os.Exit(2)
        }

        if response.Status != "200 OK" {
                fmt.Println(response.Status)
                os.Exit(2)
        }

        fmt.Println("The response header is")
        b, _ := httputil.DumpResponse(response, false)
        fmt.Print(string(b))

        contentTypes := response.Header["Content-Type"]
        if !acceptableCharset(contentTypes) {Arial
                fmt.Println("Cannot handle", contentTypes)
                os.Exit(4)
        }

        fmt.Println("The response body is")
        var buf [512]byte
        reader := response.Body
        for {
                n, err := reader.Read(buf[0:])
                if err != nil {
                        os.Exit(0)
                }
                fmt.Print(string(buf[0:n]))
        }
        os.Exit(0)
}

func acceptableCharset(contentTypes []string) bool {
        // each type is like [text/html; charset=utf-8]
        // we want the UTF-8 only
        for _, cType := range contentTypes {
                if strings.Index(cType, "utf-8") != -1 {
                        return true
                }
        }
        return false

}

```

[`http://www.golang.com`](http://www.golang.com) 运行时为:

```go
go run Get.go  http://www.golang.com

```

响应标头是:

```go
HTTP/2.0 200 OK
Content-Length: 7902
Alt-Svc: quic=":443"; ma=2592000; v="35,34"
Content-Type: text/html; charset=utf-8
Date: Fri, 06 Jan 2017 11:29:12 GMT
Server: Google Frontend
Strict-Transport-Security: max-age=31536000; preload
X-Cloud-Trace-Context: ea9b41b4796f379af487388b1474ed4e

```

响应正文是:

```go
<!DOCTYPE html>
<html>
<head>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8">
<meta name="viewport" content="width=device-width, initial-scale=1">
<meta name="theme-color" content="#375EAB">

  <title>The Go Programming Language</title>
...

```

(注意，这是通过 HTTP/2 发送的。Go 库已经为你进行了版本协商。)

请注意，有一些重要的字符集问题，就像上一章讨论的那样。服务器将使用某种字符集编码，可能还有某种传输编码来传递内容。通常这是用户代理和服务器之间的协商问题，但是我们使用的简单的`GET`命令不包括协商的用户代理部分。因此服务器可以发送它想要的任何字符编码。

第一次写作的时候，我在中国(可以访问谷歌)。当我在 [`www.google.com`](http://www.google.com) 上试用这个程序时，谷歌的服务器试图通过猜测我的位置并向我发送中文字符集 Big5 的文本来帮助我。如何告诉服务器什么样的字符编码对我来说是可以的将在后面讨论。

## 配置 HTTP 请求

Go 还为用户代理提供了一个与 HTTP 服务器通信的底层接口。如您所料，它不仅让您对客户机请求有更多的控制权，而且还要求您在构建请求时花费更多的精力。然而，复杂度只有很小的增加。

用于构建请求的数据类型是类型`Request`。这是一个复杂的类型，我们现在只显示主要字段。省略了几个字段和完整的 Go 文档。

```go
type Request struct {
    Method     string     // GET, POST, PUT, etc.
    URL        *url.URL   // Parsed URL.
    Proto      string // "HTTP/1.0"
    ProtoMajor int    // 1
    ProtoMinor int    // 0

    // A header maps request lines to their values.
    Header Header  // map[string][]string

    // The message body.
    Body io.ReadCloser

    // ContentLength records the length of the associated content.
    // The value -1 indicates that the length is unknown.
    // Values >= 0 indicate that the given number of bytes may be read from Body.
    ContentLength int64

    // TransferEncoding lists the transfer encodings from outermost to innermost.
    // An empty list denotes the "identity" encoding.
    TransferEncoding []string

    // The host on which the URL is sought.
    // Per RFC 2616, this is either the value of the Host: header
    // or the host name given in the URL itself.
    Host string
}

```

请求中可以存储许多信息。您不需要填写所有字段，只需填写感兴趣的字段。创建具有默认值的请求的最简单方法是使用，例如:

```go
request, err := http.NewRequest("GET", url.String(), nil)

```

创建请求后，您可以修改字段。例如，要指定您只想接收 UTF-8，向请求添加一个`Accept-Charset`字段，如下所示:

```go
request.Header.Add("Accept-Charset", "UTF-8;q=1, ISO-8859-1;q=0")

```

(请注意，除非列表中明确提到，否则默认设置 ISO-8859-1 的值始终为 1。HTTP 1.1 规范可以追溯到 1999 年！)

客户端设置一个`charset`请求很简单。但是对于服务器的字符集返回值会发生什么情况，还有些困惑。返回的资源应该有一个指定内容媒体类型的`Content-Type`，比如`text/html`。如果合适，媒体类型应该说明字符集，例如`text/html; charset=UTF-8`。如果没有字符集规范，那么根据 HTTP 规范，它应该被视为默认的 ISO8859-1 字符集。但是 HTML4 规范声明，由于很多服务器不符合这一点，所以你不能做任何假设。

如果在服务器的`Content-Type`中有指定的字符集，那么假设它是正确的。如果没有指定，因为超过 50%的页面是 UTF-8，有些是 ASCII，它是安全的假设 UTF-8。少于 10%的页面可能是错误的:-(。

## 客户端对象

要向服务器发送请求并获得回复，便利对象`Client`是最简单的方法。这个对象可以管理多个请求，并将处理诸如服务器是否保持 TCP 连接活动等问题。

这在下面的程序`ClientGet.go`中有说明。

该程序显示了如何添加 HTTP 头，因为我们添加头`Accept-Charset`只接受 UTF-8。这里有一个小问题，是由 Go 中的一个 bug 引起的，这个 bug 只在 Go 1.8 中得到修复。如果得到 301、302、303 或 307 响应，`Client.Do`函数将自动进行重定向。在 Go 1.8 之前，它不会在这个重定向中跨 HTTP 头进行复制。

如果你尝试访问一个像 [`http://www.google.com`](http://www.google.com) 这样的站点，那么它会重定向到一个像 [`http://www.google.com.au`](http://www.google.com.au) 这样的站点，但是会丢失`Accept-Charset`头并返回 ISO8859-1(根据 1999 HTTP 1.1 规范它应该这样做！).附带条件是，该程序在 Go 1.8 之前的版本中可能不会给出正确的结果，该程序如下所示:

```go
/* ClientGet
 */

package main

import (
        "fmt"
        "net/http"
        "net/http/httputil"
        "net/url"
        "os"
        "strings"
)

func

main() {
        if len(os.Args) != 2 {
                fmt.Println("Usage: ", os.Args[0], "http://host:port/page")
                os.Exit(1)
        }
        url, err := url.Parse(os.Args[1])
        checkError(err)

        client := &http.Client{}

        request, err := http.NewRequest("HEAD", url.String(), nil)

        // only accept UTF-8
        request.Header.Add("Accept-Charset", "utf-8;q=1, ISO-8859-1;q=0")
        checkError(err)

        response, err := client.Do(request)
        checkError(err)
        if response.Status != "200 OK" {
                fmt.Println(response.Status)
                os.Exit(2)
        }

        fmt.Println("The response header is")
        b, _ := httputil.DumpResponse(response, false)
        fmt.Print(string(b))

        chSet := getCharset(response)
        if chSet != "utf-8" {
                fmt.Println("Cannot handle", chSet)
                os.Exit(4)
        }

        var buf [512]byte
        reader := response.Body
        fmt.Println("got body")
        for {
                n, err := reader.Read(buf[0:])
                if err != nil {
                        os.Exit(0)
                }
                fmt.Print(string(buf[0:n]))
        }

        os.Exit(0)
}

func getCharset(response *http.Response) string {
        contentType := response.Header.Get("Content-Type")
        if contentType == "" {
                // guess
                return "utf-8"
        }
        idx := strings.Index(contentType, "charset=")
        if idx == -1 {
                // guess
                return "utf-8"
        }
        chSet := strings.Trim(contentType[idx+8:], " ")
        return strings.ToLower(chSet)
}

func

checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

程序运行如下，例如:

```go
go run ClientGet.go  http://www.golang.com

```

## 代理处理

现在 HTTP 请求通过特定的 HTTP 代理是很常见的。这是构成 TCP 连接并在应用层起作用的服务器的补充。公司使用代理来限制他们自己的员工可以看到的内容，而许多组织使用 Cloudflare 等代理服务来充当缓存，从而减少组织自己的服务器上的负载。通过代理访问网站需要客户端进行额外的处理。

### 简单代理

HTTP 1.1 阐述了 HTTP 应该如何通过代理工作。应该向代理发出一个`GET`请求。但是，请求的 URL 应该是目的地的完整 URL。此外，HTTP 头应该包含一个设置为代理的`Host`字段。只要代理被配置为传递这样的请求，那么这就是所有需要做的事情。

Go 认为这是 HTTP 传输层的一部分。为了管理这个，它有一个类`Transport`。这包含一个可以设置为返回代理 URL 的函数的字段。如果我们有一个 URL 作为代理的字符串，则会创建适当的传输对象，然后将其提供给一个客户端对象，如下所示:

```go
proxyURL, err := url.Parse(proxyString)
transport := &http.Transport{Proxy: http.ProxyURL(proxyURL)}
client := &http.Client{Transport: transport}

```

然后，客户端可以像以前一样继续。

以下程序`ProxyGet.go`说明了这一点

```go
/* ProxyGet
 */

package main

import (
        "fmt"
        "io"
        "net/http"
        "net/http/httputil"
        "net/url"
        "os"
)

func main() {
        if len(os.Args) != 3 {
                fmt.Println("Usage: ", os.Args[0], "http://proxy-host:port http://host:port/page")
                os.Exit(1)
        }
        proxyString := os.Args[1]
        proxyURL, err := url.Parse(proxyString)
        checkError(err)
        rawURL := os.Args[2]
        url,err := url.Parse(rawURL)
        checkError(err)

        transport := &http.Transport{Proxy: http.ProxyURL(proxyURL)}
        client := &http.Client{Transport: transport}

        request, err := http.NewRequest("GET", url.String(), nil)

        urlp, _ := transport.Proxy(request)
        fmt.Println("Proxy ", urlp)
        dump, _ := httputil.DumpRequest(request, false)
        fmt.Println(string(dump))

        response, err := client.Do(request)

        checkError(err)
        fmt.Println("Read ok")

        if response.Status != "200 OK" {
                fmt.Println(response.Status)
                os.Exit(2)
        }
        fmt.Println("Response ok")

        var buf [512]byte
        reader := response.Body
        for {
                n, err := reader.Read(buf[0:])
                if err != nil {
                        os.Exit(0)
                }
                fmt.Print(string(buf[0:n]))
        }

        os.Exit(0)
}

func checkError(err error) {
        if err != nil {
                if err == io.EOF {
                        return
                }
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

如果您在端口 8080 上有一个代理，比如说`XYZ.com`,您可以按如下方式进行测试:

```go
go run ProxyGet.go http://XYZ.com:8080/ http://www.google.com

```

如果你没有合适的代理来测试这个，那么下载并安装 Squid 代理( [`http://www.squid-cache.org/`](http://www.squid-cache.org/) )到你自己的电脑上。

这个程序使用了一个已知的代理作为程序的参数。有许多方法可以让应用程序知道代理。大多数浏览器都有一个配置菜单，您可以在其中输入代理信息:这样的信息对于 Go 应用程序是不可用的。一些应用程序可能使用 Web 代理自动发现协议( [`https://en.wikipedia.org/wiki/Web_Proxy_Autodiscovery_Protocol`](https://en.wikipedia.org/wiki/Web_Proxy_Autodiscovery_Protocol) )从网络中某个通常称为`autoproxy.pac`的文件中获取代理信息。Go 还不知道如何解析这些 JavaScript 文件，所以不能使用它们。特定的操作系统可能具有指定代理的系统特定的方式。Go 无法访问这些。但是如果在操作系统环境变量如`HTTP_PROXY`或`http_proxy`中设置了代理信息，它可以使用该函数找到代理信息:

```go
func ProxyFromEnvironment(req *Request) (*url.URL, error)

```

如果您的程序运行在这样的环境中，您可以使用这个函数，而不必显式地知道代理参数。

### 认证代理

一些代理需要通过用户名和密码进行身份验证，以传递请求。一种常见的方案是“基本认证”，其中用户名和密码被连接成一个字符串`"user:password"`，然后进行 Base64 编码。然后由 HTTP 请求头“`Proxy-Authorization`”将它提供给代理，并标记它是基本认证

下面的程序`ProxyAuthGet.go`说明了这一点，它将`Proxy-Authentication`头添加到前面的代理程序中:

```go
/* ProxyAuthGet
 */

package main

import (
        "encoding/base64"
        "fmt"
        "io"
        "net/http"
        "net/http/httputil"
        "net/url"
        "os"
)

const auth = "jannewmarch:mypassword"

func main() {
        if len(os.Args) != 3 {
                fmt.Println("Usage: ", os.Args[0], "http://proxy-host:port http://host:port/page")
                os.Exit(1)
        }
        proxy := os.Args[1]
        proxyURL, err := url.Parse(proxy)
        checkError(err)
        rawURL := os.Args[2]
        url, err := url.Parse(rawURL)
        checkError(err)

        // encode the auth
        basic := "Basic " + base64.StdEncoding.EncodeToString([]byte(auth))

        transport := &http.Transport{Proxy: http.ProxyURL(proxyURL)}
        client := &http.Client{Transport: transport}

        request, err := http.NewRequest("GET", url.String(), nil)

        request.Header.Add("Proxy-Authorization", basic)
        dump, _ := httputil.DumpRequest(request, false)
        fmt.Println(string(dump))

        // send the request
        response, err := client.Do(request)

        checkError(err)
        fmt.Println("Read ok")

        if response.Status != "200 OK" {
                fmt.Println(response.Status)
                os.Exit(2)
        }
        fmt.Println("Response ok")

        var buf [512]byte
        reader := response.Body
        for {
                n, err := reader.Read(buf[0:])
                if err != nil {
                        os.Exit(0)
                }
                fmt.Print(string(buf[0:n]))
        }

        os.Exit(0)
}

func checkError(err error) {
        if err != nil {
                if err == io.EOF {
                        return
                }
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

这个项目似乎没有公开的测试场地。我在使用认证代理的工作中测试了它。设置这样的代理超出了本书的范围。有一个关于如何做到这一点的讨论叫做“如何建立一个具有基本用户名和密码认证的 Squid 代理”(见 [`http://stackoverflow.com/questions/3297196/how-to-set-up-a-squid-proxy-with-basic-username-and-password-authentication`](http://stackoverflow.com/questions/3297196/how-to-set-up-a-squid-proxy-with-basic-username-and-password-authentication) )。

## 客户的 HTTPS 连接

对于安全、加密的连接，HTTP 使用 TLS，这在第 7 章[中有所描述。HTTP+TLS 的协议被称为 HTTPS，它使用`https://`URL 而不是`http://`URL。](07.html)

在客户端接受来自服务器的数据之前，服务器需要返回有效的 X.509 证书。如果证书是有效的，那么 Go 将处理幕后的一切，并且之前给定的客户端可以正常运行 https URLs。也就是说，像前面的`ClientGet.go`这样的程序不变地运行——你只需给它们一个 HTTPS URL。

许多网站都有无效的证书。它们可能已经过期，它们可能是自签名的，而不是由公认的证书颁发机构签名的，或者它们可能只是有错误(如服务器名不正确)。像 Firefox 这样的浏览器放了一个大大的警告通知，上面写着“让我离开这里！”按钮，但你可以继续冒险，许多人都这样做。

Go 在遇到证书错误时会立即退出。但是，您可以将客户端配置为忽略证书错误。当然，这是不可取的——证书配置错误的站点可能会有其他问题。

在第 7 章中，我们生成了自签名的 X.509 证书。在本章的后面，我们将给出一个使用 X.509 证书的 HTTPS 服务器，如果使用了自签名证书，那么`ClientGet.go`将生成这个错误:

```go
 x509: certificate signed by unknown authority

```

客户端通过打开传输配置标志`InsecureSkipVerify`来移除这些错误并继续。不安全程序是`TLSUnsafeClientGet.go`:

```go
/* TLSUnsafeClientGet
 */

package main

import (
        "fmt"
        "net/http"
        "net/url"
        "os"
        "strings"
        "crypto/tls"
)

func main() {
        if len(os.Args) != 2 {
                fmt.Println("Usage: ", os.Args[0], "https://host:port/page")
                os.Exit(1)
        }
        url, err := url.Parse(os.Args[1])
        checkError(err)
        if url.Scheme != "https" {
              fmt.Println("Not https scheme ", url.Scheme)
              os.Exit(1)
        }

        transport := &http.Transport{}
        transport.TLSClientConfig = &tls.Config{InsecureSkipVerify: true}
        client := &http.Client{Transport: transport}

        request, err := http.NewRequest("GET", url.String(), nil)
        // only accept UTF-8
        checkError(err)

        response, err := client.Do(request)
        checkError(err)

        if response.Status != "200 OK" {
                fmt.Println(response.Status)
                os.Exit(2)
        }
        fmt.Println("get a response")

        chSet := getCharset(response)
        fmt.Printf("got charset %s\n", chSet)
        if chSet != "UTF-8" {
                fmt.Println("Cannot handle", chSet)
                os.Exit(4)
        }

        var buf [512]byte

        reader := response.Body
        fmt.Println("got body")
        for {
                n, err := reader.Read(buf[0:])
                if err != nil {
                        os.Exit(0)
                }
                fmt.Print(string(buf[0:n]))
        }

        os.Exit(0)
}

func getCharset(response *http.Response) string {
        contentType := response.Header.Get("Content-Type")
        if contentType == "" {
                // guess
                return "UTF-8"
        }
        idx := strings.Index(contentType, "charset:")
        if idx == -1 {
                // guess
                return "UTF-8"
        }
        return strings

.Trim(contentType[idx:], " ")
}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

## 服务器

构建客户机的另一面是处理 HTTP 请求的 web 服务器。最简单也是最早的服务器只是返回文件的副本。然而，在当前的服务器中，任何 URL 现在都可以触发任意计算。

### 文件服务器

我们从一个基本的文件服务器开始。Go 提供了一个多路复用器，即一个可以读取和解释请求的对象。它向在自己线程中运行的`handlers,`发出请求。因此，读取 HTTP 请求、对它们进行解码，以及在它们自己的线程中分支到合适的函数的大部分工作已经为我们完成了。

对于文件服务器，Go 也给出了一个`FileServer`对象，它知道如何从本地文件系统传递文件。它需要一个“根”目录，即本地系统中文件树的顶部，以及一个匹配 URL 的模式。最简单的模式是`/`，它是任何 URL 的顶部。这将匹配所有 URL。

考虑到这些对象，从本地文件系统传送文件的 HTTP 服务器几乎是令人尴尬的琐碎。是`FileServer.go`:

```go
/* File Server
 */

package main

import (
        "fmt"
        "net/http"
        "os"
)

func main() {
        // deliver files from the directory /var/www
        fileServer := http.FileServer(http.Dir("/var/www"))

        // register the handler and deliver requests to it
        err := http.ListenAndServe(":8000", fileServer)
        checkError(err)
        // That's it!
}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

服务器按如下方式运行:

```go
go run FileServer.go

```

这个服务器甚至发送`"` `404 not found` `"`消息来请求不存在的文件资源！如果请求的文件是一个目录，它返回一个包含在`<pre> ... </pre>`标签中的列表，没有其他 HTML 头或标记。如果使用 Wireshark 或简单的 telnet 客户端，目录以`text/html`的形式发送，HTML 文件以`text/html`的形式发送，Perl 文件以`text/x-perl`的形式发送，Java 文件以`text/x-java`的形式发送，等等。`FileServer`采用了一些类型识别，并将其包含在 HTTP 请求中，但是它不像 Apache 这样的服务器那样提供对标记的控制。

### 处理函数

在最后一个程序中，处理程序在第二个参数中给定给了`ListenAndServe`。可以通过调用`Handle`或`HandleFunc`首先注册任意数量的处理程序，签名如下:

```go
func Handle(pattern string, handler Handler)
func HandleFunc(pattern string, handler func(ResponseWriter, *Request))

```

`ListenAndServe`的第二个参数可以是`nil`，然后调用被分派给所有注册的处理程序。每个处理程序应该有不同的 URL 模式。例如，文件处理器可能有 URL 模式`/`，而函数处理器可能有 URL 模式`/cgi-bin`。更具体的模式优先于更一般的模式。

常见的 CGI 程序有`test-cgi`(写在 shell 中)和`printenv`(写在 Perl 中)，它们打印环境变量的值。可以编写一个处理程序，以类似于`PrintEnv.go`的方式工作。

```go
/* Print Env
 */

package main

import (
        "fmt"
        "net/http"
        "os"
)
Arial
func main() {
        // file handler for most files
        fileServer := http.FileServer(http.Dir("/var/www"))
        http.Handle("/", fileServer)

        // function handler for /cgi-bin/printenv
        http.HandleFunc("/cgi-bin/printenv", printEnv)

        // deliver requests to the handlers
        err := http.ListenAndServe(":8000", nil)
        checkError(err)
        // That's it!
}

func printEnv(writer http.ResponseWriter, req *http.Request) {
        env := os.Environ()
        writer.Write([]byte("<h1>Environment</h1>\n<pre>"))
        for _, v := range env {
                writer.Write([]byte(v + "\n"))
        }
        writer.Write([]byte("</pre>"))
}

func

checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

Note

为了简单起见，这个程序不提供格式良好的 HTML。它缺少 html、head 和 body 标签。在本地主机上运行程序并将浏览器指向`http://localhost/cgi-bin/printenv`会在我的电脑上产生如下输出:

```go
Environment

XDG_VTNR=7
XDG_SESSION_ID=c2
CLUTTER_IM_MODULE=xim
XDG_GREETER_DATA_DIR=/var/lib/lightdm-data/newmarch
SESSION=gnome-flashback-compiz
GPG_AGENT_INFO=/home/newmarch/.gnupg/S.gpg-agent:0:1
TERM=xterm-256color
SHELL=/bin/bash
...

```

在这个程序中使用`cgi-bin`目录有点厚脸皮:它不像 CGI 脚本那样调用外部程序。它只是调用 Go 函数`printEnv`。Go 确实有能力使用`os.ForkExec`调用外部程序，但是还不支持像 Apache 的`mod_perl`这样的动态链接模块。

### 旁路默认多路复用器

Go 服务器收到的 HTTP 请求通常由多路复用器处理，多路复用器检查 HTTP 请求中的路径并调用适当的文件处理程序等。您可以定义自己的处理程序。这些可以通过调用`http.HandleFunc`注册到默认的多路复用器中，它采用一个模式和一个函数。然后，像`ListenAndServe`这样的函数接受一个`nil`处理函数。这在上一个例子中已经完成了。

然而，如果您想接管多路复用器的角色，那么您可以给一个非`nil`函数作为`ListenAndServe`的处理函数。这个函数将负责管理请求和响应。

下面的例子很简单，但是说明了它的用法。多路复用器功能简单地为所有对`ServerHandler.go`的请求返回一个`"204 No content"`:

```go
/* ServerHandler
 */

package main

import (
        "net/http"
)

func main() {

        myHandler := http.HandlerFunc(func(rw http.ResponseWriter, request *http.Request) {
                // Just return no content - arbitrary headers can be set, arbitrary body
                rw.WriteHeader(http.StatusNoContent)
        })

        http.ListenAndServe(":8080", myHandler)
}

```

可以通过对服务器运行`telnet`来测试服务器，给出如下输出:

```go
$telnet localhost 8080
Trying 127.0.0.1...
Connected to localhost.
Escape character is '^]'.
GET / HTTP/1.0

HTTP/1.0 204 No Content
Date: Tue, 10 Jan 2017 05:32:53 GMT

```

或者通过使用这个:

```go
curl -v localhost:8080

```

要给出这个输出:

```go
* Rebuilt URL to: localhost:8080/
*   Trying 127.0.0.1...
* Connected to localhost (127.0.0.1) port 8080 (#0)
> GET / HTTP/1.1
> Host: localhost:8080
> User-Agent: curl/7.47.0
> Accept: */*
>
< HTTP/1.1 204 No Content
< Date: Wed, 08 Mar 2017 08:46:35 GMT
<
* Connection #0 to host localhost left intact

```

相反，可以构建任意复杂的行为。

## 安全超文本传输协议

对于安全、加密的连接，HTTP 使用 TLS，这在第 7 章[中有所描述。HTTP+TLS 的协议被称为 HTTPS，它使用`https://`URL 而不是`http://`URL。](07.html)

对于使用 HTTPS 的服务器，它需要一个 X.509 证书和该证书的私钥文件。Go 目前要求这些是在第 7 章[中使用的 PEM 编码。然后用 HTTPS (HTTP+TLS)函数`ListenAndServeTLS`替换 HTTP 函数`ListenAndServe`。](07.html)

前面给出的文件服务器程序可以写成一个 HTTPS 服务器为`HTTPSFileServer.go`:

```go
/* HTTPSFileServer
 */

package main

import (
        "fmt"
        "net/http"
        "os"
)

func main() {
        // deliver files from the directory /var/www
        fileServer := http.FileServer(http.Dir("/var/www"))

        // register the handler and deliver requests to it
        err := http.ListenAndServeTLS(":8000", "jan.newmarch.name.pem",
                "private.pem", fileServer)
        checkError(err)
        // That's it!
}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

例如，该服务器由`https://localhost:8000/index.html`访问。如果证书是自签名证书，则需要不安全的客户端来访问服务器内容。例如:

```go
curl -kv https://localhost:8000

```

如果您想要一个同时支持 HTTP 和 HTTPs 的服务器，那么在它自己的`go`例程中运行每个监听器。

## 结论

Go 对 HTTP 有广泛的支持。这并不奇怪，因为 Go 的发明部分是为了满足谷歌对自己服务器的需求。本章讨论了 Go 对 HTTP 和 HTTPS 的各种级别的支持。