# 十、完整的网络服务器

这一章主要是对 HTTP 一章的说明，在 Go 中构建一个完整的 Web 服务器。它还展示了如何使用模板，以便在文本文件中使用表达式来插入变量值和生成重复的部分。它处理序列化数据和 Unicode 字符集。本章中的程序足够长且复杂，所以它们并不总是完整地给出，但可以从本书的网站上下载，该网站是 [`http://www.apress.com/9781484226919`](http://www.apress.com/9781484226919) 。

我正在学习中文。相反，经过多年的努力，我仍在尝试学习中文。当然，我没有埋头苦干，而是尝试了各种技术辅助手段。我尝试了教科书、视频和许多其他教具。最终我意识到我进步缓慢的原因是没有一个好的中文抽认卡的计算机程序，所以为了学习，我需要建立一个。

我在 Python 中找到了一个程序来完成一些任务。但遗憾的是，它写得不好，在几次试图把它颠倒过来之后，我得出的结论是，最好从零开始。当然，一个网络解决方案要比一个独立的解决方案好得多，因为这样我的中文课上的其他人就可以分享它，还有其他的学习者。当然，服务器应该是用 Go 编写的。

我使用了张芃芃《汉语口语精读》一书中的词汇，但是这个程序适用于任何词汇集。

## 浏览器站点图

在浏览器中看到的结果程序有三种类型的页面，如图 [10-1](#Fig1) 所示。

![A436770_1_En_10_Fig1_HTML.gif](img/A436770_1_En_10_Fig1_HTML.gif)

图 10-1。

Browser pages

主页显示抽认卡组列表(见图 [10-2](#Fig2) )。它包括当前可用的抽认卡组列表、您希望抽认卡组如何显示(随机卡顺序、首先显示中文或英文，或随机)，以及是显示一组卡还是只显示一组卡中的单词。

![A436770_1_En_10_Fig2_HTML.jpg](img/A436770_1_En_10_Fig2_HTML.jpg)

图 10-2。

The home page of the web site

抽认卡组显示一张抽认卡，一次一张。一个看起来像图 [10-3](#Fig3) 。

![A436770_1_En_10_Fig3_HTML.jpg](img/A436770_1_En_10_Fig3_HTML.jpg)

图 10-3。

Typical flashcard showing all the components

抽认卡的单词集如图 [10-4](#Fig4) 所示。

![A436770_1_En_10_Fig4_HTML.jpg](img/A436770_1_En_10_Fig4_HTML.jpg)

图 10-4。

The list of words in a flashcard set

## 浏览器文件

浏览器端有 HTML，CSS，JavaScript 文件。这些措施如下:

*   首页(`flashcards.html`):
    *   `html/ListFlashcardsStylesheet.css`
*   抽认卡组(`ShowFlashcards.html`):
    *   `css/CardStyleSheet.css`
    *   `jscript/jquery.js`
    *   `jscript/slideviewer.js`
*   抽认卡设置单词(`ListWords.html`):无额外

## 基本服务器

该服务器是一个 HTTP 服务器，如前一章所述。它有许多功能来处理不同的网址。这些功能概述如下:

<colgroup><col> <col> <col></colgroup> 
| 小路 | 功能 | HTML 已交付 |
| --- | --- | --- |
| `/` | `listFlashCards` | `html/ListFlashcards.html` |
| `/flashcards.html` | `listFlashCards` | `html/ListFlashcards.html` |
| `/flashcardSets` | `manageFlashCards` | `html/showFlashcards.html` |
| `/flashcardSets` | `manageFlashCards` | `html/ListWords.html` |
| `/jscript/*` | `fileServer` | 目录`/jscript`中的文件 |
| `/html/*` | `fileServer` | 目录`/html`中的文件 |

暂且省略功能本身，服务器是 [`http://www.apress.com/9781484226919`](http://www.apress.com/9781484226919) 的`Ch10`下的`Server.go`。

```go
/* Server
 */

package main

import (
        "fmt"
        "net/http"
        "os"
        "html/template"
)

import (
        "dictionary"
        "flashcards"
        "templatefuncs"
)

func main() {
       if len(os.Args) != 2 {
                fmt.Fprint(os.Stderr, "Usage: ", os.Args[0], ":port\n")
                os.Exit(1)
       }
        port := os.Args[1]

        http.HandleFunc("/", listFlashCards)
        fileServer := http.StripPrefix("/jscript/", http.FileServer(http.Dir("jscript")))
        http.Handle("/jscript/", fileServer)
        fileServer = http.StripPrefix("/html/", http.FileServer(http.Dir("html")))
        http.Handle("/html/", fileServer)

        http.HandleFunc("/flashcards.html", listFlashCards)
        http.HandleFunc("/flashcardSets", manageFlashCards)

        // deliver requests to the handlers
        err := http.ListenAndServe(port, nil)
        checkError(err)
        // That's it!

}

func listFlashCards(rw http.ResponseWriter, req *http.Request) {
         ...
}

/*
 * Called from ListFlashcards.html on form submission
 */
func manageFlashCards(rw http.ResponseWriter, req *http.Request) {
        ...
}

func showFlashCards(rw http.ResponseWriter, cardname, order, half string) {
        ...
}

func listWords(rw http.ResponseWriter, cardname string) {
         ...
}

func checkError(err error) {
        if err != nil {
                  fmt.Println("Fatal error ", err.Error())
                  os.Exit(1)
        }
}

```

我们现在开始讨论单个函数。

## listFlashCards 函数

调用`listFlashCards`函数为顶层页面创建 HTML。抽认卡名称列表是可扩展的，是目录`flashcardSets`中的一组文件条目。此列表用于在顶级页面中创建表格，最好使用模板包来完成:

```go
<table>
  {{range .}}
  <tr>
    <td>
      {{.}}
    </td>
  </tr>
</table>

```

其中范围超出了名称列表。文件`html/ListFlashcards.html`包含这个模板以及卡片顺序、半卡片显示和底部表单按钮的 HTML。省略了边列表和提交按钮，HTML 如下:

```go
<html>
  <head>
    <title>
      Flashcards
    </title>
    <link type="text/css" rel="stylesheet"
          href="/html/ListFlashcardsStylesheet.css">
    </link>
  </head>
  <body>
    <h1>
      Flashcards
    </h1>
    <p>

      <div id="choose">
        <form method="GET" action="http:flashcardSets">

          <table border="1" id="sets">
            <tr>
              <th  colspan="2">
                Flashcard Sets
              </th>
            </tr>
            {{range .}}
            <tr>
              <td>
                {{.}}
              </td>
              <td>
                <input type="radio" name="flashcardSets" value="{{.}}" />
              </td>
            </tr>
            {{end}}
          </table>
       </div>
    </p>
  </body>
</html>

```

将模板应用于此的函数`listFlashCards`如下:

```go
func listFlashCards(rw http.ResponseWriter, req *http.Request) {

         flashCardsNames := flashcards.ListFlashCardsNames()
         t, err := template.ParseFiles("html/ListFlashcards.html")
         if err != nil {
                   http.Error(rw, err.Error(), http.StatusInternalServerError)
                   return
         }
         t.Execute(rw, flashCardsNames)
}

```

函数`flashcards.ListFlashCardsNames()`只是遍历抽认卡目录，返回一个字符串数组(每个抽认卡集的文件名):

```go
func ListFlashCardsNames() []string {
        flashcardsDir, err := os.Open("flashcardSets")
        if err != nil {
                return nil
        }
        files, err := flashcardsDir.Readdir(-1)

        fileNames := make([]string, len(files))
        for n, f := range files {
                fileNames[n] = f.Name()
        }
        sort.Strings(fileNames)
        return fileNames
}  

```

## manageFlashCards 功能

按下“显示集合中的卡片”按钮或“列出集合中的单词”按钮时，调用`manageFlashCards`函数来管理表单提交。它从表单请求中提取值，然后在`showFlashCards`和`listWords`之间进行选择:

```go
func manageFlashCards(rw http.ResponseWriter, req *http.Request) {

       set := req.FormValue("flashcardSets")
       order := req.FormValue("order")
       action := req.FormValue("submit")
       half := req.FormValue("half")
       cardname := "flashcardSets/" + set

       fmt.Println("cardname", cardname, "action", action)
       if action == "Show cards in set" {
                 showFlashCards(rw, cardname, order, half)
       } else if action == "List words in set" {
                 listWords(rw, cardname)
       }
}

```

## 汉语词典

前面的代码相当普通:它使用文件服务器交付静态文件，使用基于目录中文件列表的模板创建 HTML 表，并处理来自 HTML 表单的信息。为了进一步了解每张卡片上显示的内容，我们必须了解应用程序的具体细节，这意味着要了解单词的来源(字典)，如何表示单词和卡片，以及如何将抽认卡数据发送到浏览器。首先，字典。

汉语是一种复杂的语言——难道它们不都是:-(。书写形式是象形文字，也就是“象形图”，而不是使用字母表。但这种书写形式随着时间的推移而演变，甚至最近分裂成两种形式:在台湾和香港使用的“繁体”中文，以及在 mainland China 使用的“简体”中文。虽然大多数字符是相同的，但大约有 1000 个字符是不同的。因此，一部汉语词典通常会有两种相同的书写形式。

像我这样的西方人，大多看不懂这些文字。所以有一种“拉丁化”的形式，叫做拼音，它是以拉丁字母为基础，用音标书写汉字。它不完全是拉丁字母，因为汉语是一种带声调的语言，拼音形式必须显示声调(很像法语和其他欧洲语言中的重音)。所以一个典型的字典必须显示四样东西:繁体、简体、拼音和英语。另外(就像英语里一样)，一个词可能有多个意思。比如 [`http://www.mandarintools.com/worddict.html`](http://www.mandarintools.com/worddict.html) 有免费的中/英文词典，更好的是可以下载成 UTF-8 文件。在里面，这个词好有这个条目:

<colgroup><col> <col> <col> <col> <col></colgroup> 
| 传统的 | 简化了的 | 拼音 | 英语 | 含义 |
| --- | --- | --- | --- | --- |
| good | good | 嘿！嘿 | 好的 | /good/well/proper/good to/easy to/very/so/(表示完成或准备就绪的后缀)/ |

这本字典有点复杂。大多数键盘不擅长表现重音，如中的卡隆音。因此，虽然汉字是用 Unicode 书写的，但拼音字符不是。虽然像\u 这样的字母有 Unicode 字符，但包括这本在内的许多词典都使用拉丁字母 a，并将音调放在单词的末尾。这里是第三声，所以 hǎo 写成 hao3。这使得那些只有美国键盘而没有 Unicode 编辑器的人仍然可以更容易地用拼音交流。网络服务器使用的字典的副本是`cedict_ts_u8`。

这种数据格式不匹配不是什么大问题。只是在原始文本字典和浏览器显示之间的某个地方，必须执行数据消息。Go 模板允许通过定义一个自定义模板来实现这一点，所以我选择了这条路线。替代方法包括在读入字典时这样做，或者在 JavaScript 中显示最后的字符。

### 字典类型

我们使用一个`Entry`来保存一个单词的基本信息:

```go
type Entry struct {
     Traditional string
     Simplified string
     Pinyin     string
     Translations []string
}

```

上面的单词将由以下内容表示:

```go
Entry{Traditional: 好,
          Simplified: 好,
          Pinyin: `hao3`
          Translations: []string{`good`, `well`,`proper`,
                                           `good to`, `easy to`, `very`, `so`,  
                                           `(suffix indicating completion or readiness)`}
}

```

字典本身就是这些条目的数组:

```go
type Dictionary struct {
      Entries []*Entry
}

```

## 抽认卡套装

一张抽认卡代表一个中文单词和这个单词的英文翻译。我们已经看到，一个单一的中文单词可以有许多可能的英文意思。但这部词典有时也会多次出现一个中文单词。举个例子，好至少出现两次，一次带有我们已经看到的意思，但也带有另一个意思，“喜欢”。这被证明是多余的，但是考虑到这一点，每个抽认卡都有一个完整的单词字典。通常字典中只有一个条目！抽认卡的其余部分只是作为可能的密钥的简化和英语单词:

```go
type FlashCard struct {
        Simplified string
        English    string
        Dictionary *dictionary.Dictionary

}

```

抽认卡组是这些抽认卡的一个数组，加上抽认卡组的名称，以及将被发送到浏览器以显示抽认卡组的信息:随机或固定顺序，首先显示每张卡片的顶部或底部，或者随机。

```go
type FlashCards struct {
        Name      string
        CardOrder string
        ShowHalf  string
        Cards     []*FlashCard
}

```

我们已经展示了这种类型的一个函数，`ListFlashCardsNames()`。这种类型还有一个有趣的功能，为抽认卡集加载 JSON 文件。这使用了第 4 章[的技术](04.html)，连载。

```go
func LoadJSON(fileName string, key interface{}) {
        inFile, err := os.Open(fileName)
        checkError(err)
        decoder := json.NewDecoder(inFile)
        err = decoder.Decode(key)
        checkError(err)
        inFile.Close()
}

```

一套典型的抽认卡是由普通单词组成的。当 JSON 文件被 Python ( `print json.dump(string, indent=4, separators=(',', ':'))`)漂亮地打印出来时，它的一部分看起来像这样:

```go
{
    "ShowHalf":"",
    "Cards":[
        {
            "Simplified":"\u4f60\u597d",
            "Dictionary":{
                "Entries":[
                    {
                        "Traditional":"\u4f60\u597d",
                        "Pinyin":"ni3 hao3",
                        "Translations":[
                            "hello",

                            "hi",
                            "how are you?"

                        ],
                        "Simplified":"\u4f60\u597d"
                    }
                ]
            },
            "English":"hello"
        },

        {
            "Simplified":"\u5582",
            "Dictionary":{
                "Entries":[
                    {
                        "Traditional":"\u5582",
                        "Pinyin":"wei4",
                        "Translations":[
                            "hello (interj., esp. on telephone)",
                            "hey",
                            "to feed (sb or some animal)"
                        ],
                        "Simplified":"\u5582"
                    }
                ]
            },
            "English":"hello (interj., esp. on telephone)"
        },
    ],
    "CardOrder":"",
    "Name":"Common Words"

}

```

## 修正口音

在我们完成服务器的代码之前，还有最后一个主要任务。字典中给出的重音符号将重音符号放在拼音单词的末尾，如 hao3 中的 hǎo。如第 [9](09.html) 章所述，可以通过自定义模板将重音符号转换为 Unicode。

这里给出了拼音格式化程序的代码。除非你真的有兴趣了解拼音格式的规则，否则不要费心去读它。程序是`PinyinFormatter.go`:

```go
package templatefuncs

import (
        "fmt"
        "strings"
)

func PinyinFormatter(args ...interface{}) string {
        ok := false
        var s string
        if len(args) == 1 {
                s, ok = args[0].(string)
        }
        if !ok {
                s = fmt.Sprint(args...)
        }
        fmt.Println("Formatting func " + s)
        // the string may consist of several pinyin words
        // each one needs to be changed separately and then
        // added back together
        words := strings.Fields(s)

        for n, word := range words {
                // convert "u:" to "ü" if present
                uColon := strings.Index(word, "u:")
                if uColon != -1 {
                        parts := strings.SplitN(word, "u:", 2)
                        word = parts[0] + "ü" + parts[1]
                }
                println(word)
                // get last character, will be the tone if present
                chars := []rune(word)
                tone := chars[len(chars)-1]

                if tone == '5' {
                        // there is no accent for tone 5
                        words[n] = string(chars[0 : len(chars)-1])
                        println("lost accent on", words[n])
                        continue
                }
                if tone < '1' || tone > '4' {
                        // not a tone value
                        continue
                }
                words[n] = addAccent(word, int(tone))
        }
        s = strings.Join(words, ` `)
        return s
}

var (
        // maps 'a1' to '\u0101' etc
        aAccent = map[int]rune{
                '1': '\u0101',
                '2': '\u00e1',
                '3': '\u01ce',
                '4': '\u00e0'}
        eAccent = map[int]rune{
                '1': '\u0113',
                '2': '\u00e9',
                '3': '\u011b',
                '4': '\u00e8'}
        iAccent = map[int]rune{
                '1': '\u012b',
                '2': '\u00ed',
                '3': '\u01d0',
                '4': '\u00ec'}
        oAccent = map[int]rune{
                '1': '\u014d',
                '2': '\u00f3',
                '3': '\u01d2',
                '4': '\u00f2'}
        uAccent = map[int]rune{
                '1': '\u016b',
                '2': '\u00fa',
                '3': '\u01d4',
                '4': '\u00f9'}
        üAccent = map[int]rune{

                '1': 'ǖ',
                '2': 'ǘ',
                '3': 'ǚ',
                '4': 'ǜ'}
)

func addAccent(word string, tone int) string {
        /*
         * Based on "Where do the tone marks go?"
         * at http://www.pinyin.info/rules/where.html
         */

        n := strings.Index(word, "a")
        if n != -1 {
                aAcc := aAccent[tone]
                // replace 'a' with its tone version
                word = word[0:n] + string(aAcc) + word[(n+1):len(word)-1]
        } else {
                n := strings.Index(word, "e")
                if n != -1 {
                        eAcc := eAccent[tone]
                        word = word[0:n] + string(eAcc) +
                                word[(n+1):len(word)-1]
                } else {
                        n = strings.Index(word, "ou")
                        if n != -1 {
                                oAcc := oAccent[tone]
                                word = word[0:n] + string(oAcc) + "u" +
                                        word[(n+2):len(word)-1]
                        } else {
                                chars := []rune(word)
                                length := len(chars)
                                // put tone on the last vowel
                        L:
                                for n, _ := range chars {
                                        m := length - n - 1
                                        switch chars[m] {
                                        case 'i':
                                                chars[m] = iAccent[tone]
                                                break L
                                        case 'o':
                                                chars[m] = oAccent[tone]
                                                break L
                                        case 'u':
                                                chars[m] = uAccent[tone]
                                                break L
                                        case 'ü':
                                                chars[m] = üAccent[tone]
                                                break L
                                        default:
                                        }

                                }
                                word = string(chars[0 : len(chars)-1])
                        }
                }
        }
        return word
}

```

## ListWords 函数

我们现在可以回到服务器的突出功能。一个是在一套抽认卡中列出单词。这将使用抽认卡集的模板填充一个 HTML 表。HTML 使用模板包遍历一个`FlashCards`结构并插入该结构中的字段:

```go
<html>
  <head>
    <title>
      Words for {{.Name}}
    </title>

  </head>
  <body>
    <h1>
      Words for {{.Name}}
    </h1>
    <p>
      <table border="1" class="sortable">
        <tr>
          <th> English </th>
          <th> Pinyin </th>
          <th> Traditional </th>
          <th> Simplified </th>
        </tr>
      {{range .Cards}}
        <div class="card">
          <tr>
          <div class="english">
            <div class="vcenter">
              <td>
                {{.English}}

              </td>
            </div>
          </div>

          {{with .Dictionary}}
            {{range .Entries}}
              <div class="pinyin">
                <div class="vcenter">
                  <td>
                    {{.Pinyin|pinyin}}
                  </td>
                </div>
              </div>

              <div class="traditional">
                <div class="vcenter">
                  <td>
                    {{.Traditional}}
                  </td>
                </div>
              </div>

              <div class="simplified">
                <div class="vcenter">
                  <td>
                    {{.Simplified}}
                  </td>
                </div>
              </div>

            {{end}}
          {{end}}
          </tr>
        </div>
      {{end}}
      </table>
    </p>
    <p class ="return">
      <a href="http:/flashcards.html"> Return to Flash Cards list</a>
    </p>
  </body>
</html>

```

为此，`Server.go`中的 Go 函数使用了上一节讨论的`PinyinFormatter`:

```go
func listWords(rw http.ResponseWriter, cardname string) {
        cards := new(flashcards.FlashCards)
        flashcards.LoadJSON(cardname, cards)
        fmt.Println("Card name", cards.Name)

        t := template.New("ListWords.html")

        t = t.Funcs(template.FuncMap{"pinyin": templatefuncs.PinyinFormatter})
        t, err := t.ParseFiles("html/ListWords.html")

        if err != nil {
                fmt.Println("Parse error " + err.Error())
                http.Error(rw, err.Error(), http.StatusInternalServerError)
                return
        }
        err = t.Execute(rw, cards)
        if err != nil {
                fmt.Println("Execute error " + err.Error())
                http.Error(rw, err.Error(), http.StatusInternalServerError)
                return
        }
}

```

这会将填充的表格发送到浏览器，如图 [10-4](#Fig4) 所示。

## showFlashCards 功能

完成服务器的最后一个函数是`showFlashCards`。这将根据浏览器提交的表单，改变抽认卡组中`CardOrder`和`ShowHalf`的默认值。然后应用`PinyinFormatter`并将结果文档发送给浏览器。我使用 UNIX 命令`script`捕获命令行会话的输出，然后运行命令:

```go
GET /flashcardSets?flashcardSets=Common+Words&order=Random&half=Chinese&submit=Show+cards+in+set HTTP/1.0

```

部分结果如下:

```go
<html>
  <head>
    <title>
      Flashcards for Common Words

    </title>

    <link type="text/css" rel="stylesheet"
          href="/html/CardStylesheet.css">
    </link>

    <script type="text/javascript"
            language="JavaScript1.2" src="/jscript/jquery.js">
      <!-- empty -->
    </script>

    <script type="text/javascript"
            language="JavaScript1.2" src="/jscript/slideviewer.js">
      <!-- empty -->
    </script>

    <script type="text/javascript"
            language="JavaScript1.2">
      cardOrder = "RANDOM";
      showHalfCard = "CHINESE_HALF";
    </script>
  </head>
  <body onload="showSlides();">

    <h1>
      Flashcards for Common Words
    </h1>
    <p>

        <div class="card">
          <div class="english">
            <div class="vcenter">
              hello
            </div>
          </div>

              <div class="pinyin">
                <div class="vcenter">
                  nǐ hǎo
                </div>
              </div>

              <div class="traditional">
                <div class="vcenter">
                  你好
                </div>
              </div>

              <div class="simplified">
                <div class="vcenter">
                  你好
                </div>
              </div>

              <div class ="translations">

                <div class="vcenter">

                  hello <br />

                  hi <br />

                  how are you? <br />

                </div>
              </div>

        </div>

```

## 浏览器上的演示

这个系统的最后一部分是如何在浏览器中显示这个 HTML。图 [10-3](#Fig3) 显示了一个由四部分组成的屏幕，显示英语、简体中文、备选翻译和繁体/简体对。这是如何通过下载到服务器的 JavaScript 程序完成的(这是使用`FileServer` Go 对象完成的)。JavaScript `slideviewer.js`文件实际上很长，因此在文本中被省略了。它包含在 [`http://www.apress.com/9781484226919`](http://www.apress.com/9781484226919) 的程序文件中。

## 运行服务器

这是本书中第一个使用我们自己导入的文件的程序。所有以前的程序都只是使用了一个主文件和 Go 标准库。包的`dictionary`、抽认卡和`pinyin`中导入的文件需要组织好，以便`go`命令可以找到它们。

需要将环境变量`GOPATH`设置到一个目录中，该目录下有一个子目录`src`，该子目录包含适当子目录中导入的源文件:

```go
src/flashcards/FlashCards.go
src/pinyin/PinyinFormatter.go
src/dictionary/Dictionary.go

```

然后，可以使用如下命令在端口`8000`(或其他端口)上运行服务器:

```go
go run Server.go :8000

```

## 结论

本章考虑了一个相对简单但完整的 web 服务器，它使用静态和动态 web 页面处理表单，并使用模板来简化编码。