# 十三、远程过程调用

套接字和 HTTP 编程都使用消息传递范式。客户端向服务器发送消息，服务器通常会发回消息。双方都负责以双方都能理解的格式创建消息，并从这些消息中读取数据。

然而，大多数独立的应用程序很少使用消息传递技术。通常，首选机制是函数(或方法或过程)调用的机制。在这种风格中，程序将调用一个带有参数列表的函数，在函数调用完成时，将有一组返回值。这些值可能是函数值，或者如果地址已作为参数传递，则这些地址的内容可能已被更改。

远程过程调用是将这种编程风格引入网络世界的一种尝试。因此，客户端将进行看起来像正常的过程调用。客户端会把这个打包成一个网络消息，传输给服务器。服务器将对其进行解包，并将其转换回服务器端的过程调用。这个调用的结果将被打包返回给客户端。

概略地看起来如图 [13-1](#Fig1) 所示。

![A436770_1_En_13_Fig1_HTML.gif](img/A436770_1_En_13_Fig1_HTML.gif)

图 13-1。

The remote procedure call steps

步骤如下:

1.  客户端调用客户端过程存根。存根将参数打包成网络消息。这叫做编组。
2.  存根调用 O/S 内核中的网络例程来发送消息。
3.  内核将消息发送到远程系统。这可能是面向连接的，也可能是无连接的。
4.  服务器过程存根从网络消息中解组参数。
5.  服务器过程存根执行服务器过程实现。
6.  过程完成，将执行返回给服务器过程存根。
7.  服务器存根将返回值整理成网络消息。
8.  返回消息被发送回来。
9.  客户端程序存根使用网络例程读取消息。
10.  消息被解组，返回值被设置在客户端程序的堆栈上。

实现 RPC 有两种常见的方式。第一种以 Sun 的 ONC/RPC 和 CORBA 为代表。在这种情况下，服务的规范以某种抽象语言给出，例如 CORBA IDL(接口定义语言)。然后将其编译成用于客户端和服务器的代码。然后，客户端编写一个包含对过程/函数/方法的调用的普通程序，该程序链接到生成的客户端代码。服务器端代码实际上是一个服务器本身，它链接到您编写的过程实现。

在第一种方式中，客户端代码在外观上与普通的过程调用几乎相同。通常有一点额外的代码来定位服务器。在 Sun 的 ONC 中，服务器的地址必须是已知的；在 CORBA 中，调用一个命名服务来查找服务器的地址；在 Java RMI 中，IDL 就是 Java 本身，命名服务用于查找服务的地址。

在第二种风格中，您必须使用特殊的客户端 API。您将函数名及其参数传递给客户端的这个库。在服务器端，您必须自己显式地编写服务器，以及远程过程实现。

第二种方法被许多 RPC 系统使用，比如 Web 服务。这也是 Go 的 RPC 使用的方法。

## Go 的 RPC

Go 的 RPC 是迄今为止唯一的 Go。它不同于其他的 RPC 系统，所以 Go 客户端只与 Go 服务器对话。它使用第 [4](04.html) 章中讨论的 Gob 序列化系统，该系统定义了可以使用的数据类型。

RPC 系统通常对可以通过网络调用的函数做一些限制。这是为了让 RPC 系统可以正确地确定发送哪些值参数，哪些引用参数接收答案，以及如何发出错误信号。

在 Go 中，限制是

*   方法的类型被导出(以大写字母开始)。
*   该方法被导出。
*   该方法有两个参数，都是导出(或内置)类型。第一个是传递给方法的数据；第二个是返回的结果。
*   该方法的第二个参数是一个指针。
*   它具有类型为`error`的返回值。

例如，下面是一个有效的函数:

```go
F(T1, &T2) error

```

对参数的限制意味着您通常必须定义一个结构类型。Go 的 RPC 使用`gob`包来编组和解组数据，因此参数类型必须遵循 Gob 的规则，如前一章所述。

我们将遵循 Go 文档中给出的例子，因为它说明了重要的几点。服务器执行两个简单的操作——它们不需要 RPC 的“咕哝”,但是很容易理解。这两个操作是将两个整数相乘，并找出第一个整数除以第二个整数后的商和余数。

要操作的两个值在结构中给出:

```go
type Values struct {
    A, B int
}

```

总和只是一个`int`，而商/余数是另一个结构:

```go
type Quotient struct {
    Quo, Rem int
}

```

我们将有两个函数，`multiply`和`divide`，可以在 RPC 服务器上调用。这些函数需要在 RPC 系统中注册。`Register`函数接受单个参数，这是一个接口。所以我们需要一个具有这两个功能的类型:

```go
type Arith int

func (t *Arith) Multiply(args *Args, reply *int) error {
        *reply = args.A * args.B
        return nil
}

func (t *Arith) Divide(args *Args, quo *Quotient) error {
        if args.B == 0 {
                return errors.New("divide by zero")
        }
        quo.Quo = args.A / args.B
        quo.Rem = args.A % args.B
        return nil
}

```

`Arith`的底层类型给定为`int`。没关系，任何类型都可以。

现在可以使用`Register`注册这种类型的对象，然后 RPC 系统可以调用它的方法。

### HTTP RPC 服务器

任何 RPC 都需要一种传输机制来通过网络获取消息。Go 可以使用 HTTP 或者 TCP。HTTP 机制的优点是它可以利用 HTTP 支持库。您需要向 HTTP 层添加一个 RPC 处理程序，这是使用`HandleHTTP`完成的，然后启动一个 HTTP 服务器。完整的代码是`ArithServer.go`:

```go
/**
 * ArithServer
 */

package main

import (
        "fmt"
        "net/rpc"
        "errors"
        "net/http"
)

type Values struct {
        A, B int
}

type Quotient struct {
        Quo, Rem int
}

type Arith int

func (t *Arith) Multiply(args *Values, reply *int) error {
        *reply = args.A * args.B

        return nil
}

func (t *Arith) Divide(args *Values, quo *Quotient) error {
        if args.B == 0 {
                return errors.New("divide by zero")
        }
        quo.Quo = args.A / args.B
        quo.Rem = args.A % args.B
        return nil
}

func main() {

        arith := new(Arith)
        rpc.Register(arith)
        rpc.HandleHTTP()

        err := http.ListenAndServe(":1234", nil)
        if err != nil {
                fmt.Println(err.Error())
        }
}

```

它是由

```go
go run ArithServer.go

```

### HTTP RPC 客户端

客户端需要建立到 RPC 服务器的 HTTP 连接。它需要准备一个包含要发送的值的结构，以及存储结果的变量的地址。然后它可以用这些参数做一个`Call`:

*   要执行的远程函数的名称
*   要发送的值
*   存储结果的变量的地址

调用算术服务器的两个功能的客户端是`ArithClient.go`:

```go
/**
* ArithClient
 */

package main

import (
        "net/rpc"
        "fmt"
        "log"
        "os"
)

type Args struct {
        A, B int
}

type Quotient struct {
        Quo, Rem int
}

func main() {
        if len(os.Args) != 2 {
                fmt.Println("Usage: ", os.Args[0], "server")
                os.Exit(1)
        }
        serverAddress := os.Args[1]

        client, err := rpc.DialHTTP("tcp", serverAddress+":1234")
        if err != nil {
                log.Fatal("dialing:", err)
        }
        // Synchronous call
        args := Args{17, 8}
        var reply int
        err = client.Call("Arith.Multiply", args, &reply)
        if err != nil {
                log.Fatal("arith error:", err)
        }
        fmt.Printf("Arith: %d*%d=%d\n", args.A, args.B, reply)

        var quot Quotient
        err = client.Call("Arith.Divide", args, &quot)
        if err != nil {
                log.Fatal("arith error:", err)
        }
        fmt.Printf("Arith: %d/%d=%d remainder %d\n", args.A, args.B, quot.Quo, quot.Rem)

}

```

运行时:

```go
go run  ArithClient.go localhost

```

它产生以下内容:

```go
Arith: 17*8=136
Arith: 17/8=2 remainder 1      

```

### TCP RPC 服务器

使用 TCP 套接字的服务器版本是`TCPArithServer.go`:

```go
/**
* TCPArithServer
 */

package main

import (
        "fmt"
        "net/rpc"
        "errors"
        "net"
        "os"
)

type Args struct {
        A, B int
}

type Quotient struct {
        Quo, Rem int
}

type Arith int

func (t *Arith) Multiply(args *Args, reply *int) error {
        *reply = args.A * args.B
        return nil
}

func (t *Arith) Divide(args *Args, quo *Quotient) error

{
        if args.B == 0 {
                return errors.New("divide by zero")
        }
        quo.Quo = args.A / args.B
        quo.Rem = args.A % args.B
        return nil
}

func main() {

        arith := new(Arith)

        rpc.Register(arith)

        tcpAddr, err := net.ResolveTCPAddr("tcp", ":1234")
        checkError(err)

        listener, err := net.ListenTCP("tcp", tcpAddr)
        checkError(err)

        /* This works:
        rpc.Accept(listener)
        */
        /* and so does this:
         */
        for {
                conn, err := listener.Accept()
                if err != nil {
                        continue
                }
                rpc.ServeConn(conn)
        }

}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

注意对`Accept`的调用是阻塞的，只处理客户端连接。如果服务器还想做其他工作，它应该在一个`go`例程中调用它。

### TCP RPC 客户端

使用 TCP 服务器并调用算术服务器的两个功能的客户端是`TCPArithClient.go`:

```go
/**
* TCPArithClient
 */

package main

import (
        "net/rpc"
        "fmt"
        "log"
        "os"
)

type Args struct {
        A, B int
}

type Quotient struct {
        Quo, Rem int
}

func main() {
        if len(os.Args) != 2 {
                fmt.Println("Usage: ", os.Args[0], "server:port")
                os.Exit(1)
        }
        service := os.Args[1]

        client, err := rpc.Dial("tcp", service)
        if err != nil {
                log.Fatal("dialing:", err)
        }
        // Synchronous call
        args := Args{17, 8}
        var reply int
        err = client.Call("Arith.Multiply", args, &reply)
        if err != nil {
                log.Fatal("arith error:", err)
        }
        fmt.Printf("Arith: %d*%d=%d\n", args.A, args.B, reply)

        var quot Quotient

        err = client.Call("Arith.Divide", args, &quot)
        if err != nil {
                log.Fatal("arith error:", err)
        }
        fmt.Printf("Arith: %d/%d=%d remainder %d\n", args.A, args.B, quot.Quo, quot.Rem)

}

```

当它运行时:

```go
go run TCPArithClient.go localhost:1234

```

它产生以下内容:

```go
Arith: 17*8=136

Arith: 17/8=2 remainder 1

```

### 匹配值

您可能已经注意到，在 HTTP 客户机和 HTTP 服务器上，值参数的类型是不同的。在服务器端，我们使用了`Values`，而在客户端，我们使用了`Args`。这没关系，因为我们遵循 Gob 序列化的规则，并且两个结构的字段的名称和类型是匹配的。更好的编程实践会说，名字当然应该是相同的！

然而，这确实指出了使用 Go RPC 的一个可能的陷阱。如果我们将服务器的结构改为这样:

```go
type Values struct {
        C, B int
}

```

那 Gob 就没问题了。在客户端，解组将忽略服务器给定的值`C`，并使用默认的零值`A`。比如说，如果除以`A`(零)就会产生问题。

使用 Go RPC 将需要程序员严格执行字段名和类型的稳定性。我们注意到没有版本控制机制来做到这一点，Gob 中也没有机制来通知任何可能的不匹配。也不需要外部表示作为参考。如果你只是添加字段，这可能是好的，但它仍然需要控制。也许在数据结构中添加一个`version`字段会有所帮助。

## 数据

这一节没有给前面的概念增加新的内容。它只是对数据使用了不同的“有线”格式，用 JSON 代替 Gob。因此，客户机或服务器可以用理解套接字和 JSON 的其他语言编写。

### JSON RPC 服务器

使用 JSON 编码的服务器版本是`JSONArithServer.go`:

```go
/* JSONArithServer
 */

package main

import (
        "fmt"
        "net/rpc"
        "net/rpc/jsonrpc"
        "os"
        "net"
        "errors"
)

type Args struct {
        A, B int
}

type Quotient struct {
        Quo, Rem int
}

type Arith int

func (t *Arith) Multiply(args *Args, reply *int) error {
        *reply = args.A * args.B
        return nil
}

func (t *Arith) Divide(args *Args, quo *Quotient) error {
        if args.B == 0 {
                return errors.New("divide by zero")
        }
        quo.Quo = args.A / args.B
        quo.Rem = args.A % args.B
        return nil
}

func main() {

        arith := new(Arith)

        rpc.Register(arith)

        tcpAddr, err := net.ResolveTCPAddr("tcp", ":1234")
        checkError(err)

        listener, err := net.ListenTCP("tcp", tcpAddr)
        checkError(err)

        /* This works:
        rpc.Accept(listener)
        */
        /* and so does this:
         */
        for {
                conn, err := listener.Accept()
                if err != nil {
                        continue
                }
                jsonrpc.ServeConn(conn)
        }

}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

它是由

```go
go run JSONArithServer.go

```

### JSON RPC 客户端

调用算术服务器的两个功能的客户端是`JSONArithClient.go`:

```go
/* JSONArithCLient
 */

package main

import (
        "net/rpc/jsonrpc"
        "fmt"
        "log"
        "os"
)

type Args struct {
        A, B int
}

type Quotient struct {
        Quo, Rem int
}

func main() {
        if len(os.Args) != 2 {
                fmt.Println("Usage: ", os.Args[0], "server:port")
                log.Fatal(1)
        }
        service := os.Args[1]

        client, err := jsonrpc.Dial("tcp", service)
        if err != nil {
                log.Fatal("dialing:", err)
        }
        // Synchronous call
        args := Args{17, 8}
        var reply int
        err = client.Call("Arith.Multiply", args, &reply)
        if err != nil {
                log.Fatal("arith error:", err)
        }
        fmt.Printf("Arith: %d*%d=%d\n", args.A, args.B, reply)

        var quot Quotient
        err = client.Call("Arith.Divide", args, &quot)
        if err != nil {
                log.Fatal("arith error:", err)
        }
        fmt.Printf("Arith: %d/%d=%d remainder %d\n", args.A, args.B, quot.Quo, quot.Rem)

}

```

它是这样运行的:

```go
go run JSONArithClient.go localhost:1234

```

它产生以下输出:

```go
Arith: 17*8=136
Arith: 17/8=2 remainder 1

```

## 结论

RPC 是一种流行的分发应用程序的方法。这里介绍了几种方法，基于 Gob 或 JSON 序列化技术，使用 HTTP 和 TCP 进行传输。