# 十四、REST

在前几章中，我们研究了 HTTP 并给出了一个 web 系统的例子。然而，我们并没有给这个系统任何特别的结构，只是给出了对这个问题来说足够简单的结构。HTTP 1.1 的主要作者之一(Roy Fielding)开发了一种架构风格，称为 REST(表述性状态转移)。在这一章中，我们来看看 REST 风格以及它对构建 web 应用程序的意义。为此，我们必须回到基本面。

如果术语 REST 可以被正确地应用，REST 有许多必须遵循的组件。不幸的是，它已经成为一个时髦的词，许多应用程序有“一些”REST，但不是全部。我们讨论 Richardson 成熟度模型，它表明了 API 在通往 RESTful-ness 的道路上已经走了多远。

在上一章中，我们看了远程过程调用。这是和 REST 完全不同的风格。我们还比较了这两种风格，看看什么时候适合使用每种风格。

## URIs 和资源

资源是我们希望在网络或互联网上与之交互的“东西”。我喜欢把它们看作对象，但是并不要求它们的实现应该是基于对象的——它们应该只是“看起来”像一个东西，可能有组件。

每个资源都有一个或多个称为 URIs(统一资源标识符)的地址。

Note

国际化形式是 IRIs—国际化资源标识符。

它们的一般形式如下:

```go
scheme:[//[user:password@]host[:port]][/]path[?query][#fragment]

```

典型的例子是 URL(统一资源定位符)，其中方案是`http`或`https`，主机通过其 IP 地址或 DNS 名称来指代计算机，如下所示:

```go
https://jan.newmarch.name/IoT/index.html

```

有非 HTTP URL 方案，如 telnet、news 和 ipp(互联网打印协议)。这些还包含位置组件。还有其他的，比如 urn(统一资源名)，它们通常是其他标识系统的包装器，并且不包含位置信息。例如，IETF 有一个由 ISBN 标识的图书的标准 URN 方案，例如这本书的 ISBN:

```go
urn:ISBN:978-1-4842-2692-6

```

这些骨灰盒往往不会被广泛使用，但仍然存在。在 [`https://www.iana.org/assignments/urn-namespaces/urn-namespaces.xhtml`](https://www.iana.org/assignments/urn-namespaces/urn-namespaces.xhtml) 由 IANA 统一资源名称(URN)命名空间给出列表。最初的方案，如 ISBN，仍在广泛使用。

资源的正式定义可能很难确定。例如， [`http://www.google.com`](http://www.google.com) 在某种意义上代表谷歌(它是一个 URL 的方案和主机部分)，但主机肯定不是某个地方的某台固定电脑。类似地，这本书的 ISBN 代表了这本书的一些内容，但肯定不是任何现存的副本(在写这一章的时候，没有副本存在，即使 ISBN 存在！).

然而，我们认为资源的概念是原始的，URIs 是这些资源的标识符。IETF at 统一资源标识符(URI):通用语法( [`https://www.ietf.org/rfc/rfc3986.txt`](https://www.ietf.org/rfc/rfc3986.txt) )也同样含糊:“术语“资源”在一般意义上用于 URI 可能标识的任何东西”。

一个资源可以有多个 URI。作为一个人，我有许多不同的标识符:我的税务档案号指的是我的一个方面，我的财务；我的 Medicare 号码将我视为健康治疗的接受者；我的名字(相当独特)经常被用来指我的不同方面。我的网址指的是我选择在我的网站上展示的那些方面。以及谷歌、领英、脸书、推特等。大概也有某种 URIs 给我那些他们选择保存的方面贴上了标签。

达成共识的是，资源是名词，而不是动词或形容词。一个写着`http://mybank/myaccount/withdraw`的银行账户的 URL 不被算作资源，因为它包含动词`withdraw`。类似地，`http://amazon.com/buy/book-id`不会标记包含动词`buy`的资源(亚马逊没有这样的 URL)。

这是 REST for HTTP 的第一个关键:识别信息系统中的资源，并为它们分配 URL。这里有一些约定，最常见的是如果有一个层次结构，那么它应该反映在 URL 路径中。然而，这是不必要的，因为信息也应该以其他方式给出。

设计 URIs 的 REST 方法仍然是一种艺术形式。合法的(并且完全合法的)URIs 不一定是“好的”REST URIs，许多所谓 RESTful APIs 的例子都有一点都不 RESTful 的 URIs。2PartsMagic in RESTful URI 设计( [`http://blog.2partsmagic.com/restful-uri-design/`](http://blog.2partsmagic.com/restful-uri-design/) )提供了设计合适的 URIs 的好建议。

## 陈述

资源的表示是以某种形式捕获关于资源的一些信息。例如，我在 URI 的税务局的代表可能是我在澳洲的纳税申报表。我在当地披萨咖啡馆的表现就是购买披萨的记录。在我的网站上，我的一个表现就是一个 HTML 文档。

这是 REST 的关键之一:URIs 识别资源，对该资源的请求返回该资源的表示。资源本身保留在服务器上，根本不会发送到客户端。事实上，这种资源甚至可能根本不存在于任何具体的形式中。例如，一个表示可能是从一个 SQL 查询的结果生成的，该查询是通过向 URI 发出请求而触发的。

REST 并不特别讨论协商资源表示的可能性。HTTP 1.1 有一个关于如何做到这一点的详细章节，考虑了服务器、客户机和透明协商。客户端可以使用`Accept`头来指定，例如:

```go
Accept: application/xml; q=1.0, application/json; q=0.5
Accept-Language: fr
Accept-Charset: utf8

```

这表明它倾向于使用`application/xml`格式，但会接受`application/json`。服务器可以接受其中的一个，或者用它将接受的格式回复。

## REST 动词

你可以向 URI 提出某些要求。如果你向一个 URL 发出一个 HTTP 请求，HTTP 定义了可以发出的请求:`GET`、`PUT`、`POST`、`DELETE`、`HEAD`、`OPTIONS`、`TRACE`和`CONNECT`，以及像`PATCH`这样的扩展。这些数量有限！这与我们对 O/O 编程的期望大相径庭。比如 Java JLabel 有大约 250 个方法，比如`getText`、`setHorizontalAlignment`等。

REST 现在通常被解释为只从 HTTP 中提取四个动词:`GET`、`PUT`、`POST`、`DELETE. GET`大致对应于 O/O 语言的 getter-methods，而`PUT`大致对应于 O/O 语言的 setter-methods。如果 JLabel 是一个 REST 资源(事实并非如此)，那么一个动词如何组成 JLabel 的上百个 getter-methods 呢？

答案在于 URIs 的组成部分。标签具有文本、对齐等属性。这些实际上是标签的子资源，可以写成标签的子 URIs。因此，如果标签有一个 URI`http://jan.newmarch.name/my_label`，那么子资源可以有 URIs:

```go
http://jan.newmarch.name/my_label/text
http://jan.newmarch.name/my_label/horizontalAlignment

```

如果只想操作标签的文本，可以使用文本资源的 URI，而不是标签本身的 getter/setter 方法。

### GET 动词

要检索一个资源的表示，您需要`GET`这个资源。这将返回资源的一些表示。这个选择可能有无数种可能性。例如，对这本书的索引的请求可能返回一个用法语表示的索引，使用 UTF 8 字符集，作为 XML 文档，或者许多其他的可能性。客户端和服务器可以协商这些可能性。

`GET`动词要求是幂等的。也就是说，重复的请求应该返回相同的结果(在表示类型内)。例如，对传感器温度的多次请求应该返回相同的结果(当然，除非温度已经改变)。

缺省情况下，幂等性允许缓存。这有助于减少网络流量，并可节省传感器的电池电量。缓存并不总是有保证的:一个返回其被访问次数的资源在每次被访问时都会给出不同的结果。这是不寻常的行为，将使用 HTTP `Cache-Control`头发出信号。

### PUT 动词

如果你想改变一个资源的状态，你可以`PUT`新的值。`PUT`有两个主要限制:

*   您只能更改 URI 已知的资源的状态
*   您发送的表示必须涵盖资源的所有组件

例如，如果你只想改变一个标签中的文本，你发送`PUT`消息到 URL `http://jan.newmarch.name/my_label/text`，而不是发送到`http://jan.newmarch.name/my_label`。发送到标签需要发送大约一百个字段。

`PUT`是幂等的，但不安全。也就是说，它改变了资源的状态，但是重复的调用会将其改变到相同的状态。

`PUT`和`DELETE`不是 HTML 的一部分，大多数浏览器都不直接支持。可以在支持 Ajax 的浏览器中调用它们。关于为什么不包括它们，有几种讨论。例如，请参见“为什么 HTML 表单上没有 PUT 和 DELETE 方法？” [`http://softwareengineering.stackexchange.com/questions/114156/why-are-there-are-no-put-and-delete-methods-on-html-forms`](http://softwareengineering.stackexchange.com/questions/114156/why-are-there-are-no-put-and-delete-methods-on-html-forms) 。

### 删除动词

这将删除资源。它是幂等的，但不安全。

### 后置动词

是 do-everything-else 动词，用于处理其他动词未涵盖的情况。关于`POST`的两种用法有共识:

*   如果你想创建一个新的资源，但你不知道它的 URI，那么`POST`一个知道如何创建资源的 URI 的资源表示。返回的表示应该包含新资源的 URI。这很重要。要与一个新资源交互，你必须知道它的 URI，从`POST`的返回告诉你这一点。
*   如果一个资源有许多属性，而你只想改变其中的一个或几个，那么`POST`一个只包含改变后的值的表示

关于边缘案例中`PUT`和`POST`各自的作用存在激烈的争论。如果您想创建一个新资源，并且知道它将拥有的 URI，那么您可以使用`PUT`或`POST`。你选择哪一个似乎取决于其他因素…

SOAP 被设计成 HTTP 之上的 RPC 系统。它使用`POST`做任何事情。HTML 继续在表单中使用`POST`，而它应该可以选择使用`PUT`。出于这些原因，除非万不得已，否则我不会使用`POST`。我想其他人使用`POST`而不是`PUT`有他们自己的原则原因，但我不知道他们可能是什么:-)。

由于其开放的范围，`POST`几乎可以用于任何事情。正如 SOAP 充分说明的那样，这些使用中的许多都可能违反 REST 模型。但是其中一些使用可能是合法的。`POST`通常是非幂等的，也不安全，尽管特殊情况下可能是这两种情况之一。

## 无维护状态

让我们预先确定这一点:cookies 是过时的。Cookies 通常用于通过与服务器的交互来跟踪用户的状态，典型的例子是购物车。在服务器端创建一个结构，并返回一个 cookie 来表示这是要使用的购物车。

REST 决定不在服务器上维护任何客户端状态。这简化了交互，也回避了客户机或服务器崩溃后如何恢复一致性的棘手问题。如果服务器不需要维护任何状态，那么它会导致更健壮的服务器模型。

如果你不能使用 cookies，你会怎么做？这其实很简单:在服务器上创建一个购物车。在 REST 下，这只能在响应一个`POST`请求时发生，该请求返回新资源的新 URI。这就是你用的——新的 URI。你可以`GET`、`PUT`、`POST`和`DELETE`到这个 URI，直接在资源上做所有你想做的事情，而不需要用 cookies 做变通。

## 恨死我了

HATEOAS 代表“作为应用程序状态引擎的超媒体”。人们普遍认为这是一个糟糕的首字母缩写，但它一直存在。基本原则是，从一个 URI 导航到以某种方式相关的另一个，不应该由任何带外机制来完成，而是新链接必须以某种方式作为超链接嵌入第一个 URI 的表示中。

REST 没有说明链接的格式。它们可以通过 HTML 链接标签、嵌入在 PDF 文档中的 URL 或 XML 文档中的链接给出。没有简单 URL 表示的格式不被认为是超媒体语言，也不包含在 REST 中。

另外，REST 也没有明确说明链接的含义，也没有说明如何提取适当的链接。菲尔丁在 [`http://roy.gbiv.com/untangled/2008/rest-apis-must-be-hypertext-driven`](http://roy.gbiv.com/untangled/2008/rest-apis-must-be-hypertext-driven) 的博客《REST APIs 必须是超文本驱动的》中写道:

*   输入 REST API 时，除了初始 URI(书签)和一组适用于目标受众的标准化媒体类型(即，任何可能使用该 API 的客户端都应该理解)之外，不应该有任何先验知识。从那时起，所有应用程序状态转换都必须由客户端选择服务器提供的选项来驱动，这些选项出现在接收到的表示中，或者由用户对这些表示的操作来暗示。

IANA 维护一个可以使用的关系类型注册表(IANA:在 [`http://www.iana.org/assignments/link-relations/link-relations.xhtml`](http://www.iana.org/assignments/link-relations/link-relations.xhtml) 的链接关系)。网页链接 RFP5988 描述了网页链接注册表。HTML5 规范定义了少量的关系，并在 [`http://microformats.org/wiki/existing-rel-values#HTML5_link_type_extensions`](http://microformats.org/wiki/existing-rel-values#HTML5_link_type_extensions) 处指向微格式`rel`值以获得更大的列表。

像 cookies 这样的机制，或者像 SOAP 的 WSDL 这样的外部 API 规范，都被 REST 有效地排除了。它们不是包含在资源表示中的超链接。

## 表示链接

HTML 文档中的链接是标准化的。标签定义了一个只能出现在 HTML 头部分的 HTML 元素。比如有章节的书等等。，如果链接以 HTML `link`元素的形式给出，可能是这样的:

```go
<html>
  <head>
   <link rel= "author" title="Jan Newmarch" href="https://jan.newmarch.name">
   <link rel="chapter" title="Introduction" href="Introduction/">
    ...

```

HTML 中的链接关系有两种类型:一种是当前文档需要的，比如 CSS 文件，另一种是指向相关资源的，如上所述。第一种类型通常是用户不可见地下载的。第二种类型一般不会被浏览器显示，但是遵循 HATEOAS 原则的用户代理会使用它们。

XML 有各种各样的链接规范。其中包括 XLink 和 Atom。Atom 似乎更受欢迎。

基于 XLink 的链接如下所示:

```go
<People xmlns:xlink="http://www.w3.org/1999/xlink">
 <Person xlink:type="simple" xlink:href="http://...">
   ...
 </Person>
   ...
</People>

```

基于 Atom 的链接如下所示:

```go
<People xmlns:atom="http://www.w3.org/2005/Atom">
 <Person>
   <link  atom:href="http://..."/>
   ...
 </Person>
 ...
</People>

```

对于 JSON，这种格式是不规范的。REST cookbook ( [`http://restcookbook.com/Mediatypes/json/`](http://restcookbook.com/Mediatypes/json/) )指出了标准化的缺乏，并指出 W3C 规范 JSON-LD 1.0:“基于 JSON 的链接数据序列化”和 HAL(超文本应用语言)。像开放连接基金会这样的机构似乎使用他们自己开发的格式，但那是针对 CoAP 的，另一个基于 REST 的系统。

JSON-LD 使用术语`@id`来表示 URL，如:

```go
{
  "name": Jan Newmarch:,
  "homepage": {"@id": "https://jan.newmarch.name/"}
}

```

在这方面值得注意的是，W3C 也有一个在 [`https://www.w3.org/wiki/LinkHeader`](https://www.w3.org/wiki/LinkHeader) 的 HTTP 链接头的规范，它可以由服务器返回给客户端。例如，JSON-LD 使用它来指向包含在 HTTP 响应主体中的 JSON 文档的规范。

这可能会影响将链接信息从服务器传递到用户代理的序列化方法。用户代理和服务器必须就要使用的格式达成一致。对于 HTML(或 XHTML)，这是标准化的。对于 XML，可以在文档中引用链接系统。对于 JSON-LD，这可以在`Accept HTTP`头中作为`application/ld+json`发出信号。

## REST 交易

REST 是如何处理事务以及其他进程的？在菲尔丁的原始论文中没有讨论它们。

HATEOAS 的 Wikipedia 条目给出了一个管理事务的糟糕例子。它从这样一个 HTTP 请求开始:

```go
GET /account/12345 HTTP/1.1
Host: somebank.org
Accept: application/xml
 ...

```

它返回一个 XML 文档作为帐户的表示:

```go
HTTP/1.1 200 OK
    Content-Type: application/xml
    Content-Length: ...

    <?xml version="1.0"?>
    <account>
       <account_number>12345</account_number>
       <balance currency="usd">100.00</balance>
       <link rel="deposit" href="http://somebank.org/account/12345/deposit" />
       <link rel="withdraw" href="http://somebank.org/account/12345/withdraw" />
       <link rel="transfer" href="http://somebank.org/account/12345/transfer" />
       <link rel="close" href="http://somebank.org/account/12345/close" />
     </account>

```

这提供了相关资源存放、提取、转移和关闭的 URIs。然而，资源是动词而不是名词，这一点也不好。它们如何与 HTTP 动词交互？你要退出吗？`POST`它？`PUT`它？如果您`DELETE`撤销，会发生什么——是要回滚事务还是什么？

更好的方法，例如，在 Stackoverflow 发布的“REST 中的事务？”(参见 [`http://stackoverflow.com/questions/147207/transactions-in-rest`](http://stackoverflow.com/questions/147207/transactions-in-rest) )是向账户`POST`请求创建新的交易:

```go
POST /account/12345/transaction HTTP/1.1

```

这将返回新交易的 URL:

```go
http://account/12345/txn123

```

现在使用这个事务 URL 进行交互，例如通过使用一个新的值来执行和提交事务。

```go
PUT /account/12345/txn123
<transaction>
  <from>/account/56789</from>
  <amount>100</amount>
</transaction>  

```

Mihindukulasooriya 等人给出了关于事务和 REST 的更详细的讨论。al 在《RESTful 事务模型的七大挑战》(见 [`http://ws-rest.org/2014/sites/default/files/wsrest2014_submission_4.pdf`](http://ws-rest.org/2014/sites/default/files/wsrest2014_submission_4.pdf) )。类似的模型被提议用于管理不仅仅是单一步骤的过程。

## 理查森成熟度模型

许多系统声称是 RESTful 的。大多数都不是。我甚至碰到一个声称 SOAP 是 RESTful 的，这是一个扭曲的精神状态的明显例子。Martin Fowler 讨论了 Richardson 成熟度模型，该模型根据系统对 REST 的符合性对系统进行分类。(见 [`https://martinfowler.com/articles/richardsonMaturityModel.html`](https://martinfowler.com/articles/richardsonMaturityModel.html) )。)

0 级

*   该模型的起点是使用 HTTP 作为远程交互的传输系统，但不使用任何 web 机制。本质上，您在这里所做的是使用 HTTP 作为您自己的远程交互机制的隧道机制，通常基于远程过程调用。

级别 1:资源

*   走向 RMM 中 Rest 辉煌的第一步是引入资源。因此，现在我们开始与单个资源对话，而不是向单个服务端点发出所有请求。

第二层:HTTP 动词

*   我在 0 级和 1 级的所有交互中都使用了 HTTP POST 动词，但是有些人使用 GETs 来代替或附加。在这些级别上，没有太大的区别，它们都被用作隧道机制，允许您通过 HTTP 隧道化您的交互。第 2 级远离了这一点，使用了尽可能接近 HTTP 本身使用方式的 HTTP 动词。

级别 3:超媒体控件

*   最后一层引入了你经常听到的 HATEOAS(作为应用程序状态引擎的超文本)的丑陋缩写。它解决了如何从空缺职位列表中了解如何预约的问题。

## 重温抽认卡

在第 [10](10.html) 章中，我们考虑了一个由服务器和浏览器中呈现的 HTML 页面组成的 web 系统，使用 JavaScript 和 CSS 来控制浏览器端的交互。没有试图做任何特别结构化的东西，而只是作为一个传统的网络系统。

Recap

第 10 章的网络系统被用来演示使用所谓的抽认卡学习语言。一次给用户一组卡片，显示一种语言的单词，然后希望记住翻译，通过“翻转”卡片来显示。该系统提供了一个不同卡片组的列表，然后在选定的卡片组中一次显示一张卡片。

我们现在处理与使用 REST 方法构建的 HTTP 客户机-服务器系统相同的情况。我们将做出一些改变:

*   将根据具体情况给出相应的 URL。这将包括“根”URL `/`以及每个抽认卡组的 URL，此外，还有每个抽认卡的 URL。
*   所有用户交互代码(HTML、JavaScript 和 CSS)都被省略了。服务器将与一个任意的用户代理对话，许多人不理解 UI 代码。
*   服务器不会维护或管理任何客户端状态。在 web 示例中，表单数据从浏览器发送到服务器，服务器立即以略有不同的形式返回。想要维护状态的客户端应该自己来做。
*   服务器将被设置为管理许多不同的序列化格式，并将在客户端-服务器协商后适当地交付。
*   大量使用 HTTP 机制，特别是错误处理和内容协商。

### 资源定位符

该系统的 URL 以及可以执行的操作如下:

<colgroup><col> <col> <col></colgroup> 
| 统一资源定位器 | 行动 | 影响 |
| `/` | `GET` | 获取抽认卡集的列表 |
| `POST` | 创建新的抽认卡集 |
| `/flashcardSets/<set>` | `GET` | 获取该集中的卡片列表 |
| `POST` | 为集合创建一张新卡 |
| `DELETE` | 如果抽认卡为空，则删除抽认卡集 |
| `/flashcardSets/<set>/<card>` | `GET` | 获取卡片的内容 |
| `DELETE` | 从牌集中删除该牌 |

这与第 [10](10.html) 章中描述的系统略有不同。主要的结构差异是每张卡都有自己的 URL 作为抽认卡组的成员。

将由服务器处理的示例 URL 包括:

<colgroup><col> <col> <col></colgroup> 
| 根 URL | 抽认卡组的 URL | 一个抽认卡的 URL |
| `/` | `/flashcardSet/CommonWords` | `/flashcardSet/CommonWords/你好` |

## 多路分解器

REST 基于应用于 URL 的少量动作。试图使用 REST 原则的系统必须是基于 URL 的。

服务器解复用器将根据 URL 模式检查客户端和调用处理程序请求的 URL。标准的 Go demuxer `ServeMux`使用了一种特殊的模式匹配机制:如果一个 URL 以`/`结尾，那么它表示以该 URL 为根的 URL 的子树。如果它以没有`/`结尾，那么它只代表那个 URL。将 URL 与具有最长模式匹配的处理程序进行匹配。

我们需要一个根 URL 的处理程序`/`。这也将匹配任何 URL，如`/passwords`，除非另一个处理程序捕捉到它。在这个系统中，没有其他处理程序会这样做，所以在`/`的处理程序中，我们需要为这样的尝试返回错误。

一个棘手的部分出现了，因为我们对我们的 URL 使用了层次结构。一组特殊的抽认卡是`/flashcardSets/CommonWords`。这将实际上是一个特定的一套卡的目录。我们必须注册两个处理程序:一个用于 URL `/flashcardSets/CommonWords`，这是抽认卡集资源，另一个用于`/flashcardSets/CommonWords/`(注意后面的`/`)，这是包含各个卡及其 URL 的子树。

注册这些的主函数中的代码如下:

```go
http.HandleFunc(`/`, handleFlashCardSets)
files, err := ioutil.ReadDir(`flashcardSets`)
checkError(err)
for _, file := range files {
        cardset_url := `/flashcardSets/` + url.QueryEscape(file.Name())
        http.HandleFunc(cardset_url, handleOneFlashCardSet)
        http.HandleFunc(cardset_url + `/`, handleOneFlashCard)
}        

```

注意我们有函数`QueryEscape`。这是为了转义 URL 中可能出现的任何特殊字符。例如，文件名中的`$`应该编码为`%44;`。我们确实需要使用这样一个函数:我们的 URL 将包含中文字符，这些字符需要进行转义编码才能在 URL 中表示。这是由`QueryEscape`完成的，只有一个例外:路径中的空格应该编码为`%20`，但是在表单数据中应该编码为`+`。`PathEscape`函数可以正确地做到这一点，但在 Go 1.8 之前不可用。我们将删除网址中的空格以避免这个问题。

## 内容协商

任何 web 用户代理都可以尝试与任何 web 服务器对话。浏览器与 HTML 服务器对话的典型情况是我们在 Web 上所熟悉的，但许多人会熟悉使用其他用户代理，如 curl、wget，甚至 telnet！浏览器和其他工具将使用 HTTP 回复中的`Content-Type`来决定如何处理提供的内容。

对于一个 Web 应用程序，用户代理必须能够理解服务器正在交付什么，因为它试图在一个可能没有用户帮助的交互中扮演一个角色。RPC 系统通常使用客户机和服务器都遵守的外部规范。这里的情况并非如此。

解决方案是双方必须就内容格式达成一致。这是在 HTTP 级别完成的。客户端将声明它将接受一系列格式。如果服务器同意，那么他们继续。如果没有，服务器将告诉客户端它可以接受哪些格式，如果可能的话，客户端可以重新开始。

协商使用 MIME 类型。标准的有几百种:`text/html`、`application/pdf`、`application/xml`、…。浏览器可以呈现它收到的任何 HTML 文档。像 VLC 这样支持 HTTP 的音乐播放器可以播放它接收到的任何 MP3 文件。但是对于抽认卡应用程序，它不能处理任何通用格式，只能处理符合预期结构的消息。这些都不是适用于为这个抽认卡应用程序协商专用协议的标准 MIME 类型。所以，我们自己编。客户机和服务器必须知道它们正在处理一个共享的 MIME 类型，否则它们就不能正常对话。

IANA 有自己创造哑剧类型的规则。我使用类型`application/x.flashcards`。服务器将能够交付 JSON 和 XML，因此两种可接受的 MIME 类型是`application/x.flashcards+xml`和`application/x.flashcards+json`。

HTTP 内容协商表示用户代理可以建议一个可接受格式的列表，权重在 0 到 1 之间，如下所示:

```go
Accept: application/x.flashcards+xml; q=0.8,
        application/x.flashcards+json; q=0.4

```

服务器可以检查请求，并决定是否可以处理该格式。我们在服务器中使用以下代码来确定用户代理是否请求了任何类型的请求，以及请求的权重(零表示未请求):

```go
const flashcard_xml string = "application/x.flashcards+xml"
const flashcard_json string = "application/x.flashcards+json"

type ValueQuality struct {
        Value   string
        Quality float64
}

/* Based on https://siongui.github.io/2015/02/22/go-parse-accept-language/ */
func parseValueQuality(s string) []ValueQuality {
        var vqs []ValueQuality

        strs := strings.Split(s, `,`)
        for _, str := range strs {
                trimmedStr := strings.Trim(str, ` `)
                valQ := strings.Split(trimmedStr, `;`)
                if len(valQ) == 1 {
                        vq := ValueQuality{valQ[0], 1}
                        vqs = append(vqs, vq)
                } else {
                        qp := strings.Split(valQ[1], `=`)
                        q, err := strconv.ParseFloat(qp[1], 64)
                        if err != nil {
                                q = 0
                        }
                        vq := ValueQuality{valQ[0], q}
                        vqs = append(vqs, vq)
                }
        }
        return vqs
}

func qualityOfValue(value string, vqs []ValueQuality) float64 {
        for _, vq := range vqs {
                if value == vq.Value {
                        return vq.Quality
                }

        }
        // not found
        return 0
}

```

如果服务器不接受用户代理请求的任何类型，它将返回一个 HTTP 代码 406 `"Not acceptable"`，并提供一个接受的格式列表。在服务器中执行此操作的代码段如下:

```go
func handleFlashCardSets(rw http.ResponseWriter, req *http.Request) {

        if req.Method == "GET" {
                acceptTypes := parseValueQuality(req.Header.Get("Accept"))

                q_xml := qualityOfValue(flashcard_xml, acceptTypes)
                q_json := qualityOfValue(flashcard_json, acceptTypes)
                if q_xml == 0 && q_json == 0 {
                        // can't find XML or JSON in Accept header
                        rw.Header().Set("Content-Type", flashcard_xml + `, ` + flashcard_json)
                        rw.WriteHeader(http.StatusNotAcceptable)
                        return
                }
                ...

```

这说明了 HTTP 服务器的一个常见 REST 模式:给定一个 HTTP 请求，检查它以查看服务器是否能够管理它。如果没有，返回一个 HTTP 错误。如果可以，尝试处理它。如果尝试失败，返回一个 HTTP 错误。成功时，返回适当的 HTTP 成功代码和结果。

### 获取/

抽认卡组都存放在目录`/flashcardSets`中。`GET /`请求需要列出所有这些文件，并为客户机准备合适的格式。该格式是一个抽认卡集合名称及其 URL 的列表。HATEOAS 需要 URL:名称列表告诉我们集合是什么，但是客户端需要它们的 URL，以便进入与其中一个进行交互的阶段。

服务器中每个`FlashcardSet`的数据类型包含集合的名称及其 URL(字符串形式):

```go
type FlashcardSet struct {
        Name string
        Link string
}

```

服务器上的抽认卡组可以从抽认卡组的目录中建立。`ioutil.ReadDir()`将创建一个`os.FileI` nfo 的数组。这需要转换成如下文件名列表:

```go
files, err := ioutil.ReadDir(`flashcardSets`)
checkError(err)
numfiles := len(files)
cardSets := make([]FlashcardSet, numfiles, numfiles)
for n, file := range files {
        cardSets[n].Name = file.Name()
        // should be PathEscape, not in go 1.6
        cardSets[n].Link = `/flashcardSets/` + url.QueryEscape(file.Name())
}

```

这将创建一个文件名数组和到服务器上资源的相对链接作为`/<name>`。对于`CommonWords`集合，相对链接 URL 将是`/flashcardSets/CommonWords`。方案(`http`或`https`)和主机(如“localhost”)由客户端自行解决。

不幸的是，文件名可能包含在 URL 路径名中不合法的字符。函数`url.PathEscape`正确地将它们全部转义，但是直到 Go 1.8 才可用。函数`url.QueryEscape`除了文件名中的空格外，其他都是正确的，它用`+`代替了`%20;`。

最后，服务器判断 JSON 或 XML 是首选，并通过一个模板运行它，为客户机生成正确的输出。对于 XML，模板代码如下:

```go
t, err := template.ParseFiles("xml/ListFlashcardSets.xml")
if err != nil {
        // parse error occurred in the template. Our error
        http.Error(rw, err.Error(), http.StatusInternalServerError)
        return
}
rw.Header().Set("Content-Type", flashcard_xml)
t.Execute(rw, cardSets)

```

XML 模板如下所示:

```go
<?xml version="1.0" encoding="UTF-8"?>

<cardsets >
  {{range .}}
  <cardset href="{{.Link}}">
    <name>
      {{.Name}}
    </name>
  </cardset>
  {{end}}
</cardsets>

```

对于只有两个集合`CommonWords`和`Lesson04`的列表，发送到客户端的内容如下:

```go
<?xml version="1.0" encoding="UTF-8"?>

<cardsets >

  <cardset href="/CommonWords">
    <name>
      Common Words
    </name>
  </cardset>

  <cardset href="/Lesson04">
    <name>
      Lesson04
    </name>
  </cardset>

</cardsets>

```

### 员额/

这里，一个客户要求创建一个新的抽认卡组。期望客户提供抽认卡组的名称。我们让它看起来像表单提交数据:

```go
name=<new flashcard set name>

```

这比这种情况下的`GET`简单多了。以表单数据的形式从请求中获取值。然后检查所请求的名字中没有不愉快的东西，比如调用抽认卡集`/etc/passwd`。如果是，返回 403 `"Forbidden"`。如果看起来没问题，用那个名字创建一个目录。如果失败，再次返回 403(该目录可能已经存在)。否则，返回 201 `"Created"`和新的相对 URL:

```go
if req.Method == "POST" {
      name := req.FormValue(`name`)
      if hasIllegalChars(name) {
              rw.WriteHeader(http.StatusForbidden)
              return
}
// lose all spaces as they are a nuisance
name = strings.Replace(name, ` `, ``, -1)

err := os.Mkdir(`flashcardSets/`+name,
              (os.ModeDir | os.ModePerm))
if err != nil {
      rw.WriteHeader(http.StatusForbidden)
      return
}
rw.WriteHeader(http.StatusCreated)
base_url := req.URL.String()
new_url := base_url + `flashcardSets/` + name
rw.Write([]byte(new_url))       

```

## 处理其他 URL

我们讨论了服务器处理带有`GET`和`POST`请求的`/` URL 的代码。这个应用程序还有另外两种类型的 URL 处理一组卡片和处理每张单独的卡片。然而，就编码而言，这并没有什么新的想法。

*   获得一组中的卡片列表是另一个目录列表。
*   向器械组发送新卡意味着在适当的目录中创建一个包含客户端内容的文件。
*   删除集合意味着删除一个目录。如果目录为空，这是可以的，否则会产生错误。
*   获取卡意味着读取卡文件并发送其内容。
*   删除卡意味着删除文件。

这些都没有什么特别新的东西。我们还没有完成一些操作的代码，比如`DELETE`:这些返回 HTTP 代码 501 `'Not implemented'`。我们也以`text/plain`的形式返回各个卡片的内容:它们有一个复杂的 JSON/Go 结构，如第 [10 章](10.html)中所使用的，但这对于讨论这个系统的其他方面是不需要的。

## 完整的服务器

接下来是处理对`/`的请求以及从那里到其他 URL 的请求的完整服务器。它需要抽认卡组和独立卡才能运行，这些都在`Ch14`文件夹中 [`http://www.apress.com/9781484226919`](http://www.apress.com/9781484226919) 的 ZIP 文件中。

```go
/* Server
 */

package main

import (
        "fmt"
        "html/template"
        "io/ioutil"
        "net/http"
        "net/url"
        "os"
        "regexp"
        "strconv"
        "strings"
)

type FlashcardSet struct {
        Name string
        Link string
}

type Flashcard struct {
        Name string
        Link string
}

const flashcard_xml string = "application/x.flashcards+xml"
const flashcard_json string = "application/x.flashcards+json"

type ValueQuality struct {
        Value   string
        Quality float64
}

/* Based on https://siongui.github.io/2015/02/22/go-parse-accept-language/ */
func parseValueQuality(s string) []ValueQuality {
        var vqs []ValueQuality

        strs := strings.Split(s, `,`)
        for _, str := range strs {
                trimmedStr := strings.Trim(str, ` `)
                valQ := strings.Split(trimmedStr, `;`)
                if len(valQ) == 1 {
                        vq := ValueQuality{valQ[0], 1}
                        vqs = append(vqs, vq)
                } else {
                        qp := strings.Split(valQ[1], `=`)
                        q, err := strconv.ParseFloat(qp[1], 64)
                        if err != nil {
                                q = 0
                        }
                        vq := ValueQuality{valQ[0], q}
                        vqs = append(vqs, vq)
                }
        }
        return vqs
}

func qualityOfValue(value string, vqs []ValueQuality) float64 {
        for _, vq := range vqs {
                if value == vq.Value {
                        return vq.Quality
                }

        }

        return 0
}

func main() {
        if len(os.Args) != 2 {
                fmt.Fprint(os.Stderr, "Usage: ", os.Args[0], ":port\n")
                os.Exit(1)
        }
        port := os.Args[1]

        http.HandleFunc(`/`, handleFlashCardSets)
        files, err := ioutil.ReadDir(`flashcardSets`)
        checkError(err)
        for _, file := range files {
                fmt.Println(file.Name())
                cardset_url := `/flashcardSets/` + url.QueryEscape(file.Name())
                fmt.Println("Adding handlers for ", cardset_url)
                http.HandleFunc(cardset_url, handleOneFlashCardSet)
                http.HandleFunc(cardset_url + `/`, handleOneFlashCard)
        }

        // deliver requests to the handlers
        err = http.ListenAndServe(port, nil)
        checkError(err)
        // That's it!
}

func hasIllegalChars(s string) bool {
        // check against chars to break out of current dir
        b, err := regexp.Match("[/$∼]", []byte(s))
        if err != nil {
                fmt.Println(err)
                return true
        }
        if b {
                return true
        }
        return false
}

func handleOneFlashCard(rw http.ResponseWriter, req *http.Request) {
        // should be PathUnescape
        path, _ := url.QueryUnescape(req.URL.String())
        // lose initial '/'
        path = path[1:]
        if req.Method == "GET" {
                fmt.Println("Handling card: ", path)
                json_contents, err := ioutil.ReadFile(path)
                if err != nil {
                        rw.WriteHeader(http.StatusNotFound)
                        rw.Write([]byte(`Resource not found`))
                        return
                }
                // Be lazy here, just return the content as text/plain
                rw.Write(json_contents)

                return
        } else if req.Method == "DELETE" {
                rw.WriteHeader(http.StatusNotImplemented)
        } else {
                rw.WriteHeader(http.StatusMethodNotAllowed)
        }
        return
}

func handleFlashCardSets(rw http.ResponseWriter, req *http.Request) {
        if req.URL.String() != `/` {
                // this function only handles '/'
                rw.WriteHeader(http.StatusNotFound)
                rw.Write([]byte("Resource not found\n"))
                return
        }
        if req.Method == "GET" {
                acceptTypes := parseValueQuality(req.Header.Get("Accept"))
                fmt.Println(acceptTypes)

                q_xml := qualityOfValue(flashcard_xml, acceptTypes)
                q_json := qualityOfValue(flashcard_json, acceptTypes)
                if q_xml == 0 && q_json == 0 {
                        // can't find XML or JSON in Accept header
                        rw.Header().Set("Content-Type", flashcard_xml + `, ` + flashcard_json)
                        rw.WriteHeader(http.StatusNotAcceptable)
                        return
                }

                files, err := ioutil.ReadDir(`flashcardSets`)
                checkError(err)
                numfiles := len(files)
                cardSets := make([]FlashcardSet, numfiles, numfiles)
                for n, file := range files {
                        fmt.Println(file.Name())
                        cardSets[n].Name = file.Name()
                        // should be PathEscape, not in go 1.6
                        cardSets[n].Link = `/flashcardSets/` + url.QueryEscape(file.Name())
                }

                if q_xml >= q_json {
                        // XML preferred
                        t, err := template.ParseFiles("xml/ListFlashcardSets.xml")
                        if err != nil {
                                fmt.Println("Template error")
                                http.Error(rw, err.Error(), http.StatusInternalServerError)
                                return
                        }
                        rw.Header().Set("Content-Type", flashcard_xml)
                        t.Execute(rw, cardSets)
                } else {
                        // JSON preferred

                        t, err := template.ParseFiles("json/ListFlashcardSets.json")
                        if err != nil {
                                fmt.Println("Template error")
                                http.Error(rw, err.Error(), http.StatusInternalServerError)
                                return
                        }
                        rw.Header().Set("Content-Type", flashcard_json)
                        t.Execute(rw, cardSets)

                }
        } else if req.Method == "POST" {
                name := req.FormValue(`name`)
                if hasIllegalChars(name) {
                        rw.WriteHeader(http.StatusForbidden)
                        return
                }
                // lose all spaces as they are a nuisance
                name = strings.Replace(name, ` `, ``, -1)
                err := os.Mkdir(`flashcardSets/`+name,
                        (os.ModeDir | os.ModePerm))
                if err != nil {
                        rw.WriteHeader(http.StatusForbidden)
                        return
                }
                rw.WriteHeader(http.StatusCreated)
                base_url := req.URL.String()
                new_url := base_url + `flashcardSets/` + name
                // add handlers for the resources
                http.HandleFunc(new_url, handleOneFlashCardSet)
                http.HandleFunc(new_url + `/`, handleOneFlashCard)
                rw.Write([]byte(new_url))
        } else {
                rw.WriteHeader(http.StatusMethodNotAllowed)
        }
        return
}

func handleOneFlashCardSet(rw http.ResponseWriter, req *http.Request) {
        cooked_url, _ := url.QueryUnescape(req.URL.String())
        fmt.Println("Handling one set for: ", cooked_url)

        if req.Method == "GET" {
                acceptTypes := parseValueQuality(req.Header.Get("Accept"))
                fmt.Println(acceptTypes)

                q_xml := qualityOfValue(flashcard_xml, acceptTypes)
                q_json := qualityOfValue(flashcard_json, acceptTypes)
                if q_xml == 0 && q_json == 0 {
                        // can't find XML or JSON in Accept header
                        rw.Header().Set("Content-Type", flashcard_xml + `, ` + flashcard_json)
                        rw.WriteHeader(http.StatusNotAcceptable)
                        return

                }

                path := req.URL.String()
                // lose leading /
                relative_path := path[1:]
                files, err := ioutil.ReadDir(relative_path)
                checkError(err)
                numfiles := len(files)
                cards := make([]Flashcard, numfiles, numfiles)
                for n, file := range files {
                        fmt.Println(file.Name())
                        cards[n].Name = file.Name()
                        // should be PathEscape, not in go 1.6
                        cards[n].Link = path + `/` + url.QueryEscape(file.Name())
                }

                if q_xml >= q_json {
                        // XML preferred
                        t, err := template.ParseFiles("xml/ListOneFlashcardSet.xml")
                        if err != nil {
                                fmt.Println("Template error")
                                http.Error(rw, err.Error(), http.StatusInternalServerError)
                                return
                        }
                        rw.Header().Set("Content-Type", flashcard_xml)
                        t.Execute(os.Stdout, cards)
                        t.Execute(rw, cards)
                } else {
                        // JSON preferred
                        t, err := template.ParseFiles("json/ListOneFlashcardSet.json")
                        if err != nil {
                                fmt.Println("Template error")
                                http.Error(rw, err.Error(), http.StatusInternalServerError)
                                return
                        }
                        rw.Header().Set("Content-Type", flashcard_json)
                        t.Execute(rw, cards)

                }
        } else if req.Method == "POST" {
                name := req.FormValue(`name`)
                if hasIllegalChars(name) {
                        rw.WriteHeader(http.StatusForbidden)
                        return
                }
                err := os.Mkdir(`flashcardSets/`+name,
                        (os.ModeDir | os.ModePerm))
                if err != nil {
                        rw.WriteHeader(http.StatusForbidden)
                        return

                }
                rw.WriteHeader(http.StatusCreated)
                base_url := req.URL.String()
                new_url := base_url + `flashcardSets/` + name
                _, _ = rw.Write([]byte(new_url))
        } else if req.Method == "DELETE" {
                rw.WriteHeader(http.StatusNotImplemented)
        } else {
                rw.WriteHeader(http.StatusMethodNotAllowed)
        }
        return
}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

它的运行方式如下:

```go
go run Server.go :8000

```

## 客户

客户端相对简单，不提供任何新的东西。这个客户端只要求 XML 格式的内容。一个新的部分是抽认卡集的内容在一个`cardset`标签中包含了作为超文本属性的链接。这可以通过`Card`结构中的标签标签`xml:"href,attr"`转换成结构的字段。

这个客户端在`getFlashcardSets()`函数中获得抽认卡集合及其 URL 的列表(步骤 1)。这将返回一个`FlashcardSets`结构。这可以用来向用户呈现一个列表，比如说，供用户选择一个特定的集合。一旦选中，该集合的 URL 就可以用来与资源进行交互。

然后，这个客户端在`createFlashcardSet()`函数中创建一个名为`NewSet`的新抽认卡集(步骤 2)。第一次运行时，它将创建集合并返回该集合的 URL。第二次运行时，它将从服务器得到一个错误，作为禁止的操作，因为该集合已经存在。

然后，这个客户机从服务器给出的 URL 中取出第一组抽认卡，并请求它持有的那组抽认卡(步骤 3)。然后，它从该组中挑选第一张牌，并获取其内容(步骤 4)。

客户是`Client.go`:

```go
/* Client
 */

package main

import (
        //"encoding/json"
        "encoding/xml"
        "fmt"
        "io/ioutil"
        "net/http"
        "net/http/httputil"
        "net/url"
        "os"
        "strings"
)

const flashcard_xml string = "application/x.flashcards+xml"
const flashcard_json string = "application/x.flashcards+json"

type FlashcardSets struct {
        XMLName string `xml:"cardsets"`
        CardSet    []CardSet `xml:"cardset"`
}

type CardSet struct {
        XMLName string `xml:"cardset"`
        Name string `xml:"name"`
        Link string `xml:"href,attr"`
        Cards []Card `xml:"card"`
}

type Card  struct {
        Name string `xml:"name"`
        Link string `xml:"href,attr"`
}

func getOneFlashcard(url *url.URL, client *http.Client) string {
        // Get the card as a string, don't do anything with it
        request, err := http.NewRequest("GET", url.String(), nil)
        checkError(err)

        response, err := client.Do(request)
        checkError(err)
        if response.Status != "200 OK" {
                fmt.Println(response.Status)
                fmt.Println(response.Header)

                os.Exit(2)
        }

        fmt.Println("The response header is")
        b, _ := httputil.DumpResponse(response, false)
        fmt.Print(string(b))

        body, err := ioutil.ReadAll(response.Body)
        content := string(body[:])
        //fmt.Printf("Body is %s", content)

        return content
}

func getOneFlashcardSet(url *url.URL, client *http.Client) CardSet {
        // Get one set of cards
        request, err := http.NewRequest("GET", url.String(), nil)
        checkError(err)

        // only accept our media types
        request.Header.Add("Accept", flashcard_xml)
        response, err := client.Do(request)
        checkError(err)
        if response.Status != "200 OK" {
                fmt.Println(response.Status)
                fmt.Println(response.Header)

                os.Exit(2)
        }

        fmt.Println("The response header is")
        b, _ := httputil.DumpResponse(response, false)
        fmt.Print(string(b))

        body, err := ioutil.ReadAll(response.Body)
        content := string(body[:])
        fmt.Printf("Body is %s", content)

        var sets CardSet
        contentType := getContentType(response)
        if contentType == "XML" {

                err = xml.Unmarshal(body, &sets)
                checkError(err)
                fmt.Println("XML: ", sets)
                return sets
        }
        /* else if contentType == "JSON" {
                var sets FlashcardSetsJson
                err = json.Unmarshal(body, &sets)
                checkError(err)
                fmt.Println("JSON: ", sets)
        }
        */
        return sets

}

func getFlashcardSets(url *url.URL, client *http.Client) FlashcardSets {
        // Get the toplevel /
        request, err := http.NewRequest("GET", url.String(), nil)
        checkError(err)

        // only accept our media types
        request.Header.Add("Accept", flashcard_xml)
        response, err := client.Do(request)
        checkError(err)
        if response.Status != "200 OK" {
                fmt.Println(response.Status)
                fmt.Println(response.Header)

                os.Exit(2)
        }

        fmt.Println("The response header is")
        b, _ := httputil.DumpResponse(response, false)
        fmt.Print(string(b))

        body, err := ioutil.ReadAll(response.Body)
        content := string(body[:])
        fmt.Printf("Body is %s", content)

        var sets FlashcardSets
        contentType := getContentType(response)
        if contentType == "XML" {

                err = xml.Unmarshal(body, &sets)
                checkError(err)
                fmt.Println("XML: ", sets)
                return sets
        }
        return sets
}

func createFlashcardSet(url1 *url.URL, client *http.Client, name string) string {
        data := make(url.Values)
        data[`name`] = []string{name}
        response, err := client.PostForm(url1.String(), data)
        checkError(err)
        if response.StatusCode != http.StatusCreated {
                fmt.Println(`Error: `, response.Status)
                return ``
                //os.Exit(2)
        }
        body, err := ioutil.ReadAll(response.Body)
        content := string(body[:])
        return content

}

func main() {
        if len(os.Args) != 2 {
                fmt.Println("Usage: ", os.Args[0], "http://host:port/page")
                os.Exit(1)
        }
        url, err := url.Parse(os.Args[1])
        checkError(err)

        client := &http.Client{}

        // Step 1: get a list of flashcard sets
        flashcardSets := getFlashcardSets(url, client)
        fmt.Println("Step 1: ", flashcardSets)

        // Step 2: try to create a new flashcard set
        new_url := createFlashcardSet(url, client, `NewSet`)
        fmt.Println("Step 2: New flashcard set has URL: ", new_url)

        // Step 3: using the first flashcard set,
        //         get the list of cards in it
        set_url, _ := url.Parse(os.Args[1] + flashcardSets.CardSet[0].Link)

        fmt.Println("Asking for flashcard set URL: ", set_url.String())
        oneFlashcardSet := getOneFlashcardSet(set_url, client)
        fmt.Println("Step 3:", oneFlashcardSet)

        // Step 4: get the contents of one flashcard
        //         be lazy, just get as text/plain and
        //         don't do anything with it
        card_url, _ :=  url.Parse(os.Args[1] + oneFlashcardSet.Cards[0].Link)
        fmt.Println("Asking for URL: ", card_url.String())
        oneFlashcard := getOneFlashcard(card_url, client)
        fmt.Println("Step 4", oneFlashcard)
        os.Exit(0)
}

func getContentType(response *http.Response) string {
        contentType := response.Header.Get("Content-Type")
        if strings.Contains(contentType, flashcard_xml) {
                return "XML"
        }
        if strings.Contains(contentType, flashcard_json) {
                return "JSON"
        }
        return ""
}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

它的运行方式如下:

```go
go run Client.go http://localhost:8000/

```

## 使用 REST 或 RPC

REST 和 RPC 的主要区别在于交互风格。在 RPC 中，您调用函数，将对象或基本类型作为参数传递，并获得对象或基本类型作为回报。这些功能是动词:做这个做那个。另一方面，REST 是关于与对象的交互，要求它们显示它们的状态或者以某种方式改变它。

上一章讨论的 Go RPC 机制和本章的 REST 机制显示了这种差异。在 Go RPC over HTTP 中，服务器注册函数，而在 REST 中，服务器注册 URL 的处理程序。

哪个更好？都不是。哪个更快？都不是。受控环境哪个好？可能是 RPC。开放的环境哪个好？可能是 REST。

您将看到基于速度和资源分配的争论。基于二进制系统的 RPC 可能会比基于文本的 HTTP 系统更快。但是 SOAP 是使用 HTTP 的基于文本的 RPC 系统，可能比 REST 慢。HTTP2 使用二进制格式，在传送二进制数据(如 BSON)时，速度可能与其他二进制系统相当。更令人困惑的是，Apache Thrift RPC 允许选择数据格式(二进制、压缩二进制、JSON 和文本)和传输(套接字、文件和共享内存)。一个系统演示所有选项！

一个更重要的因素可能是操作环境的控制有多严格。RPC 系统是紧密耦合的，一个组件的故障可能会导致整个系统瘫痪。当只有一个管理机构、一组有限的硬件和软件配置以及一个清晰的解决问题的渠道时，RPC 系统就能很好地工作。

另一方面，网络是不受控制的。没有单一的权威机构——即使是像 DNS 这样的“通用”服务也是高度分散的。硬件、操作系统和软件种类繁多；几乎没有实施任何政策的前景；如果有什么东西坏了，通常没有人可以去修理它。在这种情况下，松散耦合的系统可能更好。

HTTP 上的 REST 很好地匹配了这一点。HATEOAS 允许服务器动态重新配置，根据需要改变 URL(甚至指向不同的服务器！).HTTP 被设计成在可能的时候缓存结果。防火墙通常被配置为允许 HTTP 流量并阻止大多数其他流量。在这里 REST 是个不错的选择。

应该注意的是，REST 并不是唯一可能的基于 HTTP 的系统。SOAP 已经提到了。有许多商业和非常成功的系统是“几乎”静止的——Richardson 1 级和 2 级。他们没有享受到 REST/HTTP 匹配的全部好处，但仍然可以工作。

毫无疑问，将来会出现其他模式。在物联网领域，CoAP 因低功耗无线系统而广受欢迎。它也是基于 REST 的，但与 HTTP-REST 略有不同。

## 结论

REST 是 web 的架构模型。它可以以许多不同的方式应用，特别是作为 HTTP 和 CoAP。本章演示了 REST 在 HTTP 中的应用。